#
# BEGIN CONFIGURATION
#
# YOU SHOULD INSTALL SuperLU first in order to solve Ax=b where A is a large
# sparse matrix.  After installing, we will know how to change following two
# lines.  
#
# SYSTEM should be SOMETHING that of $(SUPER)/superlu_(SOMETHING).a
# SUPER should be the path of the location of $(SUPER)/superlu_(SOMETHING).a
#
SYSTEM=sgi
SUPER=/user1/people/hunsoo/SuperLU
#
# END CONFIGURATION
#

LSUPER=$(SUPER)/superlu_$(SYSTEM).a $(SUPER)/blas_$(SYSTEM).a

LIBS=	../lib/common.c ../lib/complex.c \
	../lib/matrix.c ../lib/mat_single.c ../lib/mat_double.c \
	../lib/mat_simplex.c ../lib/mat_duplex.c \
	../lib/mat_ludecomp.c ../lib/mat_svdcmp.c ../lib/sp_2mat.c \
	../lib/sparse.c ../lib/sp_single.c ../lib/sp_double.c \
	../lib/sp_simplex.c ../lib/sp_duplex.c ../lib/sp_ops.c \
	../lib/sp_diag.c ../lib/sp_dotvector.c ../lib/sp_vector.c \
	../lib/vector.c ../lib/vec_garbage.c \
	../lib/vec_ops.c  ../lib/timestep.c

DRIVER=../drivers/m_randgen.c ../drivers/m_read.c ../drivers/m_getupper.c
LU=	../kern/super.c
SP_LU=	../sp_lu/sp_lupre.c ../sp_lu/sp_ludebug.c

SOLVERS=../solvers/solver.c ../solvers/iterations.c ../solvers/rtc.c \
	../solvers/iter.c ../solvers/conjugate.c ../solvers/it_jacob.c \
	../solvers/it_conjugate.c ../solvers/precond.c \
	../solvers/it_bicon.c ../solvers/it_qmr.c \
	../solvers/it_bicondstab.c

KERN =	../kern/proto.c ../kern/cubic8.c  ../kern/access.c\
	../kern/triangle3.c  ../kern/triangle6.c ../kern/square8.c \
	../kern/square4.c ../kern/square9.c \
	../kern/dirichlet.c ../kern/assemble.c

MESH=	../mesh/genm_linear.c ../mesh/genm_quadrat.c ../mesh/mesh_put.c\
	../mesh/dp_getmesh.c ../mesh/tri_getmesh.c\
	../mesh/dp_putmesh.c ../mesh/tri_putmesh.c\
        ../mesh/mesh.c ../mesh/mesh_buf.c \
	../mesh/mesh_reset.c ../mesh/mesh_putvector.c \
	../mesh/mesh_mkeinfo.c

OPTIONS=../options/options.c

HEADERS=../include/finite.h ../include/vector.h ../include/matrix.h \
	../include/common.h ../include/sparse.h \
	../include/options.h ../include/complex.h 

#	../include/solver.h ../include/spvector.h \
#	../include/sp_lu.h ../include/updown.h



ALL_SRC = $(LIBS) $(LU) $(SP_LU) $(KERN) $(MESH) $(OPTIONS) $(SOLVERS) $(DRIVER)
ALL_OBJ = $(ALL_SRC:.c=.o)
#../kern/super.o:
#	gcc -g -Wall -c -I$(SUPER) ../kern/super.c
#
# END OF LIST
#

.SUFFIXES: .c .o .d
CC=gcc
CFLAGS= -g -Wall -I../include -I$(SUPER)/SRC
CPPFLAGS=
LINT=lint
LINTFLAGS=
COMPILE.c=$(CC) $(CFLAGS) $(CPPFLAGS) -c
DEPEND.c=$(CC) $(CPPFLAGS)
LINK.c=$(CC) $(CFLAGS) $(CPPFLAGS) $(LDFLAGS) -target $(TARGET_ARCH:-%=%)
LINT.c=$(LINT) $(LINTFLAGS) $(CPPFLAGS) $(TARGET_ARCH)
#.c:
#	$(LINK.c) -o $@ $< $(LDLIBS)
.c.ln:
	$(LINT.c) $(OUTPUT_OPTION) -i $<
.c.o:
	$(COMPILE.c) $< -o $@
.c.d:
	$(DEPEND.c) -MMD $< 
.c.a:
	$(COMPILE.c) -o $% $<
	$(AR) $(ARFLAGS) $@ $%
	$(RM) $%
