#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "common.h"

void	gen_right(int no, int datatype, int sym, const char *filename);
void	gen_left(int no, int datatype, int sym, int band, const char *filename);
void	gen_left2nd(int no, int datatype, int sym, int (*bandfunc)(int, int),
		const char *filename);
static void gen_info(const char *funcname);
void
gen_randomfile(int no, int datatype, int sym, int band,
	const char *right, const char *left)
{
	time_t now;

	time(&now);
	srand(now);
	gen_info(__F);
	gen_left(no, datatype, sym, band, left);
	gen_right(no, datatype, sym, right);
	return;
}

void
gen_randomfile2nd(int no, int datatype, int sym, int (*bandfunc)(int, int),
	const char *right, const char *left)
{
	time_t now;

	time(&now);
	srand(now);
	gen_info(__F);
	gen_left2nd(no, datatype, sym, bandfunc, left);
	gen_right(no, datatype, sym, right);
	return;
}

static void
gen_info(const char *f)
{
	printf("%s(): A matrix and a vector consisted of Random numbers\n", f);
	printf("\twere created\n");
	return;
}

double
get_rand10(void)
{
	int	i;
	double	randno;
	unsigned int rin;

	rin = rand();
	for (i=1; i < rin; i <<= 1)
		/* void */;
	if (i != 0)
		randno = ((double) rin - (double) (i/2))/(double) i;
	if (rin % 2)
		randno *= -20.0;
	else
		randno *= +20.0;
	return randno;
}


void
gen_left2nd(int no, int datatype, int sym, int (*bandfunc)(int, int),
	const char *filename)
{
	FILE	*fp;
	int	i, j;
	double	r0, r1;

	printf("start of %s\n", __F);
	fp = fopen(filename, "w");
	if (fp == NULL)
		fatal("%s(%s): can not open", __F, filename);
	switch (datatype) {
	case SP_SINGLE:
	case SP_DOUBLE:
		if (datatype == SP_SINGLE)
			fprintf(fp, "%d %c\n", no, 's');
		else
			fprintf(fp, "%d %c\n", no, 'd');

		if (sym)
			for (i=1;i<=no;i++) {
				for (j=1;j <= i; j++) {
					if (bandfunc(i,j))
						continue;
					r0 = get_rand10();
					fprintf(fp, "%3d %3d %12.5e\n",
						i, j, r0);
					if (i != j)
					fprintf(fp, "%3d %3d %12.5e\n",
						j, i, r0);
				}
			}
		else
			for (i=1;i<=no;i++) {
				for (j=1;j <= no; j++) {
					if (bandfunc(i,j))
						continue;
					r0 = get_rand10();
					fprintf(fp, "%3d %3d %12.5e\n",
						i, j, r0);
				}
			}
		break;
	case SP_SIMPLEX:
	case SP_DUPLEX:
		if (datatype == SP_SIMPLEX)
			fprintf(fp, "%d %c\n", no, 'c');
		else
			fprintf(fp, "%d %c\n", no, 'z');

		switch (sym) {
		case 1: /* sym */
			for (i=1;i<=no;i++) {
				for (j=1;j <= i; j++) {
					if (bandfunc(i,j))
						continue;
					r0 = get_rand10();
					r1 = get_rand10();
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						i, j, r0, r1);
					if (i != j)
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						j, i, r0, r1);
				}
			}
			break;
		case 2: /* hermitian */
			for (i=1;i<=no;i++) {
				for (j=1;j <= i; j++) {
					if (bandfunc(i,j))
						continue;
					r0 = get_rand10();
					r1 = get_rand10();
					if (i == j) {
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						i, j, r0, 0.0);
						continue;
					}
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						i, j, r0,  r1);
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						j, i, r0, -r1);
				}
			}
			break;
		default:
			for (i=1;i<=no;i++) {
				for (j=1;j <= no; j++) {
					if (bandfunc(i,j))
						continue;
					r0 = get_rand10();
					r1 = get_rand10();
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						i, j, r0, r1);
				}
			}
		}
	}
	fclose(fp);
	printf("end   of %s\n", __F);
	return;
}

void
gen_left(int no, int datatype, int sym, int band,  const char *filename)
{
	FILE	*fp;
	int	i, j;
	double	r0, r1;

	printf("start of %s\n", __F);
	fp = fopen(filename, "w");
	if (fp == NULL)
		fatal("%s(%s): can not open", __F, filename);
	switch (datatype) {
	case SP_SINGLE:
	case SP_DOUBLE:
		if (datatype == SP_SINGLE)
			fprintf(fp, "%d %c\n", no, 's');
		else
			fprintf(fp, "%d %c\n", no, 'd');

		if (sym)
			for (i=1;i<=no;i++) {
				for (j=1;j <= i; j++) {
					if (band && (abs(i-j) > band))
						continue;
					r0 = get_rand10();
					fprintf(fp, "%3d %3d %12.5e\n",
						i, j, r0);
					if (i != j)
					fprintf(fp, "%3d %3d %12.5e\n",
						j, i, r0);
				}
			}
		else
			for (i=1;i<=no;i++) {
				for (j=1;j <= no; j++) {
					if (band && (abs(i-j) > band))
						continue;
					r0 = get_rand10();
					fprintf(fp, "%3d %3d %12.5e\n",
						i, j, r0);
				}
			}
		break;
	case SP_SIMPLEX:
	case SP_DUPLEX:
		if (datatype == SP_SIMPLEX)
			fprintf(fp, "%d %c\n", no, 'c');
		else
			fprintf(fp, "%d %c\n", no, 'z');

		switch (sym) {
		case 1: /* sym */
			for (i=1;i<=no;i++) {
				for (j=1;j <= i; j++) {
					if (band && (abs(i-j) > band))
						continue;
					r0 = get_rand10();
					r1 = get_rand10();
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						i, j, r0, r1);
					if (i != j)
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						j, i, r0, r1);
				}
			}
			break;
		case 2: /* hermitian */
			for (i=1;i<=no;i++) {
				for (j=1;j <= i; j++) {
					if (band && (abs(i-j) > band))
						continue;
					r0 = get_rand10();
					r1 = get_rand10();
					if (i == j) {
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						i, j, r0, 0.0);
						continue;
					}
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						i, j, r0,  r1);
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						j, i, r0, -r1);
				}
			}
			break;
		default:
			for (i=1;i<=no;i++) {
				for (j=1;j <= no; j++) {
					if (band && (abs(i-j) > band))
						continue;
					r0 = get_rand10();
					r1 = get_rand10();
					fprintf(fp, "%3d %3d %12.5e %12.5e\n",
						i, j, r0, r1);
				}
			}
		}
	}
	fclose(fp);
	printf("end   of %s\n", __F);
	return;
}


void
gen_right(int no, int datatype, int sym, const char *filename)
{
	FILE	*fp;
	int	i;
	double	r0, r1;

	
	printf("start of %s\n", __F);
	fp = fopen(filename, "w");
	if (fp == NULL)
		fatal("%s(%s): can not open", __F, filename);
	switch (datatype) {
		case SP_SINGLE:
		case SP_DOUBLE:
			if (datatype == SP_SINGLE)
				fprintf(fp, "%d %c\n", no, 's');
			else
				fprintf(fp, "%d %c\n", no, 'd');

			for (i=1;i<=no;i++) {
				r0 = get_rand10();
				fprintf(fp, "%12.5e\n", r0);
			}
			break;
		case SP_SIMPLEX:
		case SP_DUPLEX:
			if (datatype == SP_SIMPLEX)
				fprintf(fp, "%d %c\n", no, 'c');
			else
				fprintf(fp, "%d %c\n", no, 'z');

			for (i=1;i<=no;i++) {
				r0 = get_rand10();
				r1 = get_rand10();
				fprintf(fp, "%12.5e %12.5e\n", r0, r1);
			}
			break;
	}
	fclose(fp);
	printf("end   of %s\n", __F);
	return;
}
