#ifndef H_SOLVER_H
#define H_SOLVER_H

#define solver	/* just for ctags */

#include "float.h"
#include "vector.h"
#include "sparse.h"
#include "precond.h"

extern int verbose;

#define CONTINUE	1
#define STOPNOW		0

#define iszero(a) (fabs(a) < 10.0 * DBL_MIN)
#define isone(a)  (fabs(a - 1.0) < 10.0 * DBL_EPSILON)

#define iszzero(a) (iszero(zp_absolute(a))) /* complex version */
#define iszone(a)  (isone(a->real) && iszero(a->imag))


int	iter_check(int Iter, double rNorm, double bNorm);

void iter_bontriplets(const char *func, const struct sparse_t *sp,
	const struct vector_t *ap, const struct vector_t *bp);

struct vector_t *it_jacob(struct sparse_t *sp, struct vector_t *x,
	struct vector_t *b, int MaxIter, PrecondProcType Dummy, double Omega);
struct vector_t *it_jacob2nd(struct sparse_t *sp, struct vector_t *x,
	struct vector_t *b, int MaxIter, double omega);

struct vector_t *it_cgs(struct sparse_t *A, struct vector_t *x,
	struct vector_t *b, int MaxIter, PrecondProcType PrecondProc,
	double OmegaPrecond);


struct vector_t *it_solve(struct sparse_t *,struct vector_t *,
				struct vector_t *);
struct vector_t *it_congrad(struct sparse_t *,struct vector_t *,
				struct vector_t *, int);
int it_bicongrad(struct sparse_t *,struct vector_t *,
				struct vector_t *, int);
int it_bicongradcomplex(struct sparse_t *sp, struct vector_t *x,
				struct vector_t *b, int no);
void		 precond_simple(struct vector_t *xvp, const struct sparse_t *sp,
				struct vector_t *bvp);
void		 precond_simple2nd(struct vector_t *xp,
				const struct sparse_t *sp, struct vector_t *bp);
struct vector_t *it_congradreal(struct sparse_t *sp, struct vector_t *x,
			struct vector_t *b, int itno, double omega, int pcd_te);

/*
 * Now precondition types
 */
typedef struct vector_t *(*precondition)(struct sparse_t *,
		struct vector_t *, struct vector_t *, double);

struct vector_t *precond_jacob(struct sparse_t *A, struct vector_t *y,
				struct vector_t *c, double Omega);
struct vector_t *precond_ssor(struct sparse_t *A, struct vector_t *y,
				struct vector_t *c, double Omega);
struct vector_t *precond_ilu(struct sparse_t *A, struct vector_t *y,
				struct vector_t *c, double Omega);
#define PCD_JACOB	0x01	
#define PCD_SSOR	0x02	
#define PCD_ILU		0x03	

struct vector_t *
it_normalcongradcomplex(struct sparse_t *sp, struct vector_t *x,
	struct vector_t *b, int itno, double omega, int pcd_type);
struct vector_t *
it_normalcongradreal(struct sparse_t *sp, struct vector_t *x,
	struct vector_t *b, int itno, double omega, int pcd_type);

struct vector_t *
it_bicongradreal(struct sparse_t *sp, struct vector_t *x, struct vector_t *b,
	int itno, double omega, int pcd_type);

struct vector_t *
it_bicongradcom(struct sparse_t *sp, struct vector_t *x, struct vector_t *b,
	int itno, double omega, int pcd_type);
struct vector_t *
it_qmr(struct sparse_t *sp, struct vector_t *x, struct vector_t *b,
	int itno, double omega, int pcd_type);
struct vector_t *
it_bicongradstab(struct sparse_t *sp, struct vector_t *x, struct vector_t *b,
	int itno, double omega_notused, int pcd_type);
struct vector_t *
it_bicongradstabcomplex(struct sparse_t *sp, struct vector_t *x,
	struct vector_t *b, int itno, double omega_notused, int pcd_type);
#endif
