#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "common.h"
#include "vector.h"
#include "mesh.h"

/*
 * This routine will be applicable to only linear FEM
 */

struct vector_t *
vec_getderiv(struct vector_t *xp, const int which, const struct vector_t *bp)
{
	int	*checked; /* please free me */
	nodno_t *nnp;
	int	i, j, size;
	struct ement_t *ep;
	double	dderiv;
	struct duplex_t zderiv;

	float	*sdata;
	double	*ddata;
	struct simplex_t *cdata;
	struct duplex_t *zdata;

	vec_check(__F, xp, -1);
	vec_check(__F, bp, -1);

	if (vec_isdouplets(xp, bp))
		fatal("%s(): not good type vectors", __F);
	if (meshes->m_nodesno != xp->vt_size)
		fatal("%s(): meshes->m_nodesno != xp->vt_size", __F);

	size = sizeof(int) * xp->vt_size;
	checked = (int *) malloc(size);
	if (checked == NULL)
		memerr("%s(): while malloc checked", __F);
	memset(checked, 0x00, size);

	vec_null(xp);
	switch (xp->vt_type) {
		case SP_SINGLE:
			sdata = svec(xp);
			for (i=0; i< meshes->m_ementsno;i++) {
				ep = meshes->m_ements[i];
				nnp = ep->e_nodes;
				dderiv = dgetderivat(ep->e_no, 0, which, bp);
				for (j=0;j<ep->e_nodesno;j++) {
					checked[nnp[j]]++;
					sdata[nnp[j]] += dderiv;
				}
			}
			for (i=0;i<meshes->m_nodesno;i++)
				sdata[i] /= (float) checked[i];
			break;
		case SP_DOUBLE:
			ddata = dvec(xp);
			for (i=0; i< meshes->m_ementsno;i++) {
				ep = meshes->m_ements[i];
				nnp = ep->e_nodes;
				dderiv = dgetderivat(ep->e_no, 0, which, bp);
				for (j=0;j<ep->e_nodesno;j++) {
					checked[nnp[j]]++;
					ddata[nnp[j]] += dderiv;
				}
			}
			for (i=0;i<meshes->m_nodesno;i++)
				sdata[i] /= (double) checked[i];
			break;
		case SP_SIMPLEX:
			cdata = cvec(xp);
			for (i=0; i< meshes->m_ementsno;i++) {
				ep = meshes->m_ements[i];
				nnp = ep->e_nodes;
				zderiv = zgetderivat(ep->e_no, 0, which, bp);
				for (j=0;j<ep->e_nodesno;j++) {
					checked[nnp[j]]++;
					cdata[nnp[j]].real += zderiv.real;
					cdata[nnp[j]].imag += zderiv.imag;
				}
			}
			for (i=0;i<meshes->m_nodesno;i++) {
				cdata[i].real /= (double) checked[i];
				cdata[i].imag /= (double) checked[i];
			}
			break;
		case SP_DUPLEX:
			zdata = zvec(xp);
			for (i=0; i< meshes->m_ementsno;i++) {
				ep = meshes->m_ements[i];
				nnp = ep->e_nodes;
				zderiv = zgetderivat(ep->e_no, 0, which, bp);
				for (j=0;j<ep->e_nodesno;j++) {
					checked[nnp[j]]++;
					zdata[nnp[j]].real += zderiv.real;
					zdata[nnp[j]].imag += zderiv.imag;
				}
			}
			for (i=0;i<meshes->m_nodesno;i++) {
				zdata[i].real /= (double) checked[i];
				zdata[i].imag /= (double) checked[i];
			}
			break;
	}
	free(checked);
	return xp;
}
