#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include "common.h"
#include "vector.h"
#include "proto.h"
#include "mesh.h"

#define outgrids

/*
 * In the case of 2-dimensional, I want to output at mtv format
 */
void
putmtvform(const char *filename)
{
	FILE	*fp;
	int	i, j, nodeno;
	nodno_t	*nnp;
	double	*ddata;
	float	*fdata;
	struct node_t *np;
	struct ement_t *ep;
	char	*headers[]= {"$ DATA=CONTCURVE",
		"% contstyle=2  # 1=contour lines, 2=colors, 3=mesh",
		"% nsteps=50",
		"#% contours= 0.1 0.3 0.35  # each line value is specified",
		"#% comment=\"A comment in the upper right corner\"",
		"#% eyepos.x=0.5 eyepos.y=-1.5 eyepos.z=0.5 # another 3D view",
		NULL};
	char **cpp;


	if (meshes->m_spacedim != 2)
		return;

	fp = fopen(filename, "w");
	cpp = headers;

	while (*cpp)
		fprintf(fp, "%s\n", *cpp++);

	fprintf(fp, "\n");

	switch (meshes->m_datatype) {
	case SP_SINGLE:
		fdata = svec(meshes->m_bmatrix);
		for (i=0;i<meshes->m_ementsno;i++) {
			ep = meshes->m_ements[i];
			nnp = ep->e_nodes;
			for (j=0;j < ep->e_nodesno; j++) {
				nodeno = *nnp++;
				np = meshes->m_nodes[nodeno];
				fprintf(fp, "%7.4f %7.4f %7.4f\n",
					np->n_coor[0], np->n_coor[1],
					fdata[nodeno]);
			}
			fprintf(fp, "\n\n");
		}
		break;
	case SP_DOUBLE:
		ddata = dvec(meshes->m_bmatrix);
		for (i=0;i<meshes->m_ementsno;i++) {
			ep = meshes->m_ements[i];
			nnp = ep->e_nodes;
			for (j=0;j < ep->e_nodesno; j++) {
				nodeno = *nnp++;
				np = meshes->m_nodes[nodeno];
				fprintf(fp, "%7.4f %7.4f %7.4f\n",
					np->n_coor[0], np->n_coor[1],
					ddata[nodeno]);
			}
			fprintf(fp, "\n\n");
		}
		break;
	}
	fclose(fp);
	return;
}

/*
 * In the case of 3-dimensional, I want to output at mtv format
 * see plotmtv documentation
 */
void
putgrid4form(const struct vector_t *vp, const char *filename)
{
	FILE	*fp;
	int	i;
	float	*sdata;
	double	*ddata;
	char	**cpp;
	char	*headers[]= {"$ DATA=GRID4D",
		"% interp=2",
		"% axisscale=f",
		"% meshplot",
		"% nsteps=50",
		NULL,
	};
	char *minmax[3][2]={{"xmin", "xmax"}, {"ymin","ymax"}, {"zmin","zmax"}};
	char *nums[3]={"nx", "ny", "nz"};

	if (meshes->m_spacedim != 3) {
		printf("putgrid4form(): This is not 3-dim problem!!");
		return;
	}

	fp = fopen(filename, "w");
	if (fp == NULL)
		fatal("putgrid4form(\"%s\", \"%s\"): can not open file %s",
			vp->vt_name, filename, filename);

	cpp = headers;

	while (*cpp)
		fprintf(fp, "%s\n", *cpp++);

	fprintf(fp, "\n");

	if (m_for3dim)
		for (i=0;i<3;i++) {
			fprintf(fp, "%% ");
			fprintf(fp, "%s=%f ", minmax[i][0], m_coormin[i]);
			fprintf(fp, "%s=%f ", minmax[i][1], m_coormax[i]);
			fprintf(fp, "%s=%d\n",
				nums[i], m_for3dim->f3d_howmany[i]);
		}

	
	switch (meshes->m_datatype) {
		case SP_SINGLE:
			sdata = svec(vp);
			for (i=0;i<meshes->m_nodesno;i++)
				fprintf(fp, "%10.5e\n", sdata[i]);
			break;
		case SP_DOUBLE:
			ddata = dvec(vp);
			for (i=0;i<meshes->m_nodesno;i++)
				fprintf(fp, "%10.5e\n", ddata[i]);
			break;
	}
	fclose(fp);
	return;
}

/*
 * Only puts the grid information on filename
 * this function can deal only with 2-dimensional case.
 * if needno is not zero, then numbers of the nodes are given
 *
 * filename is a full name
 * Now you can view the mesh using plotmtv utility.
 */
 
void
put_2dgrids(const char *filename, const int needno)
{
	FILE	*fp;
	char	**cpp;
	int	marker, subdo, *lseg;

	int	i, j, nodeno;
	nodno_t	*nnp;
	struct node_t *np;
	struct ement_t *ep;
	struct proto_t *proto;
	char	*headers[]= {"$ DATA=CURVE2D", "% equalscale=True", NULL};
	char	*foreach="% filltype=1 linecolor=5";

	if (meshes->m_spacedim != 2)
		return;

	fp = fopen(filename, "w");

	cpp = headers;
	while (*cpp)
		fprintf(fp, "%s\n", *cpp++);
	fprintf(fp, "\n");

	for (i=0;i<meshes->m_ementsno;i++) {
		ep    = meshes->m_ements[i];
		nnp   = ep->e_nodes;
		proto = ep->e_proto;
		lseg  = proto->p_lineseg;

		subdo = (ep->e_sdomain % 10)+1;
		fprintf(fp, "%s fillcolor=%d\n", foreach, subdo);
		for (j=0;j < (proto->p_linesegno+1); j++) {
			nodeno = ep->e_nodesno;
			np = meshes->m_nodes[nnp[lseg[j]]];
			fprintf(fp, "%7.4f %7.4f\n",np->n_coor[0],
					np->n_coor[1]);
		}
		fprintf(fp, "\n\n");
	}
	for (i=0;i<meshes->m_nodesno;i++) {
		np = meshes->m_nodes[i];
		marker = 13;
		if (np->n_bdflag)
			marker = 4;
		if (needno)
		fprintf(fp,
		    "@ point x1=%7.4f y1=%7.4f markertype=%2d linelabel=%2d\n",
			np->n_coor[0], np->n_coor[1], marker, i);
		else
			fprintf(fp,"@ point x1=%7.4f y1=%7.4f markertype=%2d\n",
				np->n_coor[0], np->n_coor[1], marker);

	}
	fclose(fp);
	printf("mtv-formatted grid info was printed at %s\n", filename);
	return;
}
