#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include "common.h"
#include "proto.h"
#include "mesh.h"

struct genmesh_t {
	int	gn_spacedim;		/* spacedimension */
	float	gn_mins[3], gn_maxs[3];	/* min and max'es of mesh */
	int	gn_steps[3];		/* how many slices? */
	float	gn_grades[3];		/* grading... */
};

void	gmesh_quad2(const char *geometry, const char *partition);
void	gmesh_quad2proc(struct genmesh_t *gp);

void
quadratic(const char *geo, const char *part)
{
	char	*p;
	int	dim, readn;

	printf("+++++++++++++START of genmesh()+++++++++++++++++\n");
	if (meshes == NULL)
		fatal("%s(\"%s\", \"%s\"): you must call mesh_init first",
			__F, geo, part);
	if (meshes->m_nodes != NULL || meshes->m_ements != NULL) {
		fprintf(stderr, "%s():you can not call genmesh() 'cause ", __F);
		fprintf(stderr, "you already have meshes->m_nodes structure\n");
		exit(1);
	}

	p = index(geo, '=');
	if (p == NULL)
		fatal("%s(\"%s\", b): not good typing", __F, geo);

	readn = sscanf(++p, "%d", &dim);
	if (readn != 1 || dim < 2 || dim > 3)
		fatal("%s(\"%s\", b): not good typing", __F, geo);

	switch (dim) {
		case 2:
			meshes->m_spacedim = 2;
			gmesh_quad2(geo, part);
			break;
		case 3:
			printf("%s(): not implemented\n", __F);
			exit(0);
	}
	meshes->m_genmeshed = 1;
	printf("+++++++++++++END   of genmesh()+++++++++++++++++\n");
	return;
}


void
gmesh_quad2(const char *geo, const char *par)
{
	char	*geom, *part;
	char	*p;
	struct genmesh_t genm;
	int	i, readn;

	geom = strdup(geo);	/* please free me */
	part = strdup(par);
	if (geom == NULL || part == NULL)
		memerr("%s(): at strdup()ing", __F);

	p = index(geom, '[');	/* dont need to check p == NULL */
	for (i=0; i<2;i++) {
		readn = sscanf(p++, "[%f:%f]",
			&genm.gn_mins[i], &genm.gn_maxs[i]);
		if (readn != 2)
			fatal("%s(\"%s\", b): wrong typing", __F, geo);
		p = index(geom, '[');
	}

	p = index(part, '=');
	if (p == NULL)
		fatal("%s(\"%s\", \"%s\"): not good 2nd argument",
			__F, geo, par);

	readn = sscanf(++p, "%d", &i);
	if (i != 2)
		fatal("%s(\"%s\", \"%s\"): two have different dim", __F,
			geo, par);

	p = index(part, '[');
	if (p == NULL)
		fatal("%s(\"%s\", \"%s\"): bad 2nd argum", __F, geo, par);

	readn = sscanf(p, "[%d:%d]", &genm.gn_steps[0], &genm.gn_steps[1]);
	if (readn != 2)
		fatal("gmesh_quad2(\"%s\", \"%s\"): bad 2nd argum", geo, par);

	if (genm.gn_steps[0] < 1 || genm.gn_steps[1] < 1)
		fatal("%s(\"%s\", \"%s\"): bad steps", __F, geo, par);

	/*
	 * read gradings
	 */
	p++;
	p = index(p, '[');
	readn = sscanf(p, "[%f:%f]", &genm.gn_grades[0], &genm.gn_grades[1]);
	if (readn != 2)
		fatal("gmesh_quad2(\"%s\", \"%s\"): bad grading", geo, par);

	/*
	 * The end of reading of geometry and partitions
	 */
	gmesh_quad2proc(&genm);
	free((void *) geom);
	free((void *) part);
	return;
}

/*
 * Do not call this function explicitly, 'cause this is a sort of private
 * function of gmesh_quad2();
 */
void
gmesh_quad2proc(struct genmesh_t *gp)
{
	int	i, j, total, cnt, steps[2];
	int	hasbd_x, hasbd_y;	/* for ement_t->e_hasbd; */

	float	mins[2], maxs[2], units[2], grading[2], plus[2], minus[2];
	float	x, y, x1, y1, var;
	nodno_t *nnp;
	struct ement_t *ep;
	int	ix, indx, xchecked=1;
	float	*xcoor, *ycoor;
	ushort	*xminnodes, *yminnodes;

	total = (2*gp->gn_steps[0] + 1) * (2*gp->gn_steps[1] + 1);

	/*
	 * To facilate the calculation of...
	 */
	meshes->m_for3dim = (struct for3dimcase *)
		malloc(sizeof(struct for3dimcase));

	for (i=0; i < 2; i++) {
		meshes->m_for3dim->f3d_howmany[i] = (2*gp->gn_steps[i] + 1);
		meshes->m_for3dim->f3d_nonodes[i] = (2*gp->gn_steps[(i+1)%2]+1);
		meshes->m_for3dim->f3d_coors[i] = (float *)
		    malloc(sizeof(float)*meshes->m_for3dim->f3d_howmany[i]);
		meshes->m_for3dim->f3d_minnodes[i] = (ushort *)
		    malloc(sizeof(ushort)*meshes->m_for3dim->f3d_howmany[i]);
	}

	meshes->m_for3dim->f3d_nincres[0] =  1;
	meshes->m_for3dim->f3d_nincres[1] = 2*meshes->m_for3dim->f3d_howmany[0];

	meshes->m_for3dim->f3d_eincres[0] =  1;
	meshes->m_for3dim->f3d_eincres[1] = gp->gn_steps[0];
		/* (meshes->m_for3dim->f3d_howmany[0] - 1); */

	/*
	 * real allocation of the node information
	 */
	meshes->m_nodes = mesh_nodesbuf(total);
	meshes->m_nodesfreeme = meshes->m_nodes[0];

	for (i=0;i<2;i++) {
		grading[i]= gp->gn_grades[i];
		steps[i]  = gp->gn_steps[i];
		mins[i]   = gp->gn_mins[i];
		maxs[i]   = gp->gn_maxs[i];
		plus[i]   = maxs[i] + mins[i];
		minus[i]  = maxs[i] - mins[i];
		units[i]  = 0.5 * minus[i] / (float) steps[i];
	}

	xcoor = meshes->m_for3dim->f3d_coors[0];
	ycoor = meshes->m_for3dim->f3d_coors[1];
	xminnodes = meshes->m_for3dim->f3d_minnodes[0];
	yminnodes = meshes->m_for3dim->f3d_minnodes[1];

	cnt = 0;
	xchecked = 1;

	/*
	 * NODE
	 */
	for (i=0; i < (2*gp->gn_steps[1]+1); i++) { /* Y */
		y  = mins[1] + units[1] * (float) i;
		y1 = y;
#if TEMPO
		var = 2.0 * y - plus[1];
		y1 = 0.5 * sign(var) * minus[1]
			* pow(fabs(var/minus[1]), 1.0/grading[1])
			+ 0.5 * plus[1];
#endif

		ycoor[i] = y1;
		yminnodes[i] = cnt;

		if (i == 0)
			meshes->m_coormin[1] = y1;
		if (i == (2*gp->gn_steps[1]))
			meshes->m_coormax[1] = y1;

		for (j=0; j < (2*gp->gn_steps[0]+1); j++) { /* X */
			x  = mins[0] + units[0] * (float) j;
			x1 = x;
#if TEMPO
			var = 2.0 * x - plus[0];
			x1 = 0.5 * sign(var) * minus[0]
				* pow(fabs(var/minus[0]), 1.0/grading[0])
				+ 0.5 * plus[0];
#endif
			if (xchecked) {
				xcoor[j] = x1;
				xminnodes[j] = cnt;

				if (j == 0)
					meshes->m_coormin[0] = x1;
				if (j == (2*gp->gn_steps[0]))
					meshes->m_coormax[0] = x1;
			}

			meshes->m_nodes[cnt]->n_coor[0] = x1;
			meshes->m_nodes[cnt]->n_coor[1] = y1;

			meshes->m_nodes[cnt]->n_plane[0] = j;	/* X */
			meshes->m_nodes[cnt]->n_plane[1] = i;	/* Y */

			meshes->m_nodes[cnt]->n_no = cnt;

			/*
			 * boundary manipulations
			 *      |
			 *      | y  2 
			 *      +---------+
			 *      |         |
			 *      |         |
			 *    3 |         |  1
			 *      |         |
			 *      |         |
			 *      +---------+----> x
			 *           0
			 */
			if (i == 0)
				setbit(meshes->m_nodes[cnt]->n_bdflag, 0);
			if (i == (2*gp->gn_steps[0]))
				setbit(meshes->m_nodes[cnt]->n_bdflag, 2);
			if (j == 0)
				setbit(meshes->m_nodes[cnt]->n_bdflag, 3);
			if (j == (2*gp->gn_steps[1]))
				setbit(meshes->m_nodes[cnt]->n_bdflag, 1);

			/* XXX */
			cnt++;
		}
		xchecked = 0;
	}
	meshes->m_nodesno = cnt;

	/*
	 * start of ement structure
	 * 
	 *   6----7----8
	 *   |    |    |
	 *   |    |    |
	 *   3----4----5
	 *   |    |    |
	 *   |    |    |
	 *   0----1----2
	 */

	total = (gp->gn_steps[0]) * (gp->gn_steps[1]);
	meshes->m_ements = mesh_ementsbuf(total);
	meshes->m_ementsfreeme = meshes->m_ements[0];

	cnt = 0;
	for (i=0; i < gp->gn_steps[1]; i++) {

		ix = i * 2 * (2*gp->gn_steps[0]+1);

		if (i == 0 || i == (gp->gn_steps[1]-1))
			hasbd_x = 1;
		else
			hasbd_x = 0;

		for (j=0; j < gp->gn_steps[0]; j++) {
			indx = ix + 2*j;
			ep = meshes->m_ements[cnt];
			ep->e_proto = getproto("ElmB9n2D");
			ep->e_nodesno = 9;
			ep->e_no = cnt;
			nnp = ep->e_nodes = (nodno_t *)
				malloc(sizeof(nodno_t) * 9);
			if (ep->e_nodes == NULL)
				fatal("%s(): Hmm", __F);
			*nnp++ = indx;
			*nnp++ = indx + 1;
			*nnp++ = indx + 2;
			*nnp++ = indx + (2*gp->gn_steps[0]+1) + 0;
			*nnp++ = indx + (2*gp->gn_steps[0]+1) + 1;
			*nnp++ = indx + (2*gp->gn_steps[0]+1) + 2;
			*nnp++ = indx + (4*gp->gn_steps[0]+2) + 0; /* XXX */
			*nnp++ = indx + (4*gp->gn_steps[0]+2) + 1; /* XXX */
			*nnp++ = indx + (4*gp->gn_steps[0]+2) + 2; /* XXX */

			if (j==0 || j==(gp->gn_steps[0] - 1))
				hasbd_y = 1;
			else
				hasbd_y = 0;

			if (hasbd_x || hasbd_y)
				ep->e_hasbd = 1;
			else
				ep->e_hasbd = 0;

			/*
			 * XXX
			 */
			cnt++;
		}
	}
	meshes->m_ementsno = cnt;

	return;
}
