#include <stdio.h>
#include "common.h"
#include "vector.h"
#include "matrix.h"
#include "proto.h"
#include "mesh.h"

static void     tri6_setunit(void);
static void	tri6_integpoints(void);
static void	tri6_mkhalfjacobm(void);
static void	tri6_mktrans(void);
static double	tri6_ugetall(int no);

struct integral_t *tri6_intg;

int
tri6_init(struct proto_t *p)
{
	int	*lseg;
	static int tri6_inited = 0;

	if (tri6_inited)
		return 0;

	tri6_inited++;
	tri6_intg = p->p_intg;

	/*
	 * ORDER is important
	 */
	tri6_setunit();
	tri6_integpoints();
	tri6_mkhalfjacobm();
	tri6_mktrans();

	lseg = p->p_lineseg;
	lseg[0] = 0; lseg[1] = 1; lseg[2] = 2; lseg[3] = 0;

	printf("END in tri6_init()--------THIS IS INCOMPLETE!!!\n");

	return 0;	/* 0 means special meaning!! dont change it */
}

/*
 *    o 2
 *    |\
 *    | \
 *  4 o  o 3
 *    |   \
 *    |    \
 *    o--o--o----ksi
 *    0  5  1
 */

static void
tri6_setunit(void)
{
	struct matrix_t *matf;
	double	**mat;

	matf = tri6_intg->intg_unit;
	mat  = dmat(matf);

	mat[0][0] =  0.0; mat[0][1] =  0.0;
	mat[1][0] =  0.0; mat[1][1] =  1.0;
	mat[2][0] =  1.0; mat[2][1] =  0.0;
	mat[3][0] =  0.5; mat[3][1] =  0.0;
	mat[4][0] =  0.0; mat[4][1] =  0.5;
	mat[5][0] =  0.5; mat[5][1] =  0.5;

	return;
}


void
tri6_integpoints(void)
{
	double	**intgcoor;
	double	 *weights;
	int	i;

	weights = tri6_intg->intg_weights;
	intgcoor = dmat(tri6_intg->intg_intgcoors);
	for (i=0; i < 3;i++)
		weights[i] = 1.0/6.0;

	intgcoor[0][0] = 1.0/2.0; intgcoor[0][1] = 0.0/2.0;
	intgcoor[1][0] = 0.0/2.0; intgcoor[1][1] = 1.0/2.0;
	intgcoor[2][0] = 1.0/2.0; intgcoor[2][1] = 1.0/2.0;
	return;
}

static void
tri6_mktrans(void)
{
	double	**mat;
	double	**unit, **intp;
	double	elle1, elle2, elle3;
	double	ksi, eta;
	int	i;

	mat = dmat(tri6_intg->intg_trans);
	intp = dmat(tri6_intg->intg_intgcoors);
	unit = dmat(tri6_intg->intg_unit);

	for (i=0; i < 3; i++) {	/* for three points */
		ksi = intp[i][0];
		eta = intp[i][1];

		elle1 = 1.0 - ksi - eta;
		elle2 = ksi;
		elle3 = eta;

		mat[i][0] = elle1 * (2.0 * elle1 - 1.0);
		mat[i][1] = elle2 * (2.0 * elle2 - 1.0);
		mat[i][2] = elle3 * (2.0 * elle3 - 1.0);

		mat[i][5] = 4.0 * elle1 * elle2;
		mat[i][3] = 4.0 * elle2 * elle3;
		mat[i][4] = 4.0 * elle3 * elle1;
	}
	m_put(tri6_intg->intg_trans);
	return;
}

static void
tri6_mkhalfjacobm(void)
{
	double	ksi, eta, elle1, elle2, elle3;
	double	**intp, **unit;
	struct matrix_t	*fmat;
	double	**fmatp;
	int	i;

	unit = dmat(tri6_intg->intg_unit);
	intp = dmat(tri6_intg->intg_intgcoors);
	for (i=0; i<3;i++) {
		fmat = tri6_intg->intg_halfjacobm[i];
		fmatp = dmat(fmat);

		ksi = intp[i][0];
		eta = intp[i][1];

		elle1 = 1.0 - ksi - eta;
		elle2 = ksi;
		elle3 = eta;

		fmatp[0][0] = -1.0 * (2.0 * elle1 - 1.0) + elle1 * (-2.0);
		fmatp[1][0] = -1.0 * (2.0 * elle1 - 1.0) + elle1 * (-2.0);
		fmatp[0][1] = (2.0 * elle2 - 1.0) + elle2 * 2.0;
		fmatp[1][1] = 0.0;
		fmatp[0][2] = 0.0;
		fmatp[1][2] = (2.0 * elle3 - 1.0) + elle3 * 2.0;

		fmatp[0][5] = 4.0 * (-1.0 * elle2 + elle1);
		fmatp[1][5] = 4.0 * (-1.0 * elle2);
		fmatp[0][3] = 4.0 * (elle3);
		fmatp[1][3] = 4.0 * (elle2);
		fmatp[0][4] = 4.0 * (-1.0 * elle3);
		fmatp[1][4] = 4.0 * (-1.0 * elle3 + elle1);
	}
	return;
}
		
/*
 * Once you have called tri6_einit() for a given elem, there is no further
 * calculation about the internal of that elements.  This routine is very
 * important because this routine gets the most important imformation of a
 * given element number.  This information will be saved to the global data
 */
void
tri6_einit(const int elemno)
{
	int	i;
	nodno_t	*nnp;
	double	**fpp;
	struct ement_t	*ep;
	struct node_t	*np;

	if (elemno < 0 || elemno >= meshes->m_ementsno)
		fatal("tri6_nodeinit() with OOB nodeno=%d\n", elemno);
	tri6_intg->intg_elemno = elemno;
	ep = meshes->m_ements[elemno]; 

	if (ep->e_nodesno != 6)
		fatal("%s(): OOPS!! THIS IS IMPOSSIBLE!!!", __F);

	nnp = ep->e_nodes;	/* seek the boundary nodes */

	fpp = dmat(tri6_intg->intg_coors);

	for (i=0; i < 6; i++, nnp++) {
		np = meshes->m_nodes[*nnp];
		fpp[i][0] = np->n_coor[0];
		fpp[i][1] = np->n_coor[1];
	}
	m_dot(tri6_intg->intg_realintgcoors, tri6_intg->intg_trans,
		tri6_intg->intg_coors);

	for (i=0; i < 3; i++) {
		tri6_intg->intg_jacob[i] = tri6_ugetall(i);
		printf("%6.3f ", tri6_intg->intg_jacob[i]);
	}
	printf("\n");

	return;
}


static double
tri6_ugetall(int no)
{
	double  det;            /* Determinant of jacob */

	m_dot(tri6_intg->intg_jacobm[no],
		tri6_intg->intg_halfjacobm[no], tri6_intg->intg_coors);

	det = dm_inverse(tri6_intg->intg_tmp, tri6_intg->intg_jacobm[no]);

	/* tem is now inverse of jacob at this no integration point */
	/* Let's calculate (9.42) */

	m_dot(tri6_intg->intg_deriv[no],
		tri6_intg->intg_tmp, tri6_intg->intg_halfjacobm[no]);

	return det;
}
