#include <stdio.h>
#include <math.h>

#include "common.h"
#include "vector.h"
#include "matrix.h"
#include "sparse.h"
#include "finite.h"

struct vector_t *avector;
struct vector_t *fvector;
void	radial(const char *fname);


void
problem(void)
{
	int	i;
	double	omega;

	struct matrix_t	*retmatf, *bmatf;
	struct proto_t	*prevproto = NULL;

	omega = 1.0e-40;
	if (!IS_NULL(currp))
		fatal("currp != NULL; What's wrong!! ");

	avector = vector(meshes->m_nodesno, "a vector", SP_DOUBLE);
	fvector = vector(meshes->m_nodesno, "f vector", SP_DOUBLE);


	dvec_set(avector, 1.0);
	dvec_set(fvector, 0.0);

	put_currtime(__F, NULL);
	for (i=0; i< meshes->m_ementsno; i++) {
		/*
		 * currp is only set at this location. You must not set currp
		 * at other places.
		 */
		currp = meshes->m_ements[i]->e_proto;
		if (prevproto != currp) {
			if (prevproto != NULL)
				m_mfree(2, retmatf, bmatf);/* no longer used */
			create_smalls(currp, &retmatf, &bmatf);
		}

		integration(i, retmatf, bmatf);
		/*
		 * Following codes are dependent on problem
		 */
		Kassemble(i, retmatf, meshes->m_kmatrix);
		bassemble(i, bmatf, meshes->m_bvector);
		prevproto = currp;
	}
	m_mfree(2, retmatf, bmatf);
	vec_mfree(2, avector, fvector);

	printf("before BC_impose\n");
	switch (meshes->m_datatype) {
		case SP_SINGLE:
		case SP_DOUBLE:
			ddirichlet(0, -10.0);
			ddirichlet(1, +10.0);
			break;
	}
	printf("after  BC_impose\n");
	put_currtime(__F, NULL);
	
	solveall();
	put_currtime(__F, NULL);

	put_gridandvector(meshes->m_bvector, "a.mtv");
	radial("radial.rad");
	return;
}

void
integration(const int elemno, struct matrix_t *kmatf, struct matrix_t *bmatf)
{
	int	i, j, k, l;
	double	**kmat, **bmat, **unit, **deriv;
	double	fval, aval, temp;
	struct integral_t *intg;
	double	*weights, *jacob;

	m_mnull(2, kmatf, bmatf); /* XXX */
	kmat   = dmat(kmatf);
	bmat   = dmat(bmatf);

	proto_einit(elemno); /* XXX */
	intg    = currp->p_intg;
	unit    = dmat(intg->intg_trans);
	jacob   = intg->intg_jacob;
	weights = intg->intg_weights;

	for (k=0; k<currp->p_intgno; k++) {	/* four integration points */
		deriv = dmat(intg->intg_deriv[k]);
		aval = dgetat(elemno, k, avector);
		fval = dgetat(elemno, k, fvector);

		for (i=0; i< currp->p_nodsno;i++) {
			for (j=0; j< currp->p_nodsno;j++) {
				temp = 0.0;
				for (l=0;l<currp->p_dimen;l++)
					temp += deriv[l][i] * deriv[l][j];

				kmat[i][j] += (aval*temp*jacob[k]*weights[k]);
			}
			bmat[i][0] +=
				-unit[k][i]*fval*jacob[k]*weights[k];
		}
	}
	if ((elemno%100) == 0) {
		printf("%6d\n", elemno);
	}

	return;
}

void
radial(const char *fname)
{
	FILE	*fp;
	int	i;
	float	x, y, len, *spp;
	struct node_t *np;

	fp = fopen(fname, "w");
	if (fp == NULL || meshes->m_datatype != SP_SINGLE)
		fatal("%s(): ", __F);

	spp = svec(meshes->m_bvector);
	for (i=0; i < meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		x = np->n_coor[0]; y = np->n_coor[1];
		len = sqrt(x*x+y*y);
		fprintf(fp, "%7.4f %7.4f\n", len, spp[i]);
	}
	fclose(fp);
	return;
}
