#include <stdio.h>
#include <math.h>

#include "common.h"
#include "vector.h"
#include "matrix.h"
#include "sparse.h"
#include "finite.h"

#define TINY 1.0e-10

int
whichbound(const struct node_t *np)
{
	float   x, y;

	x = np->n_coor[0];
	y = np->n_coor[1];

	if (fabs(x - (-1.0)) < TINY)
		return 3;
	if (fabs(x - (+1.0)) < TINY)
		return 1;
	if (fabs(y - (-1.0)) < TINY)
		return 0;
	if (fabs(y - (+1.0)) < TINY)
		return 2;
	return -1;
}

extern struct proto_t	*currp;

void
problem(void)
{
	int	i;

	struct matrix_t	*retmatf, *bmatf;
	struct proto_t	*prevproto = NULL;

	set_boundary(whichbound);
	put_2dgrids("grids.mtv", 0);

	for (i=0; i< meshes->m_ementsno; i++) {
		/*
		 * currp is only set at this location. You must
		 * not set currp at other places.
		 */
		currp = meshes->m_ements[i]->e_proto;
		if (prevproto != currp) {
			if (prevproto != NULL)
				m_mfree(2, retmatf, bmatf);
			create_smalls(currp, &retmatf, &bmatf);
		}

		integration(i, retmatf, bmatf);
		Kassemble(i, retmatf, meshes->m_kmatrix);
		bassemble(i, bmatf, meshes->m_bvector);
		prevproto = currp;
	}
	m_mfree(2, retmatf, bmatf);

	switch (meshes->m_datatype) {
		case SP_SINGLE:
		case SP_DOUBLE:
			ddirichlet(0, +10.0);
			ddirichlet(1, -10.0);
			ddirichlet(2, +10.0);
			ddirichlet(3, -10.0);
		break;
	}
	solveall();
	put_gridandvector(meshes->m_bvector, "a.mtv");
	return;
}

void
integration(const int elemno, struct matrix_t *kmatp, struct matrix_t *bmatp)
{
	int     i, j, k, l;
	double  **retmat, **bmat;
	double	**unit, **deriv, *weights, *jacob;
	double  beta, alpha, temp;

	m_mnull(2, kmatp, bmatp);
	retmat = dmat(kmatp);
	bmat   = dmat(bmatp);

	proto_einit(elemno);
	unit    = dmat(currp->p_intg->intg_trans);
	weights = currp->p_intg->intg_weights;
	jacob   = currp->p_intg->intg_jacob;

	currp = currp;
	for (k=0; k<currp->p_intgno; k++) {
		deriv = dmat(currp->p_intg->intg_deriv[k]);
		alpha = 1.0;
		beta  = 0.0;

		for (i=0; i< currp->p_nodsno;i++) {
			for (j=0; j< currp->p_nodsno;j++) {
				temp = 0.0;
				for (l=0;l<currp->p_dimen;l++)
					temp += deriv[l][i] * deriv[l][j];

				retmat[i][j] += alpha * temp * jacob[k]
					* weights[k];
			}
			bmat[i][0] +=
				-unit[k][i] * beta * jacob[k] * weights[k];
		}
	}
	if ((elemno%100) == 0) {
		printf("%6d\n", elemno);
	}

	return;
}
