#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <math.h>

#include "common.h"
#include "vector.h"
#include "matrix.h"
#include "sparse.h"
#include "finite.h"

struct vector_t *avector;
struct vector_t *fvector;


void
problem(void)
{
	int	i;
	time_t	currtime;
	double	omega;

	struct matrix_t	*retmatf, *bmatf;
	struct proto_t	*prevproto = NULL;

	omega = 1.0e-40;

	if (currp != NULL)
		fatal("%s():currp != NULL; What's wrong!! ", __F);

	avector = vector(meshes->m_nodesno, "a vector", SP_DOUBLE);
	fvector = vector(meshes->m_nodesno, "f vector", SP_DOUBLE);

	dvec_set(avector, 1.0);
	dvec_set(fvector, 0.0);

	time(&currtime);
	printf("\nStart Assemble TIME=%s", ctime(&currtime)); 

	for (i=0; i< meshes->m_ementsno; i++) {
		/*
		 * currp is only set at this location. You must not set currp
		 * at other places.
		 */
		currp = meshes->m_ements[i]->e_proto;
		if (prevproto != currp) {
			if (prevproto != NULL)
				m_mfree(2, retmatf, bmatf);/* no longer used */
			create_smalls(currp, &retmatf, &bmatf);
		}

		integration(i, retmatf, bmatf);
		/*
		 * Following codes are dependent on problem
		 */
		Kassemble(i, retmatf, meshes->m_kmatrix);
		bassemble(i, bmatf, meshes->m_bvector);
		prevproto = currp;
	}
	m_mfree(2, retmatf, bmatf);
	vec_mfree(2, avector, fvector);

	printf("before BC_impose\n");
	switch (meshes->m_datatype) {
		case SP_SINGLE:
		case SP_DOUBLE:
			ddirichlet(0, +10.0);
			ddirichlet(1, +10.0);
			ddirichlet(2, -10.0);
			ddirichlet(3, -10.0);
			ddirichlet(4, -10.0);
			ddirichlet(5, +10.0);
			break;
	}
	printf("after  BC_impose\n");
	sp_symcheck(meshes->m_kmatrix);
	solveall();

	if (0)
		vec_printatfile(meshes->m_bvector, "bmat");

	put_gridandvector(meshes->m_bvector, "a.mtv");
	return;
}

void
integration(const int elemno, struct matrix_t *kmatf, struct matrix_t *bmatf)
{
	int	i, j, k, l;
	double	**kmat, **bmat, **unit, **deriv;
	double	fval, aval, temp;
	struct integral_t *intg;
	double	*jacob, *weights;

	m_mnull(2, kmatf, bmatf);
	kmat   = dmat(kmatf);
	bmat   = dmat(bmatf);

	proto_einit(elemno); /*  the integral_t */
	intg   = currp->p_intg;
	unit   = dmat(intg->intg_trans);
	jacob  = intg->intg_jacob;
	weights= intg->intg_weights;

	currp = currp;
	for (k=0; k<currp->p_intgno; k++) {	/* four integration points */
		deriv = dmat(intg->intg_deriv[k]);
		aval = dgetat(elemno, k, avector);
		fval = dgetat(elemno, k, fvector);

		for (i=0; i< currp->p_nodsno;i++) {
			for (j=0; j< currp->p_nodsno;j++) {
				temp = 0.0;
				for (l=0;l<currp->p_dimen;l++)
					temp += deriv[l][i] * deriv[l][j];

				kmat[i][j] += (aval*temp*jacob[k]*weights[k]);
			}
			bmat[i][0] += (-unit[k][i]*fval*jacob[k]*weights[k]);
		}
	}
	if ((elemno%100) == 0) {
		printf("%6d\n", elemno);
	}

	return;
}
