#ifndef H_COMMON_H
#define H_COMMON_H

extern int verbose;

#include <sys/types.h>

#define __F	__FUNCTION__

/* for SuperLU */
#define __S	0
#define __D	1
#define __C	2
#define __Z	3

#define	SP_NONE		0x00000000
#define	SP_INTEGER	0x00000001
#define	SP_SINGLE	0x00000002
#define	SP_DOUBLE	0x00000003
#define	SP_SIMPLEX	0x00000004
#define	SP_DUPLEX	0x00000005

extern char *typenames[];

#define ATYPE(type)	  (type == SP_INTEGER \
			|| type == SP_SINGLE \
			|| type == SP_DOUBLE \
			|| type == SP_SIMPLEX \
			|| type == SP_DUPLEX)

#define UNKNOWN "UNKNOWN"
#define BITPERINTEGER 32

#ifndef setbit
#define setbit(a,i)     ((a) |= 1<<((i)%BITPERINTEGER))
#define clrbit(a,i)     ((a) &= ~(1<<((i)%BITPERINTEGER)))
#define isset(a,i)      ((a) & (1<<((i)%BITPERINTEGER))) 
#define isclr(a,i)      (((a) & (1<<((i)%BITPERINTEGER))) == 0)  
#endif

#define IS_NULL(p)	(p == NULL)
#define FTYPE   "%15.10f "
#define ITYPE   "%8d+ "
#define CUTOFF  1.0e-30
#define TOKEN(c)        (isalnum(c) || c == '_')

#define SETTING 0x0001232
#define GETTING 0x0001233
#define INTCMP(a, b) ((a) > (b) ? 1 : ((a) < (b)) ? -1 : 0)

extern char *ord[3];
extern char commonbuf[];
extern	int	malloc_all;
void	mstats(char *);

typedef unsigned short nodno_t;
typedef unsigned short elmno_t;

struct simplex_t;
struct duplex_t;


#ifdef _sgi
extern FILE	*tmpfile(void);
extern char	*tmpnam(char *);
extern int	fclose(FILE *);
extern int	fflush(FILE *);
extern FILE	*fopen(const char *, const char *);
extern FILE	*freopen(const char *, const char *, FILE *);
extern void	setbuf(FILE *, char *);
extern int	setvbuf(FILE *, char *, int, size_t);
extern int	fprintf(FILE *, const char *, ...);
extern int	fscanf(FILE *, const char *, ...);
extern int	printf(const char *, ...);
extern int	scanf(const char *, ...);
extern int	sscanf(const char *, const char *, ...);
extern int	vfprintf(FILE *, const char *, /* va_list */ char *);
extern int	vprintf(const char *, /* va_list */ char *);
extern int	vsprintf(char *, const char *, /* va_list */ char *);
extern size_t	fread(void *, size_t, size_t, FILE *);
extern size_t	fwrite(const void *, size_t, size_t, FILE *);
extern int	fseek(FILE *, long, int);
extern long	ftell(FILE *);
#endif

char	*get_ordinal(const int order);
void	 fatal(const char *fmt, ...);
void	 memerr(const char *fmt, ...);
void	 err(const char *fmt, ...);
int	 islineok(const char *);
char	*skipto(const char *, char);
char	*skipwhites(char *);
char	*skipnumer(char *);
void	 bitput(int, char *);
void	 patt_rgetpatt(char *t, char *sp, const char *patt, const char *term);
void	 patt_cgetpatt(char *t, char *sp, const char *patt, const char *term);
void	 not_yet(const char *function);
void	 put(double x);
double	 pythag(double a, double b);
float	 sign(float x);
void	 put_currtime(const char *func, const char *msg);
char	*get_currtime(void);

#endif
