#ifndef H_COMPLEX_H
#define H_COMPLEX_H

struct simplex_t {
	float	real;
	float	imag;
};

struct duplex_t {
	double	real;
	double	imag;
};

extern struct duplex_t duplex_null;
extern struct duplex_t duplex_unit;

extern struct simplex_t simplex_null;
extern struct simplex_t simplex_unit;


/* creator */
struct duplex_t   duplexv(const double, const double);
struct simplex_t  simplexv(const double, const double);

struct duplex_t  *zp_assign(struct duplex_t *ret, const struct duplex_t *src);
struct simplex_t *cp_assign(struct simplex_t *ret, const struct simplex_t *src);

struct duplex_t *zp_expon(struct duplex_t *ret, const struct duplex_t *z);
struct duplex_t *zp_conjug(struct duplex_t *ret, const struct duplex_t *src);
void	zp_print(const struct duplex_t *);


struct duplex_t *zp_inverse(struct duplex_t *ret, const struct duplex_t *src);

struct duplex_t *zp_plus(struct duplex_t *ret, const struct duplex_t *a,
				const struct duplex_t *b);
struct duplex_t *zp_pplus(struct duplex_t *ret, const struct duplex_t *a);

struct duplex_t *zp_3plus(struct duplex_t *ret, const struct duplex_t *a,
			const struct duplex_t *b, const struct duplex_t *c);

struct duplex_t *zp_mminus(struct duplex_t *ret, const struct duplex_t *a);

struct duplex_t *zp_minus(struct duplex_t *ret, const struct duplex_t *a,
				const struct duplex_t *b);

struct duplex_t *zp_cross(struct duplex_t *ret, const struct duplex_t *a,
				const struct duplex_t *b);

struct duplex_t *zp_3cross(struct duplex_t *ret, const struct duplex_t *a,
		const struct duplex_t *b, const struct duplex_t *c);

struct duplex_t *zp_crossdme(struct duplex_t *a, const double b);
struct duplex_t *zp_crosszme(struct duplex_t *ret, const struct duplex_t b);

struct duplex_t *zp_crossd(struct duplex_t *ret, const struct duplex_t *a,					const double b);

struct duplex_t *zp_slash(struct duplex_t *ret, const struct duplex_t *a,
				const struct duplex_t *b);

double		zp_absolute(const struct duplex_t *a);
double		zp_square(const struct duplex_t *a);

struct duplex_t *zp_conjcross(struct duplex_t *ret, const struct duplex_t *a,
			const struct duplex_t *b);



struct simplex_t *cp_expon(struct simplex_t *ret, const struct simplex_t *z);
struct simplex_t *cp_conjug(struct simplex_t *ret, const struct simplex_t *src);
void	cp_print(const struct simplex_t *);


struct simplex_t *cp_inverse(struct simplex_t *ret,
				const struct simplex_t *src);

struct simplex_t *cp_plus(struct simplex_t *ret, const struct simplex_t *a,
				const struct simplex_t *b);
struct simplex_t *cp_pplus(struct simplex_t *ret, const struct simplex_t *a);

struct simplex_t *cp_3plus(struct simplex_t *ret, const struct simplex_t *a,
			const struct simplex_t *b, const struct simplex_t *c);

struct simplex_t *cp_mminus(struct simplex_t *ret, const struct simplex_t *a);

struct simplex_t *cp_minus(struct simplex_t *ret, const struct simplex_t *a,
			const struct simplex_t *b);

struct simplex_t *cp_cross(struct simplex_t *ret, const struct simplex_t *a,
			const struct simplex_t *b);

struct simplex_t *cp_3cross(struct simplex_t *ret, const struct simplex_t *a,
			const struct simplex_t *b, const struct simplex_t *c);

struct simplex_t *cp_crossdme(struct simplex_t *a, const double b);
struct simplex_t *cp_crosszme(struct simplex_t *ret, const struct duplex_t b);

struct simplex_t *cp_crossd(struct simplex_t *ret, const struct simplex_t *a,
			const double b);

struct simplex_t *cp_slash(struct simplex_t *ret, const struct simplex_t *a,
			const struct simplex_t *b);

double		cp_absolute(const struct simplex_t *a);
double		cp_square(const struct simplex_t *a);

struct simplex_t *cp_conjcross(struct simplex_t *ret, const struct simplex_t *a,
			const struct simplex_t *b);

struct duplex_t		*cz_sim2du(struct duplex_t *, const struct simplex_t *);
struct simplex_t	*cz_du2sim(struct simplex_t *, const struct duplex_t *);
int zp_isnull(const struct duplex_t *z);
int cp_isnull(const struct simplex_t *z);
#endif
