#ifndef H_FINITE_H
#define H_FINITE_H

#include "complex.h"
#include "common.h"
struct matrix_t;

/*
 * ELEMENT AND NODE
 *
 * element number should not exceed the 64K.  e_sdomain should not exceed 256
 * I think e_hasbd only used at create_kmatrix();  nodeno can not exceed
 * 64k also. If you are inclined to eradicate this restriction, you should
 * change the typedef's of elemno_t and nodno_t somewhere..
	unsigned char	 n_bdindno;
 */
struct ement_t {
	elmno_t		 e_no;		/* element number */
	unsigned char	 e_nodesno;	/* node_no which surround this ement */
	char		 e_hasbd;	/* this ement has boundary nodes */
	unsigned int	 e_sdomain;	/* subdomain type */
	struct proto_t	*e_proto;	/* element type information */
	nodno_t		*e_nodes;	/* the nodes surrounding this ement*/
};

struct node_t {
	nodno_t		 n_no;		/* node number */
	float		 n_coor[3];	/* Set x, y, z  coor. */
	int		 n_bdflag;	/* boundary indicator */
	unsigned char	 n_plane[3];	/* indicator of which plane */
	nodno_t		*n_ements;	/* ajacent element */
	char		 n_ementsno;
};

struct bound_t {
	int      b_index;
	char    *b_name;
	int      b_type;
};

/*
 * TOP ROOT STRUCTURE
 * most of the following structure will be initialized at mesh_init().
 * Some of them are not used.  You should understand this fact by the fact
 * I am a good example of extreme stupidity.
 * Two of them were deleted.
	int	m_sametype;
	int	m_onesub;
 */
struct mesh {
	int	m_datatype;	/* one of SINGLE, DOUBLE, etc..see mesh_init()*/
	int	m_spacedim;	/* space dimension 2 or 3*/
	int	m_ementsno;	/* elment number in this mesh */
	int	m_nodesno;	/* number of node in this mesh */

	struct pairs	**m_pairs;  	/* used at assemble time */
	struct node_t	**m_nodes;	/* node info of this mesh */
	struct ement_t	**m_ements;	/* element info of this mesh */
	struct sparse_t	 *m_kmatrix;	/* is used at assemble time */
	struct vector_t	 *m_bvector;	/* THIS IS NOT MATRIX */
};

extern struct mesh *meshes;
void	mesh_init(int complexornot);

/*
 * MESH INPUT AND OUTPUT ROUTINES AND MESH GENERATION
 *
 * This package can handle diffpack grid file
 * format and triangle grid-and-element file format.  If you are inclined not
 * to use these grid formats, then you can call genmesh() with appropriate
 * arguments.
 */
int		  mesh_getbdint(char *line);
struct node_t	**mesh_nodesbuf(const int total);
struct ement_t	**mesh_ementsbuf(const int total);

int	dp_getmesh(const char *);
void	dp_putmesh(const char *filename);
int	tri_getmesh(const char *);
void	tri_putmesh(const char *nodefile, const char *ementfile);

void	putmtvform(const char *filename);
void	put_2dgrids(const char *filename, const int needno);
void	put_3dgrids(const char *filename);
int	find_nearestnode(const double x, const double y);
int	find_nearestement(const double x, const double y);
double	get_cosine(const int first, const int second);

struct for3dimcase {
	ushort	 f3d_howmany[3];/* the no of nodes of each side */
	ushort	 f3d_nonodes[3];/* the no of nodes of each plane */
	ushort	 f3d_nincres[3];/* node number incrementation */
	ushort	 f3d_eincres[3];/* element number incrementation */
	ushort	*f3d_minnodes[3];/* to facilitate the cal */
	float	*f3d_coors[3];	/* coor's of each coor; will be malloc'ed */
};
extern struct for3dimcase *m_for3dim;
extern int	m_for3dimed;
extern float	m_coormin[3], m_coormax[3];
	
struct genmesh_t {
	int	gn_spacedim;		/* spacedimension */
	float	gn_mins[3], gn_maxs[3];	/* min and max'es of mesh */
	int	gn_steps[3];		/* how many slices? */
	float	gn_grades[3];		/* grading... */
};

void	genm_linear(const char *geo, const char *part);
void	genm_quadrat(const int, const char *geo, const char *part);
int	get_elmno4point(float x, float y, float z);


/*
 * CREATE_KMATRIX AND CREATE_BMATRIX
 *
 * should be called after you have been
 * inputt'ted. the integration() function is the one you should provide
 * at a specific problem.
 */
struct sparse_t *create_kmatrix(void);
struct vector_t *create_bvector(void);

void	create_smalls(const struct proto_t *,
		struct matrix_t **, struct matrix_t **);
void	integration(const int eno, struct matrix_t *, struct matrix_t *);

/*
 * ASSEMBLE and SOLVE
 *
 * routines for the manupulation of the integration points.  getevalpoint() is
 * to get the points of the integration point, and dgetat() is to get the value
 * at the integration point of the vector, and zgetat() is a complex version for
 * the complex vector.
 */
double		getevalpoint(const int elemno, const int, const int);
double		dgetat(const int elemno, const int, const struct vector_t *);
struct duplex_t	zgetat(const int elemno, const int, const struct vector_t *);

void	Kassemble(const int, const struct matrix_t *, struct sparse_t *dst);
void	bassemble(const int, const struct matrix_t *, struct vector_t *dst);
void	solveall(void);
void	problem(void);

/*
 * for BOUNDARY CONDITIONS
 */
int	isbdset(int i);
int	isbdnoset(int, int);

void	set_subdomain(int (*func)(const struct ement_t *));
void	set_boundary(int (*func)(const struct node_t *));

void	ddirichlet(const int indicator, const double value);
void	zdirichlet(const int indicator, const struct duplex_t value);
void	di_dimpose(const int nodeno, const double value);
void	di_zimpose(const int nodeno, const struct duplex_t value);

int    *get_nearsubdos(const int nodeno);
void	mk_nearements(void);
void	put_gridandvector(const struct vector_t *vp, const char *filename);
float	get_centercoor(const int elemno, const int whatcoor);


/*
 * FOLLOWINGS ARE RELATED WITH THE GAUSSIAN QUADRATURE.
 *
 */
int	tri3_init(struct proto_t *);
int	tri6_init(struct proto_t *);
int	sq4_init(struct proto_t *);
int	sq8_init(struct proto_t *);
int	sq9_init(struct proto_t *);
int     cub8_init(struct proto_t *);

extern struct proto_t *protop[9];

struct integral_t {
	struct proto_t   *intg_proto;	/* its proto_t */
	int		  intg_elemno;	/* the emenent no of this info */
	double		 *intg_jacob;	/* jacobs of the intg points */
	nodno_t		 *intg_nodes;	/* the surrounding nodes info */

	struct matrix_t	 *intg_coors;	/* coors of the surr. nodes */
	struct matrix_t  *intg_realintgcoors; /* real integ points */
	struct matrix_t	**intg_deriv;	/* the deriv's at  */
	struct matrix_t	**intg_jacobm;	/* the jacob matrix at the intg points*/

	/*
	 * following will not be changes I think. That is, following info is
	 * common information.
	 */
	double		 *intg_weights;	/* weights */
	struct matrix_t	 *intg_intgcoors;/* coors of the integration at master*/
	struct matrix_t	**intg_halfjacobm;
	struct matrix_t	 *intg_unit;	/* at unit frame */
	int		 *intg_unitflag;
	struct matrix_t	 *intg_trans;
	struct matrix_t	 *intg_tmp;	/* temporay */

	/*
	 * following will not be used ???
	 */
	struct matrix_t	**intg_hjatcorners;
	struct matrix_t	**intg_derivatcorners;
	struct matrix_t	**intg_jacobmatcorners;
};

struct proto_t {
	int      p_index;
	char    *p_name;
	int      p_nodsno;
	int      p_dimen;
	int      p_intgno;
	int	 p_linesegno;
	int	*p_lineseg;
	struct integral_t *p_intg;
	int    (*p_init)(struct proto_t *);
};


extern struct integral_t *intp[10];
extern struct proto_t protos[];
extern struct proto_t   *currp;         /* current struct proto_t */
extern int      currp_index;

void		 integral_init(struct proto_t *);
int     	 proto_null(struct proto_t *);
struct proto_t  *getproto(const char *);
int    		 proto_init(void);
void		 proto_einit(const int elemno);

#endif
