#ifndef H_MATRIX_H
#define H_MATRIX_H

#include "common.h"
#include "complex.h"
struct vector_t;

struct matrix_t {
	int		mt_type;
	int	      **mt_data;
	char	       *mt_name;
	ushort_t	mt_row;
	ushort_t	mt_col;
	ushort_t	mt_flag;
};

struct dense_t {
	int	dm_row;
	int	dm_col;
	double	*dm_data;
};

extern char *mattypenames[];
extern char *ord[];

#define SAME_ROW(a, b)		(a->mt_row == b->mt_row)
#define SAME_COL(a, b)		(a->mt_col == b->mt_col)

#define IS_SQUARE(a)		(a->mt_row == a->mt_col)
#define SAME_MATRIX(a, b)	(a == b)

#define IS_SMATRIX(m)	(m->mt_type == SP_SINGLE)
#define IS_DMATRIX(m)	(m->mt_type == SP_DOUBLE)
#define IS_CMATRIX(m)	(m->mt_type == SP_SIMPLEX)
#define IS_ZMATRIX(m)	(m->mt_type == SP_DUPLEX)

#define IS_MATRIX(m)	(IS_SMATRIX(m) || IS_DMATRIX(m) \
			 || IS_CMATRIX(m) || IS_ZMATRIX(m))

extern	char *noname;
struct matrix_t	*matrix(const ushort_t, const ushort_t, const char *,const int);

void	m_check(const char *funct, const int arg, const struct matrix_t *mat);
int	m_sizecheck(const struct matrix_t *, int, int);
void	m_2sametypecheck(const char *func,
		const struct matrix_t *ap, const struct matrix_t *bp);
void	m_3sametypecheck(const char *func, const struct matrix_t *ap,
		const struct matrix_t *bp, const struct matrix_t *cp);
void	m_2samesizecheck(const char *func,
		const struct matrix_t *ap, const struct matrix_t *bp);
void	m_squarecheck(const char *func, const struct matrix_t *mp);
void	m_typecheck(const char *func, const int arg, const struct matrix_t *mp,
		const int type);


void	m_free(struct matrix_t *);
void	m_mfree(int, ...);
void	m_null(struct matrix_t *);
void	m_mnull(int, ...);
void	m_print(const struct matrix_t *af, const char *patt, const char *term);
void	m_put(const struct matrix_t *a);
void	m_mput(int num,...);
void	m_add(struct matrix_t *, const struct matrix_t *,
			const struct matrix_t *);
void	m_dot(struct matrix_t *, const struct matrix_t *,
			const struct matrix_t *);
void	m_dott(struct matrix_t *, const struct matrix_t *,
			const struct matrix_t *);
void	m_tdot(struct matrix_t *, const struct matrix_t *,
			const struct matrix_t *);

double	m_getdet(const struct matrix_t *);
struct matrix_t *m_puredup(const struct matrix_t *ap);
struct matrix_t *m_dup(const struct matrix_t *ap);
void	m_copy(struct matrix_t *, const struct matrix_t *);
void	m_inverse(struct matrix_t *, const struct matrix_t *);
int	m_issymmet(const struct matrix_t *);
void	m_briefinfo(const struct matrix_t *);

inline float		**smat(const struct matrix_t *);
inline double		**dmat(const struct matrix_t *);
inline struct simplex_t **cmat(const struct matrix_t *);
inline struct duplex_t	**zmat(const struct matrix_t *);

inline float		*sm_row(const struct matrix_t *, int);
inline double		*dm_row(const struct matrix_t *, int);
inline struct simplex_t *cm_row(const struct matrix_t *, int);
inline struct duplex_t  *zm_row(const struct matrix_t *, int);

void	sm_check(const char *, const int, const struct matrix_t *);
void	dm_check(const char *, const int, const struct matrix_t *);
void	cm_check(const char *, const int, const struct matrix_t *);
void	zm_check(const char *, const int, const struct matrix_t *);
int	sm_issymmet(const struct matrix_t *);
int	dm_issymmet(const struct matrix_t *);
int	cm_issymmet(const struct matrix_t *);
int	zm_issymmet(const struct matrix_t *);

float		 sm_inverse(struct matrix_t *, const struct matrix_t *);
double		 dm_inverse(struct matrix_t *, const struct matrix_t *);
struct simplex_t cm_inverse(struct matrix_t *, const struct matrix_t *);
struct duplex_t	 zm_inverse(struct matrix_t *, const struct matrix_t *);

int		 m_nonzeros(const struct matrix_t *af);
void		 m_pattern(const struct matrix_t *af);
double		 m_compare(const struct matrix_t *, const struct matrix_t *bf);
struct matrix_t *m_mkdiagmat(const struct vector_t *a);
void		 m_mkdiaginv(const struct matrix_t *a);
double		 dm_getdet(const struct matrix_t *);

void		 mat_ludecomp(struct matrix_t *am, struct vector_t *indxp,
					float *d, const int ind);
void		 mat_lubacksub(const struct matrix_t *af,
			const struct vector_t *indxv, struct vector_t *bv);
void		 m_svdcmp(struct matrix_t *am, struct vector_t *wv,
			struct matrix_t *vm);

struct vector_t *m_dotvec(struct vector_t *targ, const struct matrix_t *,
						const struct vector_t *);
struct vector_t *m_tdotvec(struct vector_t *targ, const struct matrix_t *,
						const struct vector_t *);
struct vector_t *m_vecdot(struct vector_t *targ, const struct vector_t *,
						const struct matrix_t *);
struct vector_t *m_tvecdot(struct vector_t *targ, const struct vector_t *,
						const struct matrix_t *);
#endif
