#ifndef __SUPER_H_H
#define __SUPER_H_H

/*
 * File name:		zsp_defs.h
 * Purpose:             Sparse matrix types and function prototypes
 * History:
 */
#include "Cnames.h"
#include "supermatrix.h"
#include "dcomplex.h"
#include "scomplex.h"


/* Driver routines */
extern void sgssv(SuperMatrix *, int *, int *, SuperMatrix *, SuperMatrix *, 
	SuperMatrix *, int *);
extern void dgssv(SuperMatrix *, int *, int *, SuperMatrix *, SuperMatrix *, 
	SuperMatrix *, int *);
extern void cgssv(SuperMatrix *, int *, int *, SuperMatrix *, SuperMatrix *, 
	SuperMatrix *, int *);
extern void zgssv(SuperMatrix *, int *, int *, SuperMatrix *, SuperMatrix *, 
	SuperMatrix *, int *);
extern void
zgssvx(char *, char *, char *, SuperMatrix *, factor_param_t *,
       int *, int *, int *, char *, double *, double *,
       SuperMatrix *, SuperMatrix *, void *, int, SuperMatrix *, 
       SuperMatrix *, double *, double *, double *,
       double *, mem_usage_t *, int *);

/* Supernodal LU factor related */
extern void zCreate_CompCol_Matrix(SuperMatrix *, int, int, int,
	doublecomplex *, int *, int *, Stype_t, Dtype_t, Mtype_t);
extern void zCopy_CompCol_Matrix(SuperMatrix *, SuperMatrix *);
extern void zCreate_Dense_Matrix(SuperMatrix *, int, int,
	doublecomplex *, int, Stype_t, Dtype_t, Mtype_t);
extern void zCreate_SuperNode_Matrix(SuperMatrix *, int, int, int,
	doublecomplex *, int *, int *, int *, int *, int *,
			 Stype_t, Dtype_t, Mtype_t);
extern void zCopy_Dense_Matrix(int, int, doublecomplex *, int,
	doublecomplex *, int);

extern void    Destroy_CompCol_Matrix(SuperMatrix *);
extern void    Destroy_SuperNode_Matrix(SuperMatrix *);
extern void    Destroy_CompCol_Permuted(SuperMatrix *);

extern void    zallocateA (int, int, doublecomplex **, int **, int **);
extern void    StatInit    (int, int);
extern void    StatFree    ();
extern void    sp_preorder (char*, SuperMatrix*, int*, int*, SuperMatrix*);
extern int     sp_coletree (int *, int *, int *, int, int, int *);
extern void    zgstrf (char*, SuperMatrix*, double, double, int, int, int*,
			void *, int, int *, int *, 
                        SuperMatrix *, SuperMatrix *, int *);
extern void    relax_snode  (int, int *, int, int *, int *);
extern int     zsnode_dfs (const int, const int, const int *, const int *,
			     const int *, int *, int *, GlobalLU_t *);
extern int     zsnode_bmod (const int, const int, const int, doublecomplex *,
                              doublecomplex *, GlobalLU_t *);
extern void    zpanel_dfs (const int, const int, const int, SuperMatrix *,
			   int *, int *, doublecomplex *, int *, int *, int *,
			   int *, int *, int *, int *, GlobalLU_t *);
extern void    zpanel_bmod (const int, const int, const int, const int,
                           doublecomplex *, doublecomplex *, int *, int *,
			   GlobalLU_t *);
extern int     zcolumn_dfs (const int, const int, int *, int *, int *, int *,
			   int *, int *, int *, int *, int *, GlobalLU_t *);
extern int     zcolumn_bmod (const int, const int, doublecomplex *,
			   doublecomplex *, int *, int *, int, GlobalLU_t *);
extern int     zcopy_to_ucol (int, int, int *, int *, int *,
                              doublecomplex *, GlobalLU_t *);         
extern int     zpivotL (const int, const double, int *, int *, 
                              int *, int *, int *, GlobalLU_t *);
extern void    zpruneL (const int, const int *, const int, const int,
			     const int *, const int *, int *, GlobalLU_t *);
extern void    resetrep_col (const int, const int *, int *);
extern void    countnz (const int, int *, int *, int *, GlobalLU_t *);
extern void    fixupL (const int, const int *, GlobalLU_t *);
extern int     spcoletree (int *, int *, int *, int, int, int *);
extern int     *TreePostorder (int, int *);
extern void    zreadmt (int *, int *, int *, doublecomplex **, int **, int **);
extern void    zGenXtrue (int, int, doublecomplex *, int);
extern void    zFillRHS (char *, int, doublecomplex *, int, SuperMatrix *,
			SuperMatrix *);
extern void    zgstrs (char *, SuperMatrix *, SuperMatrix *, int *, int *,
			SuperMatrix *, int *);


/* Driver related */

extern void    zgsequ (SuperMatrix *, double *, double *, double *,
			     double *, double *, int *);
extern void    zlaqgs (SuperMatrix *, double *, double *, double,
                             double, double, char *);
extern void    zgscon (char *, SuperMatrix *, SuperMatrix *,
			double, double *, int *);
extern void    zgsrfs (char *, SuperMatrix *, SuperMatrix *, 
			SuperMatrix *, int *, int *, char *, double *,
			double *, SuperMatrix *, SuperMatrix *, 
			double *, double *, int *);

extern int     sp_ztrsv (char *, char *, char *, SuperMatrix *,
			SuperMatrix *, doublecomplex *, int *);
extern int     sp_zgemv (char *, doublecomplex, SuperMatrix *, doublecomplex *,
			int, doublecomplex, doublecomplex *, int);

extern int     sp_zgemm (char *, char *, int, int, int, doublecomplex,
			SuperMatrix *, doublecomplex *, int, doublecomplex, 
			doublecomplex *, int);

/* Memory-related */
extern void    *superlu_malloc (int);
extern void    superlu_free (void*);
extern int     zLUMemInit (char *, void *, int, int, int, int, int,
			     SuperMatrix *, SuperMatrix *,
			     GlobalLU_t *, int **, doublecomplex **);
extern void    SetIWork (int, int, int, int *, int **, int **, int **,
                         int **, int **, int **, int **);
extern void    zSetRWork (int, int, doublecomplex *, doublecomplex **, doublecomplex **);
extern void    zLUWorkFree (int *, doublecomplex *, GlobalLU_t *);
extern int     zLUMemXpand (int, int, MemType, int *, GlobalLU_t *);

extern int     *intMalloc (int);
extern int     *intCalloc (int);
extern doublecomplex  *doublecomplexMalloc(int);
extern doublecomplex  *doublecomplexCalloc(int);
extern double  *doubleMalloc(int);
extern double  *doubleCalloc(int);
extern void    *superlu_malloc(int);
extern int     memory_usage();
extern int     zQuerySpace (SuperMatrix *, SuperMatrix *, int,
				mem_usage_t *);

/* Auxiliary routines */
extern double  SuperLU_timer_ ();
extern int     sp_ienv (int);
extern int     lsame_ (char *, char *);
extern int     xerbla_ (char *, int *);
extern void    ifill (int *, int, int);
extern void    zfill (doublecomplex *, int, doublecomplex);
extern void    snode_profile (int, int *);
extern void    super_stats (int, int *);
extern void    PrintSumm (char *, int, int, int);
/* Routines for debugging */
extern void    print_lu_col(char *, int, int, int *, GlobalLU_t *);
extern void    print_panel_seg(int, int, int, int, int *, int *);
extern void    check_repfnz(int, int, int, int *);

#ifdef __cplusplus
  }
#endif
/*
 * HUNSOO
 */

extern void get_perm_c(int ispec, SuperMatrix *A, int *perm_c);
void superlu_abort_and_exit(char* msg);
void *superlu_malloc(int size);
void superlu_free(void *addr);
void Destroy_SuperMatrix_Store(SuperMatrix *A);
void Destroy_CompCol_Matrix(SuperMatrix *A);
void Destroy_SuperNode_Matrix(SuperMatrix *A);
void Destroy_CompCol_Permuted(SuperMatrix *A);
void Destroy_Dense_Matrix(SuperMatrix *A);
void resetrep_col (const int nseg, const int *segrep, int *repfnz);
void countnz(const int n, int *xprune, int *nnzL, int *nnzU, GlobalLU_t *Glu);
void fixupL(const int n, const int *perm_r, GlobalLU_t *Glu);
void print_panel_seg(int n, int w,int jcol, int nseg, int *segrep, int *repfnz);
void StatInit(int panel_size, int relax);
void StatFree();
void ifill(int *a, int alen, int ival);

void sCreate_CompCol_Matrix(SuperMatrix *A, int m, int n, int nnz, 
			float *nzval, int *rowind, int *colptr,
		      	Stype_t stype, Dtype_t dtype, Mtype_t mtype);
void dCreate_CompCol_Matrix(SuperMatrix *A, int m, int n, int nnz, 
			double *nzval, int *rowind, int *colptr,
		      	Stype_t stype, Dtype_t dtype, Mtype_t mtype);
void cCreate_CompCol_Matrix(SuperMatrix *A, int m, int n, int nnz, 
			complex *nzval, int *rowind, int *colptr,
		      	Stype_t stype, Dtype_t dtype, Mtype_t mtype);
void zCreate_CompCol_Matrix(SuperMatrix *A, int m, int n, int nnz, 
			doublecomplex *nzval, int *rowind, int *colptr,
		      	Stype_t stype, Dtype_t dtype, Mtype_t mtype);
void cCreate_CompCol_Matrix(SuperMatrix *A, int m, int n, int nnz,
                        complex *nzval, int *rowind, int *colptr,
                        Stype_t stype, Dtype_t dtype, Mtype_t mtype);
void dCreate_Dense_Matrix(SuperMatrix *X, int m, int n, double *x, int ldx,
                    Stype_t stype, Dtype_t dtype, Mtype_t mtype);
void cCreate_Dense_Matrix(SuperMatrix *X, int m, int n,
		complex *x, int ldx,
		Stype_t stype, Dtype_t dtype, Mtype_t mtype);
void zCreate_Dense_Matrix(SuperMatrix *X, int m, int n,
		doublecomplex *x, int ldx,
		Stype_t stype, Dtype_t dtype, Mtype_t mtype);


#endif /* __SUPERLU_zSP_DEFS */
