#ifndef H_VECTOR_H
#define H_VECTOR_H

#include "common.h"
#include "complex.h"

struct vector_t {
	char 	*vt_data;	/* location of the datas */
	int	 vt_type;	/* the type of the datas */
	int	 vt_flag;	/* an indicator whether this is vector or not */
	int	 vt_size;	/* the length of this vector */
	char	*vt_name;	/* the name of this vector */
	int	 vt_base;	/* the starting point of this vector */
};

#define VECTOR_TYPED		0x12323f2a

#define IS_VECTOR(vp)	(vp->vt_flag == VECTOR_TYPED)

extern char *noname;

struct vector_t *vector(int, const char *, int);
struct vector_t *vec_read(const char *);
void	vec_check(const char *func, const struct vector_t *, const int);

void	vec_print(const struct vector_t *);
void	vec_pairprint(const struct vector_t *ap, const struct vector_t *bp);
void	vec_mprint(int, ...);
void	zvec_info(const struct vector_t *);
void	vec_null(struct vector_t *);
void	vec_mnull(int, ...);
void	vec_free(struct vector_t *);
void	vec_mfree(int, ...);
void	vec_setbase(struct vector_t *, int);
double	vec_unitdiff(const struct vector_t *ap, const struct vector_t *bp);

inline int		*ivec(const struct vector_t *);
inline float		*svec(const struct vector_t *);
inline double		*dvec(const struct vector_t *);
inline struct simplex_t	*cvec(const struct vector_t *);
inline struct duplex_t	*zvec(const struct vector_t *);

inline double		*dvec_at(const struct vector_t *, int);
inline struct duplex_t	*zvec_at(const struct vector_t *, int);

void	dvec_set(struct vector_t *, double);
void	zvec_set(struct vector_t *, struct duplex_t);

void	vec_printatfile(const struct vector_t *, const char *);
void	vec_printatfiles(const struct vector_t *, const char *);
void	vec_gnuputfiles(const struct vector_t *, const char *, const int many);
void	vec_gnuput(const struct vector_t *, const char *, const int many);

/*
 * from vec_garbage.c
 */
struct vector_t *vec_sumofsq(struct vector_t *xp,
			struct vector_t *ap, struct vector_t *bp);
struct vector_t *vec_sqofsum(struct vector_t *xp,
			struct vector_t *ap, struct vector_t *bp);



int vec_isdouplets(const struct vector_t *ap, const struct vector_t *bp);
int vec_istriplets(const struct vector_t *ap, const struct vector_t *bp,
				const struct vector_t *cp);
double		 vec_getnorm1(const struct vector_t *vp);
double		 vec_getnorm2(const struct vector_t *vp);
double		 vec_getmeannorm2(const struct vector_t *vp);
double		 vec_getnormsq(const struct vector_t *vp);

int		 vec_getlargestindex(const struct vector_t *vp);
struct duplex_t complex_norm(struct vector_t *a, struct vector_t *b);

/*
 * vector operations of
 *     a += b
 *     a  = b + c
 *     a -= b
 *     a  = b + c
 */

struct vector_t *vec_pplus(struct vector_t *ap, const struct vector_t *bp);
struct vector_t *vec_plus(struct vector_t *ap, const struct vector_t *bp,
				const struct vector_t *cp);
struct vector_t *vec_mminus(struct vector_t *ap, const struct vector_t *bp);
struct vector_t *vec_minus(struct vector_t *ap, const struct vector_t *bp,
				const struct vector_t *cp);


void		 vec_copy(struct vector_t *avp, const struct vector_t *bvp);
double		 vec_getabsoluteat(const struct vector_t *vp, int where);

struct vector_t *vec_dup(const struct vector_t *vp);
struct vector_t *vec_puredup(const struct vector_t *vp);

struct duplex_t  *vec_zdotprod(struct duplex_t *, const struct vector_t *,
			const struct vector_t *);
struct duplex_t  *vec_wrongzdotprod(struct duplex_t *, const struct vector_t *,
			const struct vector_t *);
struct duplex_t  *vec_zdotprod2nd(struct duplex_t *ret, int ind,
			const struct vector_t *ap, const struct vector_t *bp);

double		 vec_ddotprod(const struct vector_t *, const struct vector_t *);

struct vector_t *vec_crossd(struct vector_t *ret, const struct vector_t *ap,
				const double multicant);
struct vector_t *vec_crossz(struct vector_t *ret, const struct vector_t *ap,
				const struct duplex_t mult);
struct vector_t *vec_dcross(struct vector_t *ret, const double multicant,
				const struct vector_t *ap);
struct vector_t *vec_zcross(struct vector_t *ret, const struct duplex_t mult,
				const struct vector_t *ap);
struct vector_t *vec_zcross2nd(struct vector_t *ret, const int indicator,
			const struct duplex_t multicant,
			const struct vector_t *ap);
struct vector_t *vec_dcrossme(struct vector_t *ap, const double multicant);

struct vector_t *vec_dcombi(struct vector_t *avp, const double first,
				const struct vector_t *bvp, const double second,
				const struct vector_t *cvp);
struct vector_t *vec_inverseplus(struct vector_t *avp,
			const struct vector_t *bvp, const struct vector_t *cvp);
struct vector_t *
vec_zcombi(struct vector_t *avp,
	const struct duplex_t first, const struct vector_t *bvp,
	const struct duplex_t sec, const struct vector_t *cvp);

#endif
