#include <stdio.h>
#include <math.h>
#include "common.h"
#include "complex.h"
#include "vector.h"
#include "matrix.h"
#include "finite.h"

double
dgetat(const int elemno, const int no, const struct vector_t *vp)
{
	struct proto_t *p;
	double  **unit;
	float	*sdata;
	double	*ddata, dtmp;
	int	i;
	nodno_t	*nnp;	/* for e_nodes */

	if (vp->vt_size != meshes->m_nodesno)
		fatal("%s(): wrong vector size of %s", __F, vp->vt_name);
	if (vp->vt_type != SP_SINGLE && vp->vt_type != SP_DOUBLE)
		fatal("%s(): vector %s is not real type", __F, vp->vt_name);

	p = meshes->m_ements[elemno]->e_proto;
	unit = dmat(p->p_intg->intg_trans);

	nnp = meshes->m_ements[elemno]->e_nodes;

	dtmp = 0.0;
	switch (vp->vt_type) {
		case SP_SINGLE:
			sdata = svec(vp);
			for (i = 0; i < p->p_nodsno; i++)
				dtmp += sdata[nnp[i]] * unit[no][i];
			break;
		case SP_DOUBLE:
			ddata = dvec(vp);
			for (i = 0; i < p->p_nodsno; i++)
				dtmp += ddata[nnp[i]] * unit[no][i];
			break;
	}
	return dtmp;
}

struct duplex_t
zgetat(const int elemno, const int no, const struct vector_t *vp)
{
	struct proto_t *p;
	double  **unit;
	struct simplex_t *cdata, ctmp, ctmp1;
	struct duplex_t *zdata, ztmp, ztmp1;
	int	i;
	nodno_t	*nnp;	/* for e_nodes */

	if (vp->vt_size != meshes->m_nodesno)
		fatal("%s(): wrong vector size of %s", __F, vp->vt_name);
	if (vp->vt_type != SP_SIMPLEX && vp->vt_type != SP_DUPLEX)
		fatal("%s(): vector %s is not complex type",__F, vp->vt_name);

	p = meshes->m_ements[elemno]->e_proto;
	unit = dmat(p->p_intg->intg_trans);

	nnp = meshes->m_ements[elemno]->e_nodes;

	ztmp = duplex_null;
	ctmp = simplex_null;
	switch (vp->vt_type) {
		case SP_SIMPLEX:
			cdata = cvec(vp);
			for (i = 0; i < p->p_nodsno; i++) {
				cp_pplus(&ctmp, cp_crossd(&ctmp1,
					&cdata[nnp[i]], unit[no][i]));
			}
			ztmp.real = ctmp.real;
			ztmp.imag = ctmp.imag;
			break;
		case SP_DUPLEX:
			zdata = zvec(vp);
			for (i = 0; i < p->p_nodsno; i++)
				zp_pplus(&ztmp, zp_crossd(&ztmp1,
					&zdata[nnp[i]], unit[no][i]));
			break;
		default:
			fatal("%s(): wrong type", __F);
	}
	return ztmp;
}

/*
 * DO NOT call this function!  because after call p->p_einit();
 * p->p_intg->intg_realintgcoors will contains the information that you
 * want to find out
 */

double
getevalpoint(const int elemno, const int no, const int whichcoor)
{
	struct proto_t *p;
	double	**unit, temp;
	int	i;
	nodno_t	*nnp;
	struct ement_t *ep;

	p = meshes->m_ements[elemno]->e_proto;
	if (elemno < 0 || elemno >= meshes->m_ementsno)
		fatal("%s():Wow! you proved you are a stupid one!!", __F);
	if (whichcoor < 0 || whichcoor >= p->p_dimen)
		fatal("%s():err whichcoor=%d!!", __F, whichcoor);
	if (p->p_intgno <= no || no < 0)
		fatal("%s():err 2nd argument=%d!!", __F, no);

	unit = dmat(p->p_intg->intg_trans);

	ep = meshes->m_ements[elemno];	/* find corrres. element struct */
	nnp = ep->e_nodes;	/* find his neighbors */

	temp = 0.0;
	for (i = 0; i< p->p_nodsno; i++)
		temp += meshes->m_nodes[nnp[i]]->n_coor[whichcoor]* unit[no][i];

	return temp;
}

/*
 * WRONG!!!
 * get the derivative with respect to whichcoor at the point of itgno of a
 * vector_t *vp.  vp should be SP_SINGLE or SP_DOUBLE type as you expect.
 */
double
dgetderivat(const int elemno, const int itgno, const int whichcoor,
	const struct vector_t *vp)
{
	struct proto_t *p;
	double  **halfform;
	float	*sdata;
	double	*ddata, dtmp;
	int	i;
	nodno_t	*nnp;	/* for e_nodes */

	if (vp->vt_size != meshes->m_nodesno)
		fatal("%s(): wrong vector size of %s", __F, vp->vt_name);
	if (vp->vt_type != SP_SINGLE && vp->vt_type != SP_DOUBLE)
		fatal("%s(): vector %s is not real type", __F, vp->vt_name);

	p = meshes->m_ements[elemno]->e_proto;

	if (itgno < 0 || itgno >= p->p_intgno)
		fatal("%s(): 2nd argument %d is out of range", __F, itgno);
	if (whichcoor < 0 || whichcoor >= p->p_dimen)
		fatal("%s(): 3rd argument %d is out of range", __F, whichcoor);

	halfform = dmat(p->p_intg->intg_halfjacobm[itgno]);

	nnp = meshes->m_ements[elemno]->e_nodes;

	dtmp = 0.0;
	switch (vp->vt_type) {
		case SP_SINGLE:
			sdata = svec(vp);
			for (i = 0; i < p->p_nodsno; i++)
				dtmp += sdata[nnp[i]] * halfform[whichcoor][i];
			break;
		case SP_DOUBLE:
			ddata = dvec(vp);
			for (i = 0; i < p->p_nodsno; i++)
				dtmp += ddata[nnp[i]] * halfform[whichcoor][i];
			break;
	}
	return dtmp;
}

/*
 * WRONG!!!
 * get the derivative with respect to whichcoor at the point of itgno of a
 * vector_t *vp.  vp should be SP_SIMPLEX or SP_DUPLEX type as you expect.
 */
struct duplex_t
zgetderivat(const int elemno, const int itgno, const int whichcoor,
	const struct vector_t *vp)
{
	struct proto_t *p;
	double  **halfform;
	struct simplex_t *cdata, ctmp, cret;
	struct duplex_t *zdata, ztmp, zret;
	int	i;
	nodno_t	*nnp;	/* for e_nodes */

	if (vp->vt_size != meshes->m_nodesno)
		fatal("%s(): wrong vector size of %s", __F, vp->vt_name);
	if (vp->vt_type != SP_SIMPLEX && vp->vt_type != SP_DUPLEX)
		fatal("%s(): vector %s is not complex type", __F, vp->vt_name);

	p = meshes->m_ements[elemno]->e_proto;

	if (itgno < 0 || itgno >= p->p_intgno)
		fatal("%s(): 2nd argument %d is out of range", __F, itgno);
	if (whichcoor < 0 || whichcoor >= p->p_dimen)
		fatal("%s(): 3rd argument %d is out of range", __F, whichcoor);

	halfform = dmat(p->p_intg->intg_halfjacobm[itgno]);

	nnp = meshes->m_ements[elemno]->e_nodes;

	zret = duplex_null;
	cret = simplex_null;
	switch (vp->vt_type) {
		case SP_SIMPLEX:
			cdata = cvec(vp);
			for (i = 0; i < p->p_nodsno; i++)
				cp_pplus(&cret, cp_crossd(&ctmp,
					&cdata[nnp[i]],halfform[whichcoor][i]));

			zret.real = cret.real;
			zret.imag = cret.imag;
			break;
		case SP_DUPLEX:
			zdata = zvec(vp);
			for (i = 0; i < p->p_nodsno; i++)
				zp_pplus(&zret, zp_crossd(&ztmp,
					&zdata[nnp[i]],halfform[whichcoor][i]));
			break;
	}
	return zret;
}
/*
 * I assume that all the subdomain numbers are smaller than 32
 * returns the all the subdomain numbers which is stored at the bit form

 *
 * DONT USE THIS FUNCTION
 */
int	sd_reservoir[33];

int *
get_nearsubdos(const int nodeno)
{
	nodno_t *nnp;
	struct ement_t *ep;
	int	i, j, inx;
	int	*ret;

	if (meshes->m_nodesno <= nodeno || nodeno < 0)
		fatal("%s(): OOB", __F);
	
	ret = sd_reservoir;
	inx = 0;
	for (i=0; i<meshes->m_ementsno; i++) {
		ep = meshes->m_ements[i];
		nnp = ep->e_nodes;
		for (j=0; j < ep->e_nodesno; j++) {
			if (nnp[j] == nodeno) {
				ret[inx++] = ep->e_sdomain;
				continue;
			}
		}
	}
	if (inx == 0)
		fatal("%s(nodeno=%d): how can a node can not ...", __F, nodeno);
	ret[inx] = -1;
	return ret;
}

int
find_nearestnode(const double x, const double y)
{
	double minlen, len, delx, dely;
	struct node_t *np;
	int	i, ret;
	
	minlen = 1.0e10;
	for (i=0; i < meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		delx = np->n_coor[0] - x;
		dely = np->n_coor[1] - y;
		len = sqrt(delx*delx + dely*dely);
		if (minlen > len) {
			ret = i;
			minlen = len;
		}
	}
	return ret;
}

int
find_nearestement(const double x, const double y)
{
	float	min, delx, dely, len;
	int	i, ret;

	min = 1.0e10;
	for (i=0; i< meshes->m_ementsno; i++) {
		delx = x - get_centercoor(i, 0);
		dely = y - get_centercoor(i, 1);
		len = sqrt(delx*delx + dely*dely);
		if (min > len) {
			min = len;
			ret = i;
		}
	}
	return ret;
}

double
get_cosine(const int first, const int second)
{
	struct node_t *np0, *np1;
	double	x1, y1, x0, y0, delx, dely;
	double	len, off;
	if (first == second)
		return 1;

	np0 = meshes->m_nodes[first];
	np1 = meshes->m_nodes[second];
	x0 = np0->n_coor[0]; y0 = np0->n_coor[1];
	x1 = np1->n_coor[0]; y1 = np1->n_coor[1];

	delx = x0-x1;		dely = y0-y1;
	delx = fabs(delx);	dely = fabs(dely);

	off = delx < dely ? dely : delx;
	len = sqrt(delx*delx+dely*dely);

	return (off/len);
}
