#include <stdio.h>

#include "common.h"
#include "complex.h"
#include "vector.h"
#include "matrix.h"
#include "sparse.h"
#include "finite.h"

void	make_pairs(const int elmentno);
void	pair_print(const int size);
void	put_pair(struct pairs *p, int no);

static struct pairs pairtmp[50];	/* XXX */

/*
 * Assemble at meshes->m_kmatrix
 */
void
Kassemble(const int elemno, const struct matrix_t *retm, struct sparse_t *dst)
{
	int	size;		/* matrix size of trans */
	int	i, j;
	double	**dm, *dp;
	float	*fp;
	struct duplex_t **zm, *zp;
	struct simplex_t *cp, ctmp;
	struct pairs *pairp; /* XXX */
	struct sparse_t *sp;

	size = meshes->m_ements[elemno]->e_nodesno;
	sp   = dst;
	make_pairs(elemno); /* XXX */

	switch (meshes->m_datatype) {
	case SP_SINGLE:
		dm = dmat(retm); fp = ssp(dst);
		for (i=0;i<size;i++) {
			pairp = meshes->m_pairs[i];
			for (j=0;j<size;j++) {
				fp[pairp[j].pr_locus] += dm[i][j];
			}
		}
		break;
	case SP_DOUBLE:
		dm = dmat(retm); dp = dsp(dst);
		for (i=0;i<size;i++) {
			pairp = meshes->m_pairs[i];
			for (j=0;j<size;j++) {
				dp[pairp[j].pr_locus] += dm[i][j];
			}
		}
		break;
	case SP_SIMPLEX:
		zm = zmat(retm); cp = csp(dst);
		for (i=0;i<size;i++) {
			pairp = meshes->m_pairs[i];
			for (j=0;j<size;j++) {
				ctmp = simplexv(zm[i][j].real, zm[i][j].imag);
				cp_pplus(&cp[pairp[j].pr_locus], &ctmp);
			}
		}
		break;
	case SP_DUPLEX:
		zm = zmat(retm); zp = zsp(dst);
		for (i=0;i<size;i++) {
			pairp = meshes->m_pairs[i];
			for (j=0;j<size;j++) {
				zp_pplus(&zp[pairp[j].pr_locus], &zm[i][j]);
			}
		}
		break;
	default:
		fatal("%s():Are you sure to call mesh_init() properly?", __F);
	}
	return;
}

/*
 * Assemble at meshes->m_bmatrix
 * Remember meshes->m_bmatrix is not really matrix
 */
void
bassemble(const int elemno, const struct matrix_t *retm, struct vector_t *dst)
{
	int	size;		/* matrix size of trans */
	nodno_t	*nnp;
	int	i;
	double	**dm, *dp;
	float	*fp;
	struct simplex_t *cp, ctmp;
	struct duplex_t **zm, *zp;

	size = meshes->m_ements[elemno]->e_nodesno;
	nnp = meshes->m_ements[elemno]->e_nodes;

	switch (meshes->m_datatype) {
		case SP_SINGLE:
			dm = dmat(retm); fp = svec(dst);
			for (i=0;i<size;i++)
				fp[nnp[i]] +=  dm[i][0];

			break;
		case SP_DOUBLE:
			dm = dmat(retm); dp = dvec(dst);
			for (i=0;i<size;i++)
				dp[nnp[i]] +=  dm[i][0];

			break;
		case SP_SIMPLEX:
			zm = zmat(retm); cp = cvec(dst);
			for (i=0;i<size;i++) {
				ctmp = simplexv(zm[i][0].real, zm[i][0].imag);
				cp_pplus(&cp[nnp[i]], &ctmp);
			}
			break;

		case SP_DUPLEX:
			zm = zmat(retm); zp = zvec(dst);
			for (i=0;i<size;i++) {
				zp_pplus(&zp[nnp[i]], &zm[i][0]);
			}
			break;
		default:
			fatal("%s():Are you sure to call mesh_init()properly?",
				__F);
	}
	return;
}

/*
 * indicator starts at 0 !!!!!!
 * XXXXXXX
 */
int
isbdnoset(int nodeno, int indicator)
{
	struct node_t *np0;

	if (indicator < 0 || indicator > 31)
		fatal("%s(): Wrong boundary indicator!!", __F);

	np0 = meshes->m_nodes[nodeno];
	if (isset(np0->n_bdflag, (indicator-1)))
		return 1;
	return 0;
}
		
int
isbdset(int no)
{
	struct node_t *np0;
	np0 = meshes->m_nodes[no];
	if (np0->n_bdflag)
		return np0->n_bdflag;
	return 0;
}

/*
 * binary search for pair.
 * basic algorithm was taken from BSD source of UNIX.
 */

struct pairs *
bsearch4pair(struct pairs *origin, const int target, const int tot)
{
	struct pairs *base = origin;
	int lim, cmp, ret;
	struct pairs *p;

	for (lim = tot; lim != 0; lim >>= 1) {
		p = base + (lim >> 1);
		ret = p->pr_taxis;
		cmp = ret < target ? 1 : ret > target ? -1 : 0;
		if (cmp == 0)
			return p;
		if (cmp > 0) {  /* key > p: move right */
			base = (p + 1);
			lim--;
		} /* else move left */
	}
	return (NULL);
}

/*
 * Main purpose of creating this pairs??? is to decrease the computation time,
 * but I do not know whether or not this struct pair decreased the
 * computational time.
 */
void
make_pairs(const int elemno)
{
	int	size;		/* matrix size of trans */
	nodno_t	*nnp;
	int	allcol;
	int	i, j;
	struct pairs **duette, *rowpair, *pp;
	struct sparse_t *sp;

	sp = meshes->m_kmatrix;
	size = meshes->m_ements[elemno]->e_nodesno;
	nnp = meshes->m_ements[elemno]->e_nodes;

	duette = meshes->m_pairs;

	for (i=0; i<size; i++) {
		allcol = sp_getpairs(pairtmp, SP_FINDROWS, 50, sp, nnp[i]);
		rowpair = duette[i];
		for (j=0; j < size; j++) {
			pp = bsearch4pair(pairtmp, nnp[j], allcol);
			rowpair[j] = *pp;
		}
	}

	return;
}
/*
 * following two codes are for debugging
 */
void
pair_print(const int size)
{
	struct pairs **duette;
	int	i, j;

	duette = meshes->m_pairs;
	for (i=0; i < size; i++) {
		for (j=0; j < size; j++) {
			printf("%3d %3d+ ",
				duette[i][j].pr_taxis,
				duette[i][j].pr_locus);
		}
		printf("\n");
	}
	printf("+++++++++\n");
	return;
}

void
put_pair(struct pairs *p, int no)
{
	int	i;
	for (i=0; i<no;i++, p++)
		printf("%3d ", p->pr_taxis);
	printf("\n");
	return;
}
