#include <stdio.h>
#include <math.h>

#include "common.h"
#include "matrix.h"
#include "finite.h"

struct integral_t *cub8_intg;

static void	cub8_setunit(void);
static void	cub8_setintp(void);
static void	cub8_halfjacobmat(void);
static void	cub8_transmat(void);

int
cub8_init(struct proto_t *p)
{
	static int cub8_inited = 0;

	if (cub8_inited)
		return 0;
	cub8_inited++;

	cub8_intg = p->p_intg;

	cub8_setunit();
	cub8_setintp();
	cub8_halfjacobmat();
	cub8_transmat();

	return 0;
}


static void
cub8_setunit(void)
{
        double  **points;

        points = dmat(cub8_intg->intg_unit);

        points[0][0] = -1.0; points[0][1] = -1.0; points[0][2] = -1.0;
        points[1][0] =  1.0; points[1][1] = -1.0; points[1][2] = -1.0;
        points[2][0] = -1.0; points[2][1] =  1.0; points[2][2] = -1.0;
        points[3][0] =  1.0; points[3][1] =  1.0; points[3][2] = -1.0;
	/*
	 * Upper
	 */
        points[4][0] = -1.0; points[4][1] = -1.0; points[4][2] =  1.0;
        points[5][0] =  1.0; points[5][1] = -1.0; points[5][2] =  1.0;
        points[6][0] = -1.0; points[6][1] =  1.0; points[6][2] =  1.0;
        points[7][0] =  1.0; points[7][1] =  1.0; points[7][2] =  1.0;
        return;
}

static void
cub8_setintp(void)
{
	double	scale;
	double	*points, *corner;
	double	*weights;
	int	i;

	scale   = 1.0/sqrt(3.0);
	weights = cub8_intg->intg_weights;
	for (i=0;i < 8;i++) {
		points = dm_row(cub8_intg->intg_intgcoors, i);
		corner = dm_row(cub8_intg->intg_unit, i);

		points[0] = corner[0] * scale;
		points[1] = corner[1] * scale;
		points[2] = corner[2] * scale;
		weights[i] = 1.0;
	}
	return;
}

/*
 * Make half form of Jacobian matrices
 */
static void
cub8_halfjacobmat(void)
{
	double	**mat;
	int	i, intgno;
	double	*units, *points;

	for (intgno=0; intgno < 8; intgno++) {
		mat = dmat(cub8_intg->intg_halfjacobm[intgno]);
		points = dm_row(cub8_intg->intg_intgcoors, intgno);

		for (i=0; i < 8; i++) {
			units = dm_row(cub8_intg->intg_unit, i);
			mat[0][i] = 0.125
				* units[0]
				* (1.0 + units[1] * points[1])
				* (1.0 + units[2] * points[2]);
		}

		for (i=0; i < 8; i++) {
			units = dm_row(cub8_intg->intg_unit, i);
			mat[1][i] = 0.125
				* (1.0 + units[0] * points[0])
				* units[1]
				* (1.0 + units[2] * points[2]);
		}

		for (i=0; i < 8; i++) {
			units = dm_row(cub8_intg->intg_unit, i);
			mat[2][i] = 0.125
				* (1.0 + units[0] * points[0])
				* (1.0 + units[1] * points[1])
				* units[2];
		}
	}
	return;
}



static void
cub8_transmat(void)
{
	double	**result, *points, *coeff;
	int	i, j;

	result = dmat(cub8_intg->intg_trans);

	for (i=0; i < 8; i++) {	/* for four integration points */
		points = dm_row(cub8_intg->intg_intgcoors, i);

		for (j=0; j < 8;j++) {
			coeff = dm_row(cub8_intg->intg_unit, j);
			result[i][j] = 0.125
				* (1.0 + points[0] * coeff[0])
				* (1.0 + points[1] * coeff[1])
				* (1.0 + points[2] * coeff[2]);
		}
	}
	return;
}
