#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "common.h"
#include "matrix.h"
#include "finite.h"

struct proto_t *protop[9];

struct proto_t protos[9] = {
	{ 0, "ElmB4n2D",   4, 2, 4, 4, NULL, NULL, sq4_init},
	{ 1, "ElmB8n2D",   8, 2, 9, 4, NULL, NULL, sq8_init},
	{ 2, "ElmB9n2D",   9, 2, 9, 4, NULL, NULL, sq9_init},
	{ 3, "ElmT3n2D",   3, 2, 1, 3, NULL, NULL, tri3_init},
	{ 4, "ElmT6n2D",   6, 2, 3, 3, NULL, NULL, tri6_init},

	{ 5, "ElmB8n3D",   8, 3, 8,12, NULL, NULL, cub8_init},
	{ 6, "ElmB20n3D", 20, 3, 4,12, NULL, NULL, NULL},
	{ 7, "ElmB27n3D", 27, 3, 4,12, NULL, NULL, NULL},
	{-1, NULL, 0, 0}
}; 

struct mesh *meshes;
struct proto_t	*currp;
int	currp_index = -1;

static double	proto_ugetall(const int no);
static void	trans_check(const struct proto_t *p);
static void	deriv_check(const struct proto_t *p);

/*
 * you dont have to pay attention at this fucntion 'cause the mesh_init() will
 * call this function automatically, provided that you have called that
 * mesh_init().
 */

int
proto_init(void)
{
	struct proto_t *pt;
	int	maxnodno;	/* maximum size of the small K-matrix */
	static int proto_inited = 0;

	if (proto_inited)
		fatal("%s(): What's wrong!! %s() should not be called twice",
			__F, __F);

	proto_inited++;
		
	maxnodno = 0;
	for (pt = &protos[0]; pt->p_name; pt++) { /* for each element type */
		if (pt->p_init) {
			printf("%s(): %s starts\n", __F, pt->p_name);
			pt->p_lineseg = (int *)
				malloc(sizeof(int)*(pt->p_linesegno+1));
			if (pt->p_lineseg == NULL)
				memerr("%s(): [0]", __F);

			integral_init(pt);	/* for Gaussian quadrature */
			pt->p_init(pt);		/* for basis functions, etc. */
			trans_check(pt);
			deriv_check(pt);
			if (maxnodno < pt->p_nodsno)
				maxnodno = pt->p_nodsno;

		} else
			pt->p_init = proto_null;
	}
	return maxnodno; /* this will be used at meshes->m_pairs */
}

struct proto_t	*
getproto(const char *name)
{
	struct proto_t	*pt;

	pt = &protos[0];
	for (; pt->p_name && (strcmp(name, pt->p_name)); pt++)
		/* void */;

	if (pt->p_name == NULL)
		fatal("%s(): wrong proto!!  you gave me %s", __F, name);
	return pt;
}

int
proto_null(struct proto_t *p)
{
	char *name;

	name = p->p_name;
	printf("\n\nproto routines of %s ARE ", name);
	printf("NOT IMPLEMENTED YET!!\n");
	printf("If you have enough knowledge of FEM, I recommand you\n");
	printf("to implement this type of element\n\n");
	exit(1);
}
		

/*
 * For the Gaussian quadrature, etc.
 * p->p_nodsno means the number of the basis functions
 */

void
integral_init(struct proto_t *p)
{
	struct integral_t *intg;
	int	i;

	if (p->p_intg)
		fatal("%s", __F);

	intg = (struct integral_t *) malloc(sizeof(struct integral_t));
	if (intg == NULL)
		memerr("%s", __F);

	intg->intg_jacob = (double *) malloc(sizeof(double) * p->p_intgno);
	if (intg->intg_jacob == NULL)
		memerr("%s", __F);
	intg->intg_weights = (double *) malloc(sizeof(double) * p->p_intgno);
	if (intg->intg_weights == NULL)
		memerr("%s", __F);

	intg->intg_coors = matrix(p->p_nodsno, p->p_dimen,"coors", SP_DOUBLE);
	intg->intg_intgcoors=matrix(p->p_intgno,p->p_dimen,"intgxy",SP_DOUBLE);
	intg->intg_trans = matrix(p->p_intgno, p->p_nodsno, "trans",SP_DOUBLE);
	intg->intg_unit  = matrix(p->p_nodsno, p->p_dimen, "unit", SP_DOUBLE);
	intg->intg_unitflag = (int *) malloc(sizeof(int) * p->p_nodsno);
	intg->intg_realintgcoors=
		matrix(p->p_intgno, p->p_dimen,  "real intg", SP_DOUBLE);

	if (intg->intg_unitflag == NULL)
		memerr("%s", __F);

	intg->intg_nodes = (nodno_t *) malloc(sizeof(nodno_t) * p->p_nodsno);
	if (intg->intg_nodes == NULL)
		memerr("%s", __F);

	intg->intg_deriv = (struct matrix_t **)
		malloc(sizeof(struct matrix_t *) * p->p_intgno);
	intg->intg_jacobm = (struct matrix_t **)
		malloc(sizeof(struct matrix_t *) * p->p_intgno);
	intg->intg_halfjacobm = (struct matrix_t **)
		malloc(sizeof(struct matrix_t *) * p->p_intgno);

	intg->intg_tmp = matrix(p->p_dimen, p->p_dimen, "tmp", SP_DOUBLE);

	for (i=0; i < p->p_intgno; i++) { /* for integration points */
		intg->intg_deriv[i]
			= matrix(p->p_dimen, p->p_nodsno,"deriv", SP_DOUBLE);
		intg->intg_jacobm[i]
			= matrix(p->p_dimen, p->p_dimen, "jacobm", SP_DOUBLE);
		intg->intg_halfjacobm[i]
			= matrix(p->p_dimen, p->p_nodsno,"halfj", SP_DOUBLE);
	}
	intg->intg_elemno = -1;	/* XXX MOST OMMISIBLE line */

	/*
	 * for the corner nodes
	 */
#if 0
	intg_hjatcorners = (struct matrix_t **)
		malloc(sizeof(struct matrix_t *) * p->p_nodsno);
	intg_derivatcorners = (struct matrix_t **)
		malloc(sizeof(struct matrix_t *) * p->p_nodsno);
	intg_jacobmatcorners = (struct matrix_t **)
		malloc(sizeof(struct matrix_t *) * p->p_nodsno);

	for (i=0; i < p->p_intgno; i++) { /* for integration points */
		intg->intg_derivatcorners[i]
			= matrix(p->p_dimen, p->p_nodsno,"deriv", SP_DOUBLE);
		intg->intg_jacobmatcorners[i]
			= matrix(p->p_dimen, p->p_dimen, "jacobm", SP_DOUBLE);
		intg->intg_hjatcorners[i]
			= matrix(p->p_dimen, p->p_nodsno,"halfj", SP_DOUBLE);
	}
#endif

	/*
	 * All malloc is over
	 */
	p->p_intg = intg;
	intg->intg_proto = p;

	return;
}

static struct integral_t *intg;

/*
 * following function will be called at the time of assemble time.
 */

void
proto_einit(const int elemno)
{
	int	i, j;
	nodno_t	*nnp;
	double	**dpp;
	struct ement_t	*ep;
	struct node_t	*np;
	struct proto_t *p;

	ep   = meshes->m_ements[elemno]; 
	p    = ep->e_proto;
	if (p != currp)
		fatal("%s(): p != currp", __F);

	intg = p->p_intg;
	intg->intg_elemno = elemno;

	nnp = ep->e_nodes;	/* seek the boundary nodes */

	dpp = dmat(intg->intg_coors);

	for (i=0; i < p->p_nodsno; i++) {
		np = meshes->m_nodes[nnp[i]];
		for (j=0; j < p->p_dimen; j++)
			dpp[i][j] = np->n_coor[j];
	}
	m_dot(intg->intg_realintgcoors, intg->intg_trans,
		intg->intg_coors);

	for (i=0; i < p->p_intgno; i++) {
		intg->intg_jacob[i] = proto_ugetall(i);
	}

	return;
}


static double
proto_ugetall(const int no)
{
	double  det;            /* Determinant of jacob */

	m_dot(intg->intg_jacobm[no],
		intg->intg_halfjacobm[no], intg->intg_coors);

	det = dm_inverse(intg->intg_tmp, intg->intg_jacobm[no]);

	/* tem is now inverse of jacob at this no'th integration point */
	/* Let's calculate (9.42) */

	m_dot(intg->intg_deriv[no],
		intg->intg_tmp, intg->intg_halfjacobm[no]);

	return det;
}

/*
 * The summation of the basis functions at an integration points of the master
 * element should be UNIT exactly, I think.
 */
static void
trans_check(const struct proto_t *p)
{
	int	i, j;
	double	sum, **trans;

	trans = dmat(p->p_intg->intg_trans);
	for (i=0; i < p->p_intgno; i++) {
		sum = 0.0;
		for (j=0; j < p->p_nodsno; j++)
			sum += trans[i][j];
		if (fabs(sum-1.0) > 1.0e-10)
			fatal("%s(%s): sum should be 1.0", __F, p->p_name);
	}
	if (0)
		printf("%s(%s)  ", __F, p->p_name);
	return;
}

/*
 * The summation of derivatives with respect to x, y, or z of the basis
 * functions at an integration points of the master element should be ZERO
 * exactly, I think.
 */
static void
deriv_check(const struct proto_t *p)
{
	int	i, j, k;
	double	**dpp, **unit, **deriv, *jacob;
	double	sum;

	intg = p->p_intg;

	dpp  = dmat(intg->intg_coors);
	unit = dmat(intg->intg_unit);
	jacob= intg->intg_jacob;

	for (i=0; i < p->p_nodsno; i++) {
		for (j=0; j < p->p_dimen; j++)
			dpp[i][j] = unit[i][j];
	}

	for (i=0; i < p->p_intgno; i++)
		intg->intg_jacob[i] = proto_ugetall(i);

	for (i=0; i < p->p_intgno; i++) {
		deriv = dmat(intg->intg_deriv[i]);
		sum = 0.0;
		for (j=0; j < p->p_dimen; j++) {
			for (k=0; k < p->p_nodsno; k++)
				sum += deriv[j][k];
			if (fabs(sum) > 1.0e-10)
				fatal("%s(%s): sum should be 0.0",
					__F, p->p_name);
		}
		if (fabs(jacob[i] - 1.0) > 1.0e-10)
			fatal("%s(%s): jacobian should be 1.0",
				__F, p->p_name);
	}
	if (0)
		printf("%s(%s)  ", __F, p->p_name);

	return;
}
