#include <stdio.h>
#include <math.h>

#include "common.h"
#include "matrix.h"
#include "finite.h"

static void     sq4_setunit(void);
static void	sq4_integpoints(void);
static void	sq4_mkhalfjacobm(void);
static void	sq4_mktrans(void);

struct integral_t *sq4_intg;

int
sq4_init(struct proto_t *p)
{
	static int sq4_inited = 0;
	int	*lineseg;

	if (sq4_inited)
		return 0;

	sq4_inited++;
	sq4_intg = p->p_intg;

	/*
	 * ORDER is important
	 */
	sq4_setunit();
	sq4_integpoints();
	sq4_mkhalfjacobm();
	sq4_mktrans();

	lineseg = p->p_lineseg;
	lineseg[0] = 0; lineseg[1] = 1;
	lineseg[2] = 3; lineseg[3] = 2;
	lineseg[4] = 0;

	return 0;
}

static void
sq4_setunit(void)
{
	struct matrix_t *matf;
	double	**mat;

	matf = sq4_intg->intg_unit;
	mat  = dmat(matf);

	mat[0][0] = -1.0; mat[0][1] = -1.0;
	mat[1][0] =  1.0; mat[1][1] = -1.0;
	mat[2][0] = -1.0; mat[2][1] =  1.0;
	mat[3][0] =  1.0; mat[3][1] =  1.0;

	return;
}


void
sq4_integpoints(void)
{
	int i;
	double	scale;
	double	**unit, **intp;
	double	*weights;

	scale = 1.0 / sqrt(3.0);
	weights = sq4_intg->intg_weights;

	unit = dmat(sq4_intg->intg_unit);
	intp = dmat(sq4_intg->intg_intgcoors);
	for (i=0; i < 4; i++) {
		intp[i][0] = unit[i][0] * scale;
		intp[i][1] = unit[i][1] * scale;
		weights[i] = 1.0;
	}
	return;
}

static void
sq4_mktrans(void)
{
	double	**mat;
	double	**unit, **intp;
	double	ksi, eta;
	double	c_ksi, c_eta;
	int	i, j;

	mat = dmat(sq4_intg->intg_trans);
	intp = dmat(sq4_intg->intg_intgcoors);
	unit = dmat(sq4_intg->intg_unit);
	for (i=0; i <  4; i++) {	/* for four points */
		ksi = intp[i][0];
		eta = intp[i][1];

		for (j=0;j< 4;j++) {
			c_ksi = unit[j][0];
			c_eta = unit[j][1];

			mat[i][j] = 0.25 * (1.0 + c_ksi * ksi)
					 * (1.0 + c_eta * eta);
		}
	}
	return;
}
		

static void
sq4_mkhalfjacobm(void)
{
	double	ksi, eta;
	double	c_ksi, c_eta;
	double	**intp, **unit;
	struct matrix_t	*fmat;
	double	**fmatp;
	int	i, j;

	unit = dmat(sq4_intg->intg_unit);
	intp = dmat(sq4_intg->intg_intgcoors);
	for (i=0; i<4;i++) {
		fmat = sq4_intg->intg_halfjacobm[i];
		fmatp = dmat(fmat);

		ksi = intp[i][0];
		eta = intp[i][1];

		for (j=0;j< 4;j++) {
			c_ksi = unit[j][0];
			c_eta = unit[j][1];

			fmatp[0][j] = 0.25 * c_ksi * (1.0 + c_eta * eta);
			fmatp[1][j] = 0.25 * c_eta * (1.0 + c_ksi * ksi);
		}
	}
	return;
}
