#include "ssp_defs.h"
#include "util.h"
#include "scomplex.h"

#include "common.h"
#include "vector.h"
#include "sparse.h"
#include "finite.h"
#include "super.h"

void
solveall(void)
{
	SuperMatrix A, L, U, B;
	int	*perm_c, *perm_r;
	int	permc_spec;
	int	info;
	int	size;

	struct sparse_t *sp;

	/*
	 * solve
	 */
	sp = meshes->m_kmatrix;
	size = sp->sp_rown;
	switch (meshes->m_datatype) {
		case SP_SINGLE:
			sCreate_CompCol_Matrix(&A, sp->sp_rown,
				sp->sp_coln, sp->sp_nonzero,
				(float *)sp->sp_data, sp->sp_rowx, sp->sp_colx,
				NC, __S, GE);

			sCreate_Dense_Matrix(&B, size, 1,
				svec(meshes->m_bvector),
				meshes->m_nodesno, DN, __S, GE);
			break;

		case SP_DOUBLE:
			dCreate_CompCol_Matrix(&A, sp->sp_rown,
				sp->sp_coln, sp->sp_nonzero,
				(double *)sp->sp_data, sp->sp_rowx, sp->sp_colx,
				NC, __D, GE);

			dCreate_Dense_Matrix(&B, size, 1,
				dvec(meshes->m_bvector),
				meshes->m_nodesno, DN, __D, GE);
			break;

		case SP_SIMPLEX:
			cCreate_CompCol_Matrix(&A, sp->sp_rown,
				sp->sp_coln, sp->sp_nonzero,
				(complex *)sp->sp_data, sp->sp_rowx,
				sp->sp_colx, NC, __C, GE);

			cCreate_Dense_Matrix(&B, size, 1,
				(complex *) cvec(meshes->m_bvector),
				meshes->m_nodesno, DN, __C, GE);
			break;

		case SP_DUPLEX:
			zCreate_CompCol_Matrix(&A, sp->sp_rown,
				sp->sp_coln, sp->sp_nonzero,
				(doublecomplex *)sp->sp_data,
				sp->sp_rowx, sp->sp_colx, NC, __Z, GE);

			zCreate_Dense_Matrix(&B, size, 1,
				(doublecomplex *) zvec(meshes->m_bvector),
				meshes->m_nodesno, DN, __Z, GE);
			break;
		default:
			fatal("%s(): IMPOSSIBLE", __F);
	}

	if (!(perm_r = intMalloc(meshes->m_kmatrix->sp_rown)) )
		ABORT("Malloc fails for perm_r[].");
	if (!(perm_c = intMalloc(meshes->m_kmatrix->sp_coln)) )
		ABORT("Malloc fails for perm_c[].");

	/*
	 * Get column permutation vector perm_c[], according to permc_spec:
	 *   permc_spec = 0: use the natural ordering
	 *   permc_spec = 1: use minimum degree ordering on structure of A'*A
	 *   permc_spec = 2: use minimum degree ordering on structure of A'+A
	 */

	permc_spec = 0;
	get_perm_c(permc_spec, &A, perm_c);

	printf("%s(): Right before solving()=%s\n", __F, get_currtime());

	switch (meshes->m_datatype) {
		case SP_SINGLE:
			sgssv(&A, perm_c, perm_r, &L, &U, &B, &info);
			break;
		case SP_DOUBLE:
			dgssv(&A, perm_c, perm_r, &L, &U, &B, &info);
			break;
		case SP_SIMPLEX:
			cgssv(&A, perm_c, perm_r, &L, &U, &B, &info);
			break;
		case SP_DUPLEX:
			zgssv(&A, perm_c, perm_r, &L, &U, &B, &info);
			break;
		default:
			fatal("%s(): IMPOSSIBLE", __F);
	}

	printf("%s(): Right after  solving()=%s\n", __F, get_currtime());
	printf("%s(): Right after solving I got info = %d\n", __F, info);
	if (info) {
		putchar(7);
		printf("%s(): ERROR OCCURED!!\n", __F);
		return;
	}
		

	/*
	 * Free
	 */
	SUPERLU_FREE (perm_r);
	SUPERLU_FREE (perm_c);
#if 0
	Destroy_CompCol_Matrix(&A);
#else
	Destroy_SuperMatrix_Store(&A);
#endif
	Destroy_SuperMatrix_Store(&B);
	Destroy_SuperNode_Matrix(&L);
	Destroy_CompCol_Matrix(&U);
	return;
}
