#include <stdio.h>
#include "common.h"
#include "vector.h"
#include "matrix.h"
#include "finite.h"

static void     tri3_setunit(void);
static void	tri3_integpoints(void);
static void	tri3_mkhalfjacobm(void);
static void	tri3_mktrans(void);

struct integral_t *tri3_intg;

int
tri3_init(struct proto_t *p)
{
	int	*lseg;
	static int tri3_inited = 0;

	if (tri3_inited)
		return 0;

	tri3_inited++;
	tri3_intg = p->p_intg;

	/*
	 * ORDER is important
	 */
	tri3_setunit();
	tri3_integpoints();
	tri3_mkhalfjacobm();
	tri3_mktrans();

	lseg = p->p_lineseg;
	lseg[0] = 0; lseg[1] = 1; lseg[2] = 2; lseg[3] = 0;

	return 0;
}

static void
tri3_setunit(void)
{
	struct matrix_t *matf;
	double	**mat;

	matf = tri3_intg->intg_unit;
	mat  = dmat(matf);

	mat[0][0] =  0.0; mat[0][1] =  0.0;
	mat[1][0] =  1.0; mat[1][1] =  0.0;
	mat[2][0] =  0.0; mat[2][1] =  1.0;

	return;
}


void
tri3_integpoints(void)
{
	double	**intgcoor;
	double	 *weights;

	weights = tri3_intg->intg_weights;
	weights[0] = 1.0/2.0;

	intgcoor = dmat(tri3_intg->intg_intgcoors);
	intgcoor[0][0] = 1.0/3.0;
	intgcoor[0][1] = 1.0/3.0;
	return;
}

static void
tri3_mktrans(void)
{
	double	**mat;

	mat = dmat(tri3_intg->intg_trans);
	mat[0][0] = 1.0/3.0; mat[0][1] = 1.0/3.0; mat[0][2] = 1.0/3.0;

	return;
}
		

static void
tri3_mkhalfjacobm(void)
{
	double	**mat;

	mat = dmat(tri3_intg->intg_halfjacobm[0]);
	mat[0][0] = -1.0; mat[0][1] = +1.0; mat[0][2] =  0.0;
	mat[1][0] = -1.0; mat[1][1] =  0.0; mat[1][2] = +1.0;

	return;
}

