#include <stdio.h>
#include <sys/types.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

#include "common.h"
#include "complex.h"

char	*noname="[no name]";
int	verbose;
char	commonbuf[256];
char *typenames[]={
	"NONE type",
	"SP_INTEGER type",
	"SP_SINGLE type",
	"SP_DOUBLE type",
	"SP_SIMPLEX type",
	"SP_DUPLEX type",
	NULL,
};


struct duplex_t duplex_null    = { 0.0, 0.0};
struct duplex_t duplex_unit    = { 1.0, 0.0};

struct simplex_t simplex_null  = { 0.0, 0.0};
struct simplex_t simplex_unit  = { 1.0, 0.0};

void
not_yet(const char *function)
{
	(void) printf("function %s() has something that is not implemented yet",
		function);
	exit(1);
}

/*
 * used only at patt_rgetpatt(), patt_cgetpatt()
 */
int
patt_getwidth(const char *pattern)
{
	char	*p;
	int	readn, span;
	
	p = (char *) pattern;
	if (*p++ != '%')
		fatal("%s(\"%s\"): first char is not %%", __F, pattern);
	readn = sscanf(p, "%d", &span);
	if (readn != 1)
		fatal("%s(\"%s\"): after %% is not integer", __F, pattern);
	return span;
}


/*
 * real version
 * used begug-routines like m_print
 */
void
patt_rgetpatt(char *target, char *spaces, const char *patt, const char *term)
{
	char	*p;
	int	i, width;

	strcpy(target, patt);
	if (term)
		strcat(target, term);
	p = target + strlen(target);
	*p++ = ' ';
	*p   = 0;

	width = patt_getwidth(patt);

	p = spaces;
	for (i=0;i<width;i++)
		*p++ = ' ';
	*p++ = 0;
	if (term)
		strcat(spaces, term);
	p = spaces + strlen(spaces);
	*p++ = ' ';
	*p++ = 0;
		
	return;
}

/*
 * complex version
 * used begug-routines like m_print
 */
void
patt_cgetpatt(char *target, char *spaces, const char *patt, const char *term)
{
	char	*p;
	int	i, width;

	strcpy(target, patt);	/* real */
	p = target + strlen(target);
	*p++ = ' ';
	*p++ = 0;
	strcat(target, patt);	/* imag */
	if (term)
		strcat(target, term);
	p = target + strlen(target);
	*p++ = ' ';
	*p   = 0;

	width = patt_getwidth(patt);

	p = spaces;
	for (i=0;i<(2*width+1);i++)
		*p++ = ' ';
	*p++ = 0;
	if (term)
		strcat(spaces, term);
	p = spaces + strlen(spaces);
	*p++ = ' ';
	*p++ = 0;
		
	return;
}

/*
 * Following code is just for debugging... like
 * gdb> p put(sqrt(10.0))
 * I dont know this will be useful, ie, this will operate.
 */
void
put(double x)
{
	printf("%10.5f\n", x);
	return;
}

#define SQR(a)	(a*a)
double
pythag(double a, double b)
{
	double	absa,absb;

	absa=fabs(a);
	absb=fabs(b);
	if (absa > absb)
		return absa*sqrt(1.0+SQR(absb/absa));
	else
		return (absb == 0.0 ? 0.0 : absb*sqrt(1.0+SQR(absa/absb)));
}

void
fatal(const char *fmt, ...)
{
	char	line[128];
	va_list ap;
	va_start(ap, fmt);

	(void)fprintf(stderr, "fatal error:\n");
	(void)vfprintf(stderr, fmt, ap);
	va_end(ap);
	(void)fprintf(stderr, "\n\nCongratulation!!! You met a bug\n");
	(void)fprintf(stderr, "Type any charater!!\n");
	gets(line);
	exit(1);
	/* NOTREACHED */
}

void
memerr(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);

	(void)fprintf(stderr, "memory error: ");
	(void)vfprintf(stderr, fmt, ap);
	va_end(ap);
	(void)fprintf(stderr, "\n");
	exit(1);
	/* NOTREACHED */
}

void
err(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);

	(void)fprintf(stderr, "err error: ");
	(void)vfprintf(stderr, fmt, ap);
	va_end(ap);
	(void)fprintf(stderr, "\n");
}


int
islineok(const char *line)
{
	char *p;

	for (p = (char *) line; isspace(*p); p++)
		/* void */;
	if (*p)
		return 1;
	return 0;
}

char *
skipto(const char *line, char c)
{
	char	*p;
	if (line == NULL)
		fatal("skipto() fatal");

	p = index(line, c);
	if (p == NULL)
		return (char *) NULL;
	return p;
}


void
bitput(int n, char *buf)
{
	int	i;
	char	*q=buf;

	if (buf == NULL)
		fatal("%s(): OOPS.. you must be a stupid one!!", __F);
	q[32] = 0;
	for (i=0; i < 32; i++) {
		*q++ = (n & 0x1) ? '1' : '.';
		n >>= 1;
	}
	return;
}

float
sign(float x)
{
	if (x > 0.0)
		return 1.0;
	else
		return -1.0;
}


char *
skipwhites(char *p)
{
	char *q;

	for (q=p; isspace(*q); q++)
		/* void */ ;

	if (*q == NULL)
		fatal("White spaces??? Nothing left in this string!!");

	return q;
}

#define OK_NO(c) (isdigit(c) || (c == 'e') || index("+-", c))
char *
skipnumer(char *p)
{
	if (!OK_NO(*p) && *p != '.')
		fatal("skipnumer(): while reading %s", p);
	while (OK_NO(*p) || *p == '.')
		p++;

	return p;
}
#undef OK_NO

char *
get_ordinal(const int order)
{
	char	*p;
	static char *ord[3]={"1st", "2nd", "3rd"};
	static char tmp[5];

	if (order < 4)
		p = ord[order-1];
	else {
		sprintf(tmp, "%dth", order);
		p = tmp;
	}
	return p;
}

void
put_currtime(const char *func, const char *msg)
{
	static int steps = 0;
	char	*zeit;

	zeit = get_currtime();
	if (msg)
		printf("[%3d]: %s at %s(%s)\n", steps++, zeit, func, msg);
	else
		printf("[%3d]: %s at %s()\n", steps++, zeit, func);
	return;
}

char *
get_currtime(void)
{
	static time_t before, cette;
	static char timebuf[48], firstcalled = 1;
	time_t elapsed;
	unsigned int day, hour, min, sec;
	char	*zeit, *p;
	
	time(&cette);
	if (firstcalled) {
		elapsed = 0;
		firstcalled = 0;
	} else
		elapsed = cette - before;

	sec  = elapsed % 60; elapsed /= 60;
	min  = elapsed % 60; elapsed /= 60;
	hour = elapsed % 24; elapsed /= 24;
	day  = elapsed;

	zeit = ctime(&cette);
	strcpy(timebuf, zeit);
	p = &timebuf[strlen(timebuf)-1];
	sprintf(p, "  %02dd/%02dh/%02dm/%02ds", day, hour, min, sec);

	before = cette;

	return timebuf;
}
