#include <stdio.h>
#include <math.h>
#include "common.h"
#include "complex.h"

struct duplex_t
duplexv(const double real, const double imag)
{
	struct duplex_t dx;

	dx.real = real;
	dx.imag = imag;

	return dx;
}

struct simplex_t
simplexv(const double real, const double imag)
{
	struct simplex_t cx;

	cx.real = real;
	cx.imag = imag;

	return cx;
}

struct duplex_t *
zp_assign(struct duplex_t *ret, const struct duplex_t *src)
{
	ret->real = src->real;
	ret->imag = src->imag;
	return ret;
}

struct simplex_t *
cp_assign(struct simplex_t *ret, const struct simplex_t *src)
{
	ret->real = src->real;
	ret->imag = src->imag;
	return ret;
}


struct duplex_t *
zp_expon(struct duplex_t *ret, const struct duplex_t *z)
{
	double expx;
	expx = exp(z->real);
	ret->real = expx * cos(z->imag);
	ret->imag = expx * sin(z->imag);
	return ret;
}

struct duplex_t *
zp_conjug(struct duplex_t *ret, const struct duplex_t *src)
{
	ret->real =  src->real;
	ret->imag = -src->imag;
	return ret;
}

void
zp_print(const struct duplex_t *z)
{
	(void) printf("%10.5f %10.5f+\n", z->real, z->imag);
	return;
}

int
zp_isnull(const struct duplex_t *z)
{
	if (fabs(z->real) || fabs(z->imag))
		return 0;
	return 1;
}


struct duplex_t *
zp_inverse(struct duplex_t *ret, const struct duplex_t *src)
{
	zp_slash(ret, &duplex_unit, src);
	return ret;
}

struct duplex_t *
zp_plus(struct duplex_t *ret, const struct duplex_t *a,const struct duplex_t *b)
{
	ret->real = a->real + b->real;
	ret->imag = a->imag + b->imag;
	return ret;
}

struct duplex_t *
zp_pplus(struct duplex_t *ret, const struct duplex_t *a)
{
	ret->real += a->real;
	ret->imag += a->imag;
	return ret;
}

struct duplex_t *
zp_3plus(struct duplex_t *ret, const struct duplex_t *a,
	const struct duplex_t *b, const struct duplex_t *c)
{
	ret->real = a->real + b->real + c->real;
	ret->imag = a->imag + b->imag + c->imag;
	return ret;
}

struct duplex_t *
zp_mminus(struct duplex_t *ret, const struct duplex_t *a)
{
	ret->real -= a->real;
	ret->imag -= a->imag;
	return ret;
}

struct duplex_t *
zp_minus(struct duplex_t *ret,const struct duplex_t *a,const struct duplex_t *b)
{
	ret->real = a->real - b->real;
	ret->imag = a->imag - b->imag;
	return ret;
}

struct duplex_t *
zp_cross(struct duplex_t *ret,const struct duplex_t *a,const struct duplex_t *b)
{
	ret->real = a->real * b->real - a->imag * b->imag;
	ret->imag = a->imag * b->real + a->real * b->imag;

	return ret;
}

struct duplex_t *
zp_3cross(struct duplex_t *ret, const struct duplex_t *a,
		const struct duplex_t *b, const struct duplex_t *c)
{
	ret->real =	 (a->real * b->real * c->real)
		  	-(a->real * b->imag * c->imag)
			-(a->imag * b->imag * c->real)
			-(a->imag * b->real * c->imag);

	ret->imag =	 (a->imag * b->real * c->real)
		  	+(a->real * b->imag * c->real)
			+(a->real * b->real * c->imag)
			-(a->imag * b->imag * c->imag);

	return ret;
}

struct duplex_t *
zp_crossdme(struct duplex_t *ret, const double b)
{
	ret->real  = ret->real * b;
	ret->imag  = ret->imag * b;

	return ret;
}

struct duplex_t *
zp_crosszme(struct duplex_t *ret, const struct duplex_t b)
{
	struct duplex_t ztmp;
	ztmp.real = ret->real * b.real - ret->imag * b.imag;
	ztmp.imag = ret->imag * b.real + ret->real * b.imag;

	ret->real = ztmp.real;
	ret->imag = ztmp.imag;
	return ret;
}

struct duplex_t *
zp_crossd(struct duplex_t *ret, const struct duplex_t *a, const double b)
{
	ret->real  = a->real * b;
	ret->imag  = a->imag * b;

	return ret;
}

#if 0
struct duplex_t *
zp_slash(struct duplex_t *ret, const struct duplex_t *a,
		const struct duplex_t *b)
{
	double division;

	division = 1.0/(b->real * b->real + b->imag * b->imag);
	ret->real =  (a->real * b->real + a->imag * b->imag) * division;
	ret->imag =  (a->imag * b->real - a->real * b->imag) * division;

	return ret;
}
#else
struct duplex_t *
zp_slash(struct duplex_t *c, const struct duplex_t *a, const struct duplex_t *b)
{
	double ratio, den;
	double abr, abi, cr, ci;

	if( (abr = b->real) < 0.)
		abr = - abr;
	if( (abi = b->imag) < 0.)
		abi = - abi;
	if( abr <= abi ) {
		if (abi == 0) {
			fatal("z_div.c: division by zero");
		}
		ratio = b->real / b->imag ;
		den = b->imag * (1 + ratio*ratio);
		cr = (a->real*ratio + a->imag) / den;
		ci = (a->imag*ratio - a->real) / den;
	} else {
		ratio = b->imag / b->real ;
		den = b->real * (1 + ratio*ratio);
		cr = (a->real + a->imag*ratio) / den;
		ci = (a->imag - a->real*ratio) / den;
	}
	c->real = cr;
	c->imag = ci;

	return c;
}
#endif

double
zp_absolute(const struct duplex_t *z)
{
	double temp;
	double real = z->real;
	double imag = z->imag;

	if (real < 0)
		real = -real;
	if (imag < 0)
		imag = -imag;
	if (imag > real) {
		temp = real;
		real = imag;
		imag = temp;
	}
	if ((real+imag) == real)
		return real;

	temp = imag/real;
	temp = real*sqrt(1.0 + temp*temp);  /*overflow!!*/

	return temp;
}

double
zp_square(const struct duplex_t *a)
{
	return (a->real * a->real + a->imag * a->imag);
}

struct duplex_t *
zp_conjcross(struct duplex_t *ret, const struct duplex_t *a,
		const struct duplex_t *b)
{
	ret->real =  a->real * b->real + a->imag * b->imag;
	ret->imag = -a->imag * b->real + a->real * b->imag;

	return ret;
}

struct simplex_t *
cp_expon(struct simplex_t *ret, const struct simplex_t *z)
{
	double expx;
	expx = exp(z->real);
	ret->real = expx * cos(z->imag);
	ret->imag = expx * sin(z->imag);
	return ret;
}

struct simplex_t *
cp_conjug(struct simplex_t *ret, const struct simplex_t *src)
{
	ret->real =  src->real;
	ret->imag = -src->imag;
	return ret;
}
void
cp_print(const struct simplex_t *z)
{
	(void) printf("%10.5f %10.5f+\n", z->real, z->imag);
	return;
}


struct simplex_t *
cp_inverse(struct simplex_t *ret, const struct simplex_t *src)
{
	cp_slash(ret, &simplex_unit, src);
	return ret;
}

struct simplex_t *
cp_plus(struct simplex_t *ret, const struct simplex_t *a,const struct simplex_t *b)
{
	ret->real = a->real + b->real;
	ret->imag = a->imag + b->imag;
	return ret;
}
struct simplex_t *
cp_pplus(struct simplex_t *ret, const struct simplex_t *a)
{
	ret->real = ret->real + a->real;
	ret->imag = ret->imag + a->imag;
	return ret;
}

struct simplex_t *
cp_3plus(struct simplex_t *ret, const struct simplex_t *a,
	const struct simplex_t *b, const struct simplex_t *c)
{
	ret->real = a->real + b->real + c->real;
	ret->imag = a->imag + b->imag + c->imag;
	return ret;
}

struct simplex_t *
cp_mminus(struct simplex_t *ret, const struct simplex_t *a)
{
	ret->real = ret->real - a->real;
	ret->imag = ret->imag - a->imag;
	return ret;
}

struct simplex_t *
cp_minus(struct simplex_t *ret,const struct simplex_t *a,const struct simplex_t *b)
{
	ret->real = a->real - b->real;
	ret->imag = a->imag - b->imag;
	return ret;
}

struct simplex_t *
cp_cross(struct simplex_t *ret,const struct simplex_t *a,const struct simplex_t *b)
{
	ret->real = a->real * b->real - a->imag * b->imag;
	ret->imag = a->imag * b->real + a->real * b->imag;

	return ret;
}

struct simplex_t *
cp_3cross(struct simplex_t *ret, const struct simplex_t *a,
		const struct simplex_t *b, const struct simplex_t *c)
{
	ret->real =	 (a->real * b->real * c->real)
		  	-(a->real * b->imag * c->imag)
			-(a->imag * b->imag * c->real)
			-(a->imag * b->real * c->imag);

	ret->imag =	 (a->imag * b->real * c->real)
		  	+(a->real * b->imag * c->real)
			+(a->real * b->real * c->imag)
			-(a->imag * b->imag * c->imag);

	return ret;
}

struct simplex_t *
cp_crossdme(struct simplex_t *ret, const double b)
{
	ret->real  = ret->real * b;
	ret->imag  = ret->imag * b;

	return ret;
}

struct simplex_t *
cp_crosszme(struct simplex_t *ret, const struct duplex_t b)
{
	struct duplex_t ztmp;
	ztmp.real = ret->real * b.real - ret->imag * b.imag;
	ztmp.imag = ret->imag * b.real + ret->real * b.imag;

	ret->real = ztmp.real;
	ret->imag = ztmp.imag;
	return ret;
}


struct simplex_t *
cp_crossd(struct simplex_t *ret, const struct simplex_t *a, const double b)
{
	ret->real  = a->real * b;
	ret->imag  = a->imag * b;

	return ret;
}

#if 0
struct simplex_t *
cp_slash(struct simplex_t *ret, const struct simplex_t *a,
	const struct simplex_t *b)
{
	double division;

	division = 1.0/(b->real * b->real + b->imag * b->imag);
	ret->real =  (a->real * b->real + a->imag * b->imag)/division;
	ret->imag =  (a->imag * b->real - a->real * b->imag)/division;

	return ret;
}
#else
struct simplex_t *
cp_slash(struct simplex_t *c, const struct simplex_t *a,
	const struct simplex_t *b)
{
	double ratio, den;
	double abr, abi, cr, ci;

	if( (abr = b->real) < 0.)
		abr = - abr;
	if( (abi = b->imag) < 0.)
		abi = - abi;
	if( abr <= abi ) {
		if (abi == 0) {
			fatal("z_div.c: division by zero");
		}
		ratio = b->real / b->imag ;
		den = b->imag * (1 + ratio*ratio);
		cr = (a->real*ratio + a->imag) / den;
		ci = (a->imag*ratio - a->real) / den;
	} else {
		ratio = b->imag / b->real ;
		den = b->real * (1 + ratio*ratio);
		cr = (a->real + a->imag*ratio) / den;
		ci = (a->imag - a->real*ratio) / den;
	}
	c->real = cr;
	c->imag = ci;

	return c;
}
#endif

double
cp_absolute(const struct simplex_t *z)
{
	double temp;
	double real = z->real;
	double imag = z->imag;

	if (real < 0)
		real = -real;
	if (imag < 0)
		imag = -imag;
	if (imag > real) {
		temp = real;
		real = imag;
		imag = temp;
	}
	if ((real+imag) == real)
		return real;

	temp = imag/real;
	temp = real*sqrt(1.0 + temp*temp);  /*overflow!!*/

	return temp;
}

double
cp_square(const struct simplex_t *a)
{
	return (a->real * a->real + a->imag * a->imag);
}

struct simplex_t *
cp_conjcross(struct simplex_t *ret, const struct simplex_t *a,
		const struct simplex_t *b)
{
	ret->real =  a->real * b->real + a->imag * b->imag;
	ret->imag = -a->imag * b->real + a->real * b->imag;

	return ret;
}
int
cp_isnull(const struct simplex_t *z)
{
	if (fabs(z->real) || fabs(z->imag))
		return 0;
	return 1;
}

/*
 * CONVERSION
 */
struct duplex_t *
cz_sim2du(struct duplex_t *d, const struct simplex_t *s)
{
	d->real = s->real;
	d->imag = s->imag;
	return d;
}

struct simplex_t *
cz_du2sim(struct simplex_t *s, const struct duplex_t *d)
{
	s->real = d->real;
	s->imag = d->imag;
	return s;
}
