/*
 * code was copied from  Numeric Recipes in C
 * CAN ONLY handle single type matrix
 */
#include <stdio.h>
#include <math.h>
#include "common.h"
#include "vector.h"
#include "matrix.h"

#define TINY 1.0e-20;

#define PIVOT 1

void
mat_ludecomp(struct matrix_t *am, struct vector_t *indxv, float *d,
	const int ind)
{
	int i,imax,j,k;
	float big,dum,sum,temp;
	float	*sp;
	float	**a;
	int	*indx;
	struct vector_t *vp0;
	int	n;

	n    = am->mt_col;
	a    = smat(am);
	indx = ivec(indxv);
	vp0  = vector(n, "tmp", SP_SINGLE);
	sp   = svec(vp0);

	*d=1.0;
	if (ind)
		for (i=0;i < n;i++) {
			big=0.0;
			for (j=0;j < n;j++)
				if ((temp=fabs(a[i][j])) > big)
					big=temp;
			if (big == 0.0)
				fatal("Singular matrix in routine ludcmp");
			sp[i]=1.0/big;
		}

	for (j=0;j < n;j++) {
		for (i=0;i<j;i++) {
			sum=a[i][j];
			for (k=0;k<i;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i < n;i++) {
			sum=a[i][j];
			for (k=0;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;

			if (ind)
				if ((dum=sp[i]*fabs(sum)) >= big) {
					big=dum;
					imax=i;
				}
		}
		if (ind)
			if (j != imax) {
				printf("CHANGE!");
				for (k=0;k < n;k++) {
					dum=a[imax][k];
					a[imax][k]=a[j][k];
					a[j][k]=dum;
				}
				*d = -(*d);
				sp[imax]=sp[j];
			}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != (n-1)) { /* XXX */
			dum=1.0/(a[j][j]);
			for (i=j+1;i < n;i++)
				a[i][j] *= dum;
		}
	}
	if (ind == 0)
		for (i=0;i<indxv->vt_size;i++)
			indx[i] = i;

	vec_free(vp0);
	printf("\n");
	
}

void
mat_lubacksub(const struct matrix_t *af, const struct vector_t *indxv,
	struct vector_t *bv)
{
	int	i, ii= -1, ip, j, n;
	float	sum;
	float	**a, *b;
	int	*indx;

	n = af->mt_row;
	a = smat(af);
	b = svec(bv);
	indx = ivec(indxv);

	for (i = 0;i < n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii > -1)
			for (j=ii;j<=(i-1);j++)
				sum -= a[i][j]*b[j];
		else if (sum)
			ii=i;
		b[i]=sum;
	}
	for (i = (n-1);i >= 0;i--) {
		sum=b[i];
		for (j=i+1;j < n;j++)
			sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
	return;
}
