#include <stdio.h>
#include <math.h>
#include "common.h"
#include "matrix.h"


inline float **
smat(const struct matrix_t *mp)
{
	m_typecheck(__F, 1, mp, SP_SINGLE);
	return (float **) mp->mt_data;
}

float
sm_getdet(const struct matrix_t *mp)
{
	float	**pp;
	float	det;

	m_typecheck(__F, 1, mp, SP_SINGLE);
	m_squarecheck(__F, mp);

	pp = smat(mp);
	switch (mp->mt_row) {
		case 1: det =  pp[0][0]; break;
		case 2: det =  pp[0][0]*pp[1][1] - pp[0][1]*pp[1][0]; break;
		case 3:
			det =     pp[0][0]*pp[1][1]*pp[2][2]
				+ pp[0][1]*pp[1][2]*pp[2][0]
				+ pp[0][2]*pp[1][0]*pp[2][1]
				- pp[2][0]*pp[1][1]*pp[0][2]
				- pp[1][0]*pp[0][1]*pp[2][2]
				- pp[0][0]*pp[2][1]*pp[1][2];
			break;
		default:
			fatal("%s(): IMPOSSIBLE", __F);
	}
	
	return det;
}


float
sm_inverse(struct matrix_t *target, const struct matrix_t *af)
{
	float	**mat, **inv;
	float	det;

	mat = smat(af); inv = smat(target);

	switch (af->mt_row) {
		case 2:
			det = mat[0][0]*mat[1][1] - mat[0][1]*mat[1][0];
			inv[0][0] =        mat[1][1]/det;
			inv[0][1] = -1.0 * mat[0][1]/det;
			inv[1][0] = -1.0 * mat[1][0]/det;
			inv[1][1] =        mat[0][0]/det;
			break;
		case 3:
			det =     mat[0][0]*mat[1][1]*mat[2][2]
				+ mat[0][1]*mat[1][2]*mat[2][0]
				+ mat[0][2]*mat[1][0]*mat[2][1]
				- mat[2][0]*mat[1][1]*mat[0][2]
				- mat[1][0]*mat[0][1]*mat[2][2]
				- mat[0][0]*mat[2][1]*mat[1][2];

			inv[0][0]= mat[1][1]*mat[2][2]-mat[1][2]*mat[2][1];
			inv[0][1]= mat[0][1]*mat[2][2]-mat[2][1]*mat[0][2];
			inv[0][2]= mat[0][1]*mat[1][2]-mat[1][1]*mat[0][2];

			inv[1][0]= mat[1][0]*mat[2][2]-mat[2][0]*mat[1][2];
			inv[1][1]= mat[0][0]*mat[2][2]-mat[2][0]*mat[0][2];
			inv[1][2]= mat[0][0]*mat[1][2]-mat[1][0]*mat[0][2];

			inv[2][0]= mat[1][0]*mat[2][1]-mat[2][0]*mat[1][1];
			inv[2][1]= mat[0][0]*mat[2][1]-mat[2][0]*mat[0][1];
			inv[2][2]= mat[0][0]*mat[1][1]-mat[1][0]*mat[0][1];

			inv[0][1] *=  -1.0;     
			inv[1][0] *=  -1.0;     
			inv[2][1] *=  -1.0;     
			inv[1][2] *=  -1.0;     

			inv[0][0] /= det;
			inv[0][1] /= det;
			inv[0][2] /= det;
			inv[1][0] /= det;
			inv[1][1] /= det;
			inv[1][2] /= det;
			inv[2][0] /= det;
			inv[2][1] /= det;
			inv[2][2] /= det;

			break;
		default:
			fatal("%s(): dim of matrix should be 2 or 3", __F);
	}
	return det;
}

inline float *
sm_row(const struct matrix_t *af, int row)
{
	m_typecheck(__F, 1, af, SP_SINGLE);

	return (float *) (af->mt_data[row]);
}

int
sm_issymmet(const struct matrix_t *af)
{
	int	i, j;
	float	**data;

	data = smat(af);
	for (i=0;i< af->mt_row;i++)
		for (j=0;j < i;j++) {
			if (fabs(data[i][j] - data[j][i]) > CUTOFF)
				return 0;
		}
	return 1;
}
