#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include "common.h"
#include "complex.h"
#include "vector.h"
#include "matrix.h"

struct matrix_t *
matrix(const ushort_t row, const ushort_t col, const char *name, const int type)
{
	struct matrix_t	*mt;
	int	i, howmuch;
	double	*dp, **dpp;
	float	*sp, **spp;
	struct duplex_t *dxp, **dxpp;
	struct simplex_t *cxp, **cxpp;

	if (type != SP_SINGLE && type != SP_DOUBLE
		&& type != SP_SIMPLEX  && type != SP_DUPLEX)
		fatal("%s(\"%s\"): with wrong type", __F, name);

	if (row < 1 || col < 1)
		fatal("%s(): wrong col || row no in fmatrix()", __F);

	mt = (struct matrix_t *) malloc((size_t)(sizeof(struct matrix_t)));
	if (IS_NULL(mt))
		memerr("%s(): mt alloc", __F);

	mt->mt_row  = row;
	mt->mt_col  = col;
	mt->mt_type = type;

	/*
	 * give matrix name if set
	 */
	if (name && *name)
		mt->mt_name = strdup(name);
	else
		mt->mt_name = noname;

	switch (type) {
		case SP_SINGLE:
			spp = (float **)
				malloc((size_t)(sizeof(float *) * row));
			if (IS_NULL(spp))
				memerr("%s(\"%s\"): spp", __F, name);

			howmuch = sizeof(float) * row * col;
			sp = (float *)
				malloc((size_t) howmuch);
			if (IS_NULL(sp))
				memerr("%s(\"%s\"): sp", __F, name);

			for (i=0; i < row; i++, sp += col) {
				spp[i] = sp;
			}
			mt->mt_data = (int **) spp;
			mt->mt_type = SP_SINGLE;
			break;

		case SP_DOUBLE:
			dpp = (double **)
				malloc((size_t)(sizeof(double *) * row));
			if (IS_NULL(dpp))
				memerr("%s(\"%s\"): dpp", __F, name);

			howmuch = sizeof(double) * row * col;
			dp = (double *)
				malloc((size_t) howmuch);

			if (IS_NULL(dp))
				memerr("%s(\"%s\"): dx", __F, name);

			for (i=0; i < row; i++, dp += col) {
				dpp[i] = dp;
			}
			mt->mt_data = (int **) dpp;
			mt->mt_type = SP_DOUBLE;
			break;

		case SP_SIMPLEX:
			cxpp = (struct simplex_t **) malloc((size_t)
				(sizeof(struct simplex_t *) * row));
			if (IS_NULL(cxpp))
				memerr("%s(\"%s\"): cxp", __F, name);

			howmuch = sizeof(struct simplex_t) * row * col;
			cxp = (struct simplex_t *)
				malloc((size_t) howmuch);

			if (IS_NULL(cxp))
				memerr("%s(\"%s\"): cx", __F, name);

			for (i=0; i < row; i++, cxp += col) {
				cxpp[i] = cxp;
			}
			mt->mt_data = (int **) cxpp;
			mt->mt_type = SP_SIMPLEX;
			break;

		case SP_DUPLEX:
			dxpp = (struct duplex_t **) malloc((size_t)
				(sizeof(struct duplex_t *) * row));
			if (IS_NULL(dxpp))
				memerr("%s(\"%s\"): dxp", __F, name);

			howmuch = sizeof(struct duplex_t) * row * col;
			dxp = (struct duplex_t *)
				malloc((size_t) howmuch);

			if (IS_NULL(dxp))
				memerr("%s(\"%s\"): dx", __F, name);

			for (i=0; i < row; i++, dxp += col) {
				dxpp[i] = dxp;
			}
			mt->mt_data = (int **) dxpp;
			mt->mt_type = SP_DUPLEX;
			break;
	}

	memset((void *) mt->mt_data[0], 0x00, howmuch);

	return mt;
}

void
m_free(struct matrix_t *mp)
{
	if (mp == NULL)
		fatal("%s(): matrix is NULL", __F);


	free(mp->mt_data[0]);
	free(mp->mt_data);
	if (strcmp(mp->mt_name, noname))
		free(mp->mt_name);
	free(mp);
	return;
}

void
m_mfree(int num,...)
{
	int     i;
	va_list ap;
	struct matrix_t *mp;

	va_start(ap, num);
	for (i=0;i<num;i++)  {
		mp = va_arg(ap, struct matrix_t *);
		m_free(mp);
	}
	va_end(ap);
	return;
}

void
m_null(struct matrix_t *m)
{
	int	size;
	m_check(__F, 1, m);
	switch (m->mt_type) {
		case SP_SINGLE:
			size = m->mt_row * m->mt_col * sizeof(float);
			break;
		case SP_DOUBLE:
			size = m->mt_row * m->mt_col * sizeof(double);
			break;
		case SP_SIMPLEX:
			size = m->mt_row * m->mt_col * sizeof(struct simplex_t);
			break;
		case SP_DUPLEX:
			size = m->mt_row * m->mt_col * sizeof(struct duplex_t);
			break;
		default:
			fatal("%s(): IMPOSSIBLE type", __F);
	}
	memset(m->mt_data[0], 0x00, (size_t) size);
	return;
}

/*
 * make zero
 */
void
m_mnull(int num,...)
{
        int     i;
        va_list ap;
        struct matrix_t *mp;
        va_start(ap, num);

        for (i=0;i<num;i++)  {
                mp = va_arg(ap, struct matrix_t *);
                m_null(mp);
        }
        va_end(ap);
        return;
}

/*
 * copy but does not copy its contents.
 */

struct matrix_t *
m_puredup(const struct matrix_t *mp)
{
	struct matrix_t *retm;
	char	name[256];

	m_check(__F, 1, mp);

	sprintf(name, "puredup'ed of <%s>", mp->mt_name);
	retm = matrix(mp->mt_row, mp->mt_col, name, mp->mt_type);

	return retm;
}

/*
 * copy
 */
struct matrix_t *
m_dup(const struct matrix_t *mp)
{
	struct matrix_t *retm;

	m_check(__F, 1, mp);
	retm = m_puredup(mp);
	m_copy(retm, mp);

	return retm;
}

void
m_print(const struct matrix_t *af, const char *patt, const char *term)
{
	int	i, j;
	char	patts[64], spaces[64]; /* spaces is not used */

	float	**sdata;
	double	**ddata;
	struct simplex_t **cdata;
	struct duplex_t **zdata;

	switch (af->mt_type) {
		case SP_SINGLE:
		case SP_DOUBLE:
			patt_rgetpatt(patts, spaces, patt, term);
			break;
		case SP_SIMPLEX:
		case SP_DUPLEX:
			patt_cgetpatt(patts, spaces, patt, term);
			break;
		default:
			fatal("%s(): at line %d", __F, __LINE__);
	}

	printf("matrix[%s]------\n", af->mt_name);
	switch (af->mt_type) {
		case SP_SINGLE:
			sdata = smat(af);
			for (i=0; i< af->mt_row; i++) {
				for (j=0; j< af->mt_col; j++)
					if (fabs(sdata[i][j])>1.0e-5)
						printf(patts, sdata[i][j]);
					else
						printf(spaces);
				printf("\n");
			}
			break;
		case SP_DOUBLE:
			ddata = dmat(af);
			for (i=0; i< af->mt_row; i++) {
				for (j=0; j< af->mt_col; j++)
					printf(patts, ddata[i][j]);
				printf("\n");
			}
			break;
		case SP_SIMPLEX:
			cdata = cmat(af);
			for (i=0; i< af->mt_row; i++) {
				for (j=0; j< af->mt_col; j++)
					printf(patts, cdata[i][j].real,
						cdata[i][j].imag);
				printf("\n");
			}
			break;
		case SP_DUPLEX:
			zdata = zmat(af);
			for (i=0; i< af->mt_row; i++) {
				for (j=0; j< af->mt_col; j++)
					printf(patts, zdata[i][j].real,
						zdata[i][j].imag);
				printf("\n");
			}
			break;
		default:
			fatal("%s(): IMPOSSIBLE type", __F);
	}
	printf("++++++++\n");
	return;
}

void
m_add(struct matrix_t *c, const struct matrix_t *a,
	const struct matrix_t *b)
{
	float	**sa, **sb, **sc;
	double	**da, **db, **dc;
	struct simplex_t **ca, **cb, **cc;
	struct duplex_t **za, **zb, **zc;
	int	i, j;

	m_check(__F, 1, a);
	m_check(__F, 2, b);
	m_check(__F, 3, c);

	m_2samesizecheck(__F, a, b);
	m_2samesizecheck(__F, a, c);
	m_2sametypecheck(__F, a, b);
	m_2sametypecheck(__F, a, c);

	switch (c->mt_type) {
		case SP_SINGLE:
			sa = smat(a); sb = smat(b); sc = smat(c);
			for (i=0; i< a->mt_row; i++)
				for (j=0; j< b->mt_col; j++)
					sc[i][j] = sa[i][j] + sb[i][j];
			break;
		case SP_DOUBLE:
			da = dmat(a); db = dmat(b); dc = dmat(c);
			for (i=0; i< a->mt_row; i++)
				for (j=0; j< b->mt_col; j++)
					dc[i][j] = da[i][j] + db[i][j];
			break;
		case SP_SIMPLEX:
			ca = cmat(a); cb = cmat(b); cc = cmat(c);
			for (i=0; i< a->mt_row; i++)
				for (j=0; j< b->mt_col; j++)
					cp_plus(&cc[i][j],
						&ca[i][j], &cb[i][j]);
		case SP_DUPLEX:
			za = zmat(a); zb = zmat(b); zc = zmat(c);
			for (i=0; i< a->mt_row; i++)
				for (j=0; j< b->mt_col; j++)
					zp_plus(&zc[i][j],
						&za[i][j], &zb[i][j]);
			break;
	}
	return;
}

void
m_dot(struct matrix_t *c, const struct matrix_t *a, const struct matrix_t *b)
{
	int	i, j, k;
	float	**sa, **sb, **sc;
	double	**da, **db, **dc;
	struct simplex_t **ca, **cb, **cc, ctmp;
	struct duplex_t **za, **zb, **zc, ztmp;

	m_check(__F, 3, a);
	m_check(__F, 2, b);
	m_check(__F, 1, c);

	if (c->mt_row != a->mt_row || c->mt_col != c->mt_col
		|| a->mt_col != b->mt_row)
		fatal("%s(): [0]", __F);
	m_2sametypecheck(__F, a, b);
	m_2sametypecheck(__F, a, c);

	m_null(c);
	switch (c->mt_type) {
		case SP_SINGLE:
			sa = smat(a); sb = smat(b); sc = smat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_col; k++) 
						sc[i][j] += sa[i][k] * sb[k][j];
			break;
		case SP_DOUBLE:
			da = dmat(a); db = dmat(b); dc = dmat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_col; k++) 
						dc[i][j] += da[i][k] * db[k][j];
			break;
		case SP_SIMPLEX:
			ca = cmat(a); cb = cmat(b); cc = cmat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_col; k++) 
						cp_pplus(&cc[i][j], cp_cross(
						&ctmp, &ca[i][k], &cb[k][j]));
			break;
		case SP_DUPLEX:
			za = zmat(a); zb = zmat(b); zc = zmat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_col; k++) 
						zp_pplus(&zc[i][j], zp_cross(
						&ztmp, &za[i][k], &zb[k][j]));
			break;
	}
	return;
}

/*
 * C=A^T B
 */
void
m_tdot(struct matrix_t *c, const struct matrix_t *a, const struct matrix_t *b)
{
	int	i, j, k;
	float	**sa, **sb, **sc;
	double	**da, **db, **dc;
	struct simplex_t **ca, **cb, **cc, ctmp;
	struct duplex_t **za, **zb, **zc, ztmp;

	m_check(__F, 1, a);
	m_check(__F, 2, b);
	m_check(__F, 3, c);

	if (c->mt_row != a->mt_col || c->mt_col != b->mt_col ||
		a->mt_row != b->mt_row)
		fatal("%s(): [0]", __F);
	m_2sametypecheck(__F, a, b);
	m_2sametypecheck(__F, a, c);

	m_null(c);
	switch (c->mt_type) {
		case SP_SINGLE:
			sa = smat(a); sb = smat(b); sc = smat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_row; k++) 
						sc[i][j] += sa[k][i] * sb[k][j];
			break;
		case SP_DOUBLE:
			da = dmat(a); db = dmat(b); dc = dmat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_row; k++) 
						dc[i][j] += da[k][i] * db[k][j];
			break;
		case SP_SIMPLEX:
			ca = cmat(a); cb = cmat(b); cc = cmat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_row; k++) 
						cp_pplus(&cc[i][j], cp_cross(
						&ctmp, &ca[k][i], &cb[k][j]));
			break;
		case SP_DUPLEX:
			za = zmat(a); zb = zmat(b); zc = zmat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_row; k++) 
						zp_pplus(&zc[i][j], zp_cross(
						&ztmp, &za[k][i], &zb[k][j]));
			break;
	}
	return;
}

/*
 * C=AB^T
 */
void
m_dott(struct matrix_t *c, const struct matrix_t *a, const struct matrix_t *b)
{
	int	i, j, k;
	float	**sa, **sb, **sc;
	double	**da, **db, **dc;
	struct simplex_t **ca, **cb, **cc, ctmp;
	struct duplex_t **za, **zb, **zc, ztmp;

	m_check(__F, 1, c);
	m_check(__F, 2, a);
	m_check(__F, 3, b);

	if (c->mt_row != a->mt_row || c->mt_col != b->mt_row ||
		a->mt_col != b->mt_col)
		fatal("%s(): [0]", __F);
	m_2sametypecheck(__F, a, b);
	m_2sametypecheck(__F, a, c);

	m_null(c);
	switch (c->mt_type) {
		case SP_SINGLE:
			sa = smat(a); sb = smat(b); sc = smat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_col; k++) 
						sc[i][j] += sa[i][k] * sb[j][k];
			break;
		case SP_DOUBLE:
			da = dmat(a); db = dmat(b); dc = dmat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_col; k++) 
						dc[i][j] += da[i][k] * db[j][k];
			break;
		case SP_SIMPLEX:
			ca = cmat(a); cb = cmat(b); cc = cmat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_col; k++) 
						cp_pplus(&cc[i][j], cp_cross(
						&ctmp, &ca[i][k], &cb[j][k]));
			break;
		case SP_DUPLEX:
			za = zmat(a); zb = zmat(b); zc = zmat(c);
			for (i=0; i< c->mt_row; i++)
				for (j=0; j< c->mt_col; j++)
					for (k=0; k < a->mt_col; k++) 
						zp_pplus(&zc[i][j], zp_cross(
						&ztmp, &za[i][k], &zb[j][k]));
			break;
	}
	return;
}

void
m_copy(struct matrix_t *target, const struct matrix_t *bf)
{
	int	size;
	m_check(__F, 1, target);
	m_check(__F, 1, bf);
	if (bf == target)
		fatal("%s(): you can not store the inverse in itself!!", __F);

	m_2samesizecheck(__F, target, bf);
	m_2sametypecheck(__F, target, bf);


	switch (bf->mt_type) {
		case SP_SINGLE:
			size = sizeof(float) * bf->mt_row * bf->mt_col;
			break;
		case SP_DOUBLE:
			size = sizeof(double) * bf->mt_row * bf->mt_col;
			break;
		case SP_SIMPLEX:
			size = sizeof(struct simplex_t)*bf->mt_row * bf->mt_col;
			break;
		case SP_DUPLEX:
			size = sizeof(struct duplex_t)*bf->mt_row * bf->mt_col;
			break;
		default:
			fatal("%s(): IMPOSSIBLE type", __F);
	}
	memcpy(target->mt_data[0], bf->mt_data[0], size);
	return;
}


void
m_inverse(struct matrix_t *target, const struct matrix_t *bf)
{
	m_check(__F, 1, target);
	m_check(__F, 2, bf);
	if (bf == target)
		fatal("%s(): you can not store the inverse in itself!!", __F);

	m_2samesizecheck(__F, target, bf);
	m_2sametypecheck(__F, target, bf);

	switch (bf->mt_type) {
		case SP_SINGLE: sm_inverse(target, bf); break;
		case SP_DOUBLE: dm_inverse(target, bf); break;
		default:
			fatal("%s(): not implemented yet", __F);
	}
	return;
}


int
m_issymmet(const struct matrix_t *af)
{
	int	ret;
	switch (af->mt_type) {
		case SP_SINGLE:  ret = sm_issymmet(af); break;
		case SP_DOUBLE:  ret = dm_issymmet(af); break;
		case SP_SIMPLEX: ret = cm_issymmet(af); break;
		case SP_DUPLEX:  ret = zm_issymmet(af); break;
		default: fatal("%s(): IMPOSSIBLE type", __F);
	}
	return ret;
}


void
m_briefinfo(const struct matrix_t *af)
{
	printf("matrix name  = %s\n", af->mt_name);
	printf("matrix rowno = %d\n", af->mt_row);
	printf("matrix colno = %d\n", af->mt_col);

	return;
}

#define TOL 1.0e-6

int
m_nonzeros(const struct matrix_t *af)
{
	int	cnt;
	int	i, j;
	float	**spp;
	double	**dpp;
	struct simplex_t **cpp;
	struct duplex_t **zpp;

	m_check(__F, 1, af);
	cnt = 0;
	switch (af->mt_type) {
		case SP_SINGLE:
			spp = smat(af);
			for (i=0;i<af->mt_row;i++)
				for (j=0;j<af->mt_row;j++)
					if (fabs(spp[i][j]) > TOL)
						cnt++;
			break;
		case SP_DOUBLE:
			dpp = dmat(af);
			for (i=0;i<af->mt_row;i++)
				for (j=0;j<af->mt_row;j++)
					if (fabs(dpp[i][j]) > TOL)
						cnt++;
			break;
		case SP_SIMPLEX:
			cpp = cmat(af);
			for (i=0;i<af->mt_row;i++)
				for (j=0;j<af->mt_row;j++)
					if (cp_absolute(&cpp[i][j]) > TOL)
						cnt++;
			break;
		case SP_DUPLEX:
			zpp = zmat(af);
			for (i=0;i<af->mt_row;i++)
				for (j=0;j<af->mt_row;j++)
					if (zp_absolute(&zpp[i][j]) > TOL)
						cnt++;
			break;
		default:
			fatal("%s(): Unknown type", __F);
	}
	return cnt;
}

void
m_pattern(const struct matrix_t *af)
{
	int	i, j;
	float	**spp;
	double	**dpp;
	struct simplex_t **cpp;
	struct duplex_t **zpp;
	char *head, *p;

	m_check(__F, 1, af);
	head = (char *) malloc((af->mt_col * sizeof(char))+1);
	printf("-------------------------\n");
	switch (af->mt_type) {
		case SP_SINGLE:
			spp = smat(af);
			for (i=0;i<af->mt_row;i++) {
				p = head;
				for (j=0;j<af->mt_col;j++, p++)
					if (fabs(spp[i][j]) > TOL)
						*p = '+';
					else
						*p = '.';
				*p = 0;
				printf("%s\n", head);
			}
			break;
		case SP_DOUBLE:
			dpp = dmat(af);
			for (i=0;i<af->mt_row;i++) {
				p = head;
				for (j=0;j<af->mt_col;j++, p++)
					if (fabs(dpp[i][j]) > TOL)
						*p = '+';
					else
						*p = '.';
				*p = 0;
				printf("%s\n", head);
			}
			break;
		case SP_SIMPLEX:
			cpp = cmat(af);
			for (i=0;i<af->mt_row;i++) {
				p = head;
				for (j=0;j<af->mt_col;j++, p++)
					if (cp_absolute(&cpp[i][j]) > TOL)
						*p = '+';
					else
						*p = '.';
				*p = 0;
				printf("%s\n", head);
			}
			break;
		case SP_DUPLEX:
			zpp = zmat(af);
			for (i=0;i<af->mt_row;i++) {
				p = head;
				for (j=0;j<af->mt_col;j++, p++)
					if (zp_absolute(&zpp[i][j]) > TOL)
						*p = '+';
					else
						*p = '.';
				*p = 0;
				printf("%s\n", head);
			}
			break;
		default:
			fatal("%s(): Unknown type", __F);
	}
	free(head);
	return;
}

void
m_differat(const struct matrix_t *af, const struct matrix_t *bf,
	const int i, const int j)
{
	printf("matrices <%s> and <%s> differ at [%d][%d]\n",
		af->mt_name, bf->mt_name, i, j);
	return;
}

/*
 * returns the difference per entry
 */
double
m_compare(const struct matrix_t *af, const struct matrix_t *bf)
{
	int	i, j;
	float		 **sma, **smb;
	double		 **dma, **dmb;
	struct simplex_t **cma, **cmb, ctmp;
	struct duplex_t  **zma, **zmb, ztmp;
	double differ;

	m_check(__F, 1, af);
	m_check(__F, 1, bf);
	m_2samesizecheck(__F, af, bf);
	m_2sametypecheck(__F, af, bf);

	differ = 0.0;
	switch (af->mt_type) {
		case SP_SINGLE:
			sma = smat(af); smb = smat(bf);
			for (i=0;i<af->mt_row;i++) {
				for (j=0;j<af->mt_col;j++)
					differ += fabs(sma[i][j]-smb[i][j]);
			}
			break;
		case SP_DOUBLE:
			dma = dmat(af); dmb = dmat(bf);
			for (i=0;i<af->mt_row;i++) {
				for (j=0;j<af->mt_col;j++)
					differ += fabs(dma[i][j]-dmb[i][j]);
			}
			break;
		case SP_SIMPLEX:
			cma = cmat(af); cmb = cmat(bf);
			for (i=0;i<af->mt_row;i++) {
				for (j=0;j<af->mt_col;j++)
					differ += cp_absolute(cp_minus(&ctmp,
					    &cma[i][j], &cmb[i][j]));
			}
			break;
		case SP_DUPLEX:
			zma = zmat(af); zmb = zmat(bf);
			for (i=0;i<af->mt_row;i++) {
				for (j=0;j<af->mt_col;j++)
					differ += zp_absolute(zp_minus(&ztmp,
					    &zma[i][j], &zmb[i][j]));
			}
			break;
		default:
			fatal("%s(): Unknown type", __F);
	}
	differ /= (double) (af->mt_row * af->mt_col);
	return differ;
}


/*
 * yp = af * xp;
 */
struct vector_t *
m_dotvec(struct vector_t *yptr,
	const struct matrix_t *af, const struct vector_t *xp)
{
	int	i, j;
	float		 *sv, **sam, *sxp;
	double		 *dv, **dam, *dxp;
	struct simplex_t *cv, **cam, *cxp, ctmp;
	struct duplex_t  *zv, **zam, *zxp, ztmp;

	m_check(__F, 1, af);
	vec_check(__F, yptr, -1);
	vec_check(__F, xp, -1);

	if (af->mt_col != xp->vt_size)
		fatal("%s(): [0]", __F);
	if (yptr->vt_size != af->mt_row)
		fatal("%s(): [1]", __F);
	if (yptr->vt_type != af->mt_type || af->mt_type != xp->vt_type)
		fatal("%s(): [2]", __F);

	vec_null(yptr);
	switch (yptr->vt_type) {
		case SP_SINGLE:
			sv  = svec(yptr); sxp = svec(xp); sam = smat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					sv[i] += (sam[i][j] * sxp[j]);
			}
			break;
		case SP_DOUBLE:
			dv  = dvec(yptr); dxp = dvec(xp); dam = dmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					dv[i] += (dam[i][j] * dxp[j]);
			}
			break;
		case SP_SIMPLEX:
			cv  = cvec(yptr); cxp = cvec(xp); cam = cmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					cp_pplus(&cv[i], cp_cross(&ctmp,
						&cam[i][j], &cxp[j]));
			}
			break;
		case SP_DUPLEX:
			zv  = zvec(yptr); zxp = zvec(xp); zam = zmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					zp_pplus(&zv[i], zp_cross(&ztmp,
						&zam[i][j], &zxp[j]));
			}
			break;
		default:
			fatal("%s(): Unknown type", __F);
	}
	return yptr;
}

/*
 * yp = af^T * xp;
 */
struct vector_t *
m_tdotvec(struct vector_t *yptr,
	const struct matrix_t *af, const struct vector_t *xp)
{
	int	i, j;
	float		 *sv, **sam, *sxp;
	double		 *dv, **dam, *dxp;
	struct simplex_t *cv, **cam, *cxp, ctmp;
	struct duplex_t  *zv, **zam, *zxp, ztmp;

	m_check(__F, 1, af);
	vec_check(__F, yptr, -1);
	vec_check(__F, xp, -1);

	if (af->mt_col != xp->vt_size)
		fatal("%s(): [0]", __F);
	if (yptr->vt_size != af->mt_row)
		fatal("%s(): [1]", __F);
	if (yptr->vt_type != af->mt_type || af->mt_type != xp->vt_type)
		fatal("%s(): [2]", __F);

	vec_null(yptr);
	switch (yptr->vt_type) {
		case SP_SINGLE:
			sv  = svec(yptr); sxp = svec(xp); sam = smat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					sv[i] += (sam[j][i] * sxp[j]);
			}
			break;
		case SP_DOUBLE:
			dv  = dvec(yptr); dxp = dvec(xp); dam = dmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					dv[i] += (dam[j][i] * dxp[j]);
			}
			break;
		case SP_SIMPLEX:
			cv  = cvec(yptr); cxp = cvec(xp); cam = cmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					cp_pplus(&cv[i], cp_cross(&ctmp,
						&cam[j][i], &cxp[j]));
			}
			break;
		case SP_DUPLEX:
			zv  = zvec(yptr); zxp = zvec(xp); zam = zmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					zp_pplus(&zv[i], zp_cross(&ztmp,
						&zam[j][i], &zxp[j]));
			}
			break;
		default:
			fatal("%s(): Unknown type", __F);
	}
	return yptr;
}


/*
 * yp = xp *af^T;
 */
struct vector_t *
m_tvecdot(struct vector_t *yptr, const struct vector_t *xp,
	const struct matrix_t *af)
{
	int	i, j;
	float		 *sv, **sam, *sxp;
	double		 *dv, **dam, *dxp;
	struct simplex_t *cv, **cam, *cxp, ctmp;
	struct duplex_t  *zv, **zam, *zxp, ztmp;

	m_check(__F, 1, af);
	vec_check(__F, yptr, -1);
	vec_check(__F, xp, -1);

	if (af->mt_row != yptr->vt_size)
		fatal("%s(): [0]", __F);
	if (xp->vt_size != af->mt_col)
		fatal("%s(): [1]", __F);
	if (yptr->vt_type != af->mt_type || af->mt_type != xp->vt_type)
		fatal("%s(): [2]", __F);

	vec_null(yptr);
	switch (yptr->vt_type) {
		case SP_SINGLE:
			sv  = svec(yptr); sxp = svec(xp); sam = smat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					sv[i] += (sxp[j] * sam[i][j]);
			}
			break;
		case SP_DOUBLE:
			dv  = dvec(yptr); dxp = dvec(xp); dam = dmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					dv[i] += (dxp[j] * dam[i][j]);
			}
			break;
		case SP_SIMPLEX:
			cv  = cvec(yptr); cxp = cvec(xp); cam = cmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					cp_pplus(&cv[i], cp_cross(&ctmp,
						&cxp[j], &cam[i][j]));
			}
			break;
		case SP_DUPLEX:
			zv  = zvec(yptr); zxp = zvec(xp); zam = zmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					zp_pplus(&zv[i], zp_cross(&ztmp,
						&zxp[j], &zam[i][j]));
			}
			break;
		default:
			fatal("%s(): Unknown type", __F);
	}
	return yptr;
}
/*
 * yp = xp *af;
 */
struct vector_t *
m_vecdot(struct vector_t *yptr, const struct vector_t *xp,
	const struct matrix_t *af)
{
	int	i, j;
	float		 *sv, **sam, *sxp;
	double		 *dv, **dam, *dxp;
	struct simplex_t *cv, **cam, *cxp, ctmp;
	struct duplex_t  *zv, **zam, *zxp, ztmp;

	m_check(__F, 1, af);
	vec_check(__F, yptr, -1);
	vec_check(__F, xp, -1);

	if (af->mt_col != yptr->vt_size)
		fatal("%s(): [0]", __F);
	if (xp->vt_size != af->mt_row)
		fatal("%s(): [1]", __F);
	if (yptr->vt_type != af->mt_type || af->mt_type != xp->vt_type)
		fatal("%s(): [2]", __F);

	vec_null(yptr);
	switch (yptr->vt_type) {
		case SP_SINGLE:
			sv  = svec(yptr); sxp = svec(xp); sam = smat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					sv[i] += (sxp[j] * sam[j][i]);
			}
			break;
		case SP_DOUBLE:
			dv  = dvec(yptr); dxp = dvec(xp); dam = dmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					dv[i] += (dxp[j] * dam[j][i]);
			}
			break;
		case SP_SIMPLEX:
			cv  = cvec(yptr); cxp = cvec(xp); cam = cmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					cp_pplus(&cv[i], cp_cross(&ctmp,
						&cxp[j], &cam[j][i]));
			}
			break;
		case SP_DUPLEX:
			zv  = zvec(yptr); zxp = zvec(xp); zam = zmat(af);
			for (i=0;i< yptr->vt_size;i++) {
				for (j=0;j< yptr->vt_size;j++)
					zp_pplus(&zv[i], zp_cross(&ztmp,
						&zxp[j], &zam[j][i]));
			}
			break;
		default:
			fatal("%s(): Unknown type", __F);
	}
	return yptr;
}

struct matrix_t *
m_mkdiagmat(const struct vector_t *a)
{
	struct matrix_t *retm;
	int	i;
	float		 *sv, **sam;
	double		 *dv, **dam;
	struct simplex_t *cv, **cam;
	struct duplex_t  *zv, **zam;

	vec_check(__F, a, -1);
	retm = matrix(a->vt_size, a->vt_size, NULL, a->vt_type);

	switch (a->vt_type) {
		case SP_SINGLE:
			sv  = svec(a); sam = smat(retm);
			for (i=0;i < a->vt_size;i++)
				sam[i][i] = sv[i];
			break;
		case SP_DOUBLE:
			dv  = dvec(a); dam = dmat(retm);
			for (i=0;i < a->vt_size;i++)
				dam[i][i] = dv[i];
			break;
		case SP_SIMPLEX:
			cv  = cvec(a); cam = cmat(retm);
			for (i=0;i < a->vt_size;i++)
				cp_assign(&cam[i][i], &cv[i]);
			break;
		case SP_DUPLEX:
			zv  = zvec(a); zam = zmat(retm);
			for (i=0;i < a->vt_size;i++)
				zp_assign(&zam[i][i], &zv[i]);
			break;
	}
	return retm;
}

/*
 * Currently only used at Singular value decomposition
 * only inverse the diagonal part
 */
void
m_mkdiaginv(const struct matrix_t *a)
{
	int	i;
	float		 **sam;
	double		 **dam;
	struct simplex_t **cam;
	struct duplex_t  **zam;

	m_check(__F, 1, a);
	if (a->mt_row != a->mt_col)
		fatal("%s(): [0]", __F);

	switch (a->mt_type) {
		case SP_SINGLE: sam = smat(a);
			for (i=0;i < a->mt_row;i++)
				sam[i][i] = 1.0/sam[i][i];
			break;
		case SP_DOUBLE: dam = dmat(a);
			for (i=0;i < a->mt_row;i++)
				dam[i][i] = 1.0/dam[i][i];
			break;
		case SP_SIMPLEX: cam = cmat(a);
			for (i=0;i < a->mt_row;i++) /* XXX */
				cp_inverse(&cam[i][i], &cam[i][i]);
			break;
		case SP_DUPLEX: zam = zmat(a);
			for (i=0;i < a->mt_row;i++) /* XXX */
				zp_inverse(&zam[i][i], &zam[i][i]);
			break;
	}
	return;
}

void
m_put(const struct matrix_t *a)
{
	char line[]="%6.3f";
	m_print(a, line, "+");
	return;
}

void
m_mput(int num,...)
{
	int     i;
	va_list ap;
	struct matrix_t *p;
	va_start(ap, num);

	for (i=0;i<num;i++)  {
		p = va_arg(ap, struct matrix_t *);
		m_put(p);
	}
	va_end(ap);
	return;
}

void
m_2samesizecheck(const char *func,
	const struct matrix_t *ap, const struct matrix_t *bp)
{
	if (ap->mt_col == bp->mt_col && ap->mt_row == bp->mt_row)
		return;
	fatal("%s(): %s and %s have diff. size", __F,
		ap->mt_name, bp->mt_name);
}

void
m_2sametypecheck(const char *func,
	const struct matrix_t *ap, const struct matrix_t *bp)
{
	if (ap->mt_type == bp->mt_type)
		return;

	fatal("%s(): %s and %s have diff. type", __F,
		ap->mt_name, bp->mt_name);
}
void
m_3sametypecheck(const char *func, const struct matrix_t *ap,
	const struct matrix_t *bp, const struct matrix_t *cp)
{
	if (ap->mt_type != bp->mt_type)
		fatal("%s(): %s and %s have diff. type", __F,
			ap->mt_name, bp->mt_name);
	if (ap->mt_type != cp->mt_type)
		fatal("%s(): %s and %s have diff. type", __F,
			ap->mt_name, cp->mt_name);
	return;

}

void
m_squarecheck(const char *func, const struct matrix_t *mp)
{
	if (mp->mt_col == mp->mt_row)
		return;
	fatal("%s(): %s is not squar matrix", __F, mp->mt_name);
}

void
m_typecheck(const char *func, const int arg, const struct matrix_t *mp,
	const int type)
{
	char	*p;

	p = get_ordinal(arg);
	if (mp == NULL)
		fatal("%s argument of %s() was NULL pointer!!\n", p, func);

	if (!IS_MATRIX(mp))
		fatal("Even though %s argument of %s() was not NULL pointer,\n"
		"it was not of matrix_t type!!\n", p, func);
	if (mp->mt_type != type)
		fatal("%s, the %s arg of %s() was not matrix of type %s!!\n",
			mp->mt_name, p, func, typenames[type]);
	return;
}

void
m_check(const char *func, const int arg, const struct matrix_t *mp)
{
	char	*p;

	p = get_ordinal(arg);
	if (mp == NULL)
		fatal("%s argument of %s() was NULL pointer!!\n", p, func);

	if (!IS_MATRIX(mp))
		fatal("Even though %s argument of %s() was not NULL pointer,\n"
		"it was not of matrix_t type!!\n", p, func);
	return;
}
/* 
 * returns 0 if ok
 */
int
m_sizecheck(const struct matrix_t *mp, int row, int col)
{
	m_check(__F, 1, mp);
	if (mp->mt_row != row || mp->mt_col != col)
		return 1;
	else
		return 0;
}
