#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include "common.h"
#include "sparse.h"




/*
 * XXX-------IO routines
 */
struct d_before {
	int     b_cnt;
	int     b_row;
	int     b_col;
	float  b_data;
};
int sp_srowcolcmp(struct d_before **first, struct d_before **second);

struct d_cols {
        int     col_colno;
        short   col_norows;
        short   col_alloced;
        struct d_before **col_datas;
};

struct sparse_t *
sp_sread(const char *fname)
{
	FILE	*fp;
	int	row, col, nonzero;
	char	line[BUFSIZ], name[32];
	int	type;
	int	tmp;
        struct d_before **root, *befp;
        struct d_cols     *colsp;
        int     (*compare)() = sp_srowcolcmp;
	struct d_cols **allcols;

        size_t  entries;
        int     i, j;
        int     cnt;
        struct sparse_t *sp;
        int     *irow, *icol;
        float  *dp;
        float   fdata;


	if (fname == NULL || *fname == 0x00)
		fatal("%s(): Wrong fname", __F);
	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("%s(): Can't fopen(%s)", __F, fname);

	fgets(line, BUFSIZ, fp);
	tmp = sscanf(line, "%d %d %d %s", &row, &col, &nonzero, name);
	if (tmp != 4)
		fatal("%s(): while reading header of %s", __F, fname);

	type = SP_SINGLE;
	sp = sparse(row, col, nonzero, fname, type);

	root = (struct d_before **)
		malloc((size_t)(sizeof(struct d_before *) * nonzero));
	if (root == NULL)
		memerr("%s(%s): allocing **root", __F, fname);

	befp = (struct d_before *)
		malloc((size_t)(sizeof(struct d_before) * nonzero));
	if (befp == NULL)
		memerr("%s(%s): allocing *befp", __F, fname);

	for (i=0; i< nonzero; i++, befp++)
		root[i] = befp;

	allcols = (struct d_cols **)
		malloc((size_t) (sizeof(struct d_cols *) * col));
	if (allcols == NULL)
		memerr("%s(%s): allocing **allcols", __F, fname);
	colsp = (struct d_cols *)
		malloc((size_t) (sizeof(struct d_cols) * col));
	if (colsp == NULL)
		memerr("%s(%s): allocing *colsp", __F, fname);

	for (i=0;i<col;i++) {
		allcols[i] = colsp++;
		allcols[i]->col_colno = (i+1);
		allcols[i]->col_norows = 0;
		allcols[i]->col_alloced = 10;
		allcols[i]->col_datas = (struct d_before **)
			malloc((size_t)(sizeof(struct d_before *) * 10));
	}
		
	befp = root[0];
	entries = 1;
	while (fgets(line, BUFSIZ, fp)) {
		tmp = sscanf(line, "%d %d %f",
			&befp->b_row, &befp->b_col, &fdata);
		if (tmp != 3)
			fatal("%s(%s): reading error at %d", __F,
				sp->sp_name, (entries+1));
		if (befp->b_row == 0 && befp->b_col == 0)
			break;
		if (befp->b_row > row || befp->b_col > col)
			fatal("%s(%s):bad index(%d,%d) in at line %d", __F,
				fname, befp->b_row, befp->b_col, (entries+1));

		colsp = allcols[(befp->b_col-1)];
		/*
		 * OOPS!! Following line should be >= not >
		 */
		if (colsp->col_norows >= colsp->col_alloced) {
			colsp->col_datas = (struct d_before **) realloc(
				(void *) colsp->col_datas,
				(size_t) (sizeof(struct d_before *)
					* (colsp->col_alloced + 10)));
			colsp->col_alloced += 10;
		}
		colsp->col_datas[colsp->col_norows] = befp;
		colsp->col_norows++;

		befp->b_data = fdata; 
		befp->b_cnt = entries++;
		befp++;
	}
	fclose(fp);
	entries--;
	if (entries != nonzero)
		fatal("%s(%s): entries nonzero no %d is not correct", __F,
			fname, nonzero);

	for (i=0; i< col;i++) {
		colsp = allcols[i];
		qsort((int *) colsp->col_datas, colsp->col_norows,
			(size_t) sizeof(struct d_before *), compare);
	}
	entries = 0;
	for (i=0; i< col;i++) {
		colsp = allcols[i];
		entries += colsp->col_norows;
	}
	/*
	 * Now I want to make sparse_t matrix
	 */

	icol = sp->sp_colx;
	irow = sp->sp_rowx;
	dp = (float *) sp->sp_data;

	cnt = 0;
	for (i=0; i<sp->sp_coln;i++) {
		colsp = allcols[i];
		*icol++ = cnt;
		for (j=0;j<colsp->col_norows;j++) {
			befp = colsp->col_datas[j];
			*irow++ = (befp->b_row - 1); 	/* Zero index */
			*dp++   = (befp->b_data);
			cnt++;	/* XXX */
		}
		free(colsp->col_datas);
	}
	*icol = cnt;	/* Very Easy to ommit this line */
	free(allcols[0]);
	free(allcols);

	free(root[0]);
	free(root);

	return  sp;
}


struct sparse_t *
sp_sreadraw(const char *fname)
{
	FILE	*fp;
	int	fd;
	char	buf[BUFSIZ], typenm[32], *p;
	int	readn, newlinecnt;
	struct d_before **root, *befp;
	struct d_cols	*colsp;
	int	(*compare)() = sp_srowcolcmp;
	struct d_cols **allcols;

	int	matsize;
	int	entries;
	int	i, j;
	int	cnt;
	struct sparse_t *sp;
	int	*irow, *icol;
	float	*dp;
	float	fdata;

	if (fname == NULL || *fname == 0)
		fatal("%s(): You should provide good name!!", __F);
	fd = open(fname, O_RDONLY);
	if (fd < 0)
		fatal("%s(): at fopen(%s)", __F, fname);

	/*
	 * To order provide information of malloc, I have to scan the
	 * input file to get the approximate number of nonzero entries
	 */

	newlinecnt = 0;
	while ((readn = read(fd, (void *) buf, (unsigned) BUFSIZ))) {
		for (p = (char *) buf, i=0; i < readn;i++)
			if (*p++ == '\n')
				newlinecnt++;
	}
	close(fd);

	root = (struct d_before **)
		malloc((size_t)(sizeof(struct d_before *) * newlinecnt));
	if (root == NULL)
		memerr("%s(): allocing **root", __F);

	befp = (struct d_before *)
		malloc((size_t)(sizeof(struct d_before) * newlinecnt));
	if (befp == NULL)
		memerr("%s(): allocing *befp", __F);

	for (i=0; i< newlinecnt; i++, befp++)
		root[i] = befp;

	/*
	 * real reading
	 */
	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("%s(): You should provide good name!!", __F);

	befp = root[0];
	while (fgets(buf, BUFSIZ, fp)) {
		cnt = sscanf(buf, "%d %s", &matsize, typenm);
		if (cnt == 2)
			break;
	}

	allcols = (struct d_cols **)
		malloc((size_t) (sizeof(struct d_cols *) * matsize));
	if (allcols == NULL)
		memerr("%s(): allocing **allcols", __F);
	colsp = (struct d_cols *)
		malloc((size_t) (sizeof(struct d_cols) * matsize));
	if (colsp == NULL)
		memerr("%s(): allocing *colsp", __F);

	for (i=0;i<matsize;i++) {
		allcols[i] = colsp++;
		allcols[i]->col_colno = (i+1);
		allcols[i]->col_norows = 0;
		allcols[i]->col_alloced = 10;
		allcols[i]->col_datas = (struct d_before **)
			malloc((size_t)(sizeof(struct d_before *) * 10));
	}
		
	entries = 1;
	while (fgets(buf, BUFSIZ, fp)) {
		readn = sscanf(buf, "%d %d %f",
			&befp->b_row, &befp->b_col, &fdata);
		if (readn - 3)
			fatal("%s(%s): at line %d", __F, fname,(entries+1));

		if (befp->b_row == 0 && befp->b_col == 0)
			break;
		if (befp->b_row > matsize || befp->b_col > matsize)
			fatal("%s(): wrong index %d row %d col at %dth line",
				__F, befp->b_row, befp->b_col, (entries+1));

		colsp = allcols[(befp->b_col-1)];
		/*
		 * OOPS!! Following line should be >= not >
		 */
		if (colsp->col_norows >= colsp->col_alloced) {
			colsp->col_datas = (struct d_before **) realloc(
				(void *) colsp->col_datas,
				(size_t) (sizeof(struct d_before *)
					* (colsp->col_alloced + 10)));
			colsp->col_alloced += 10;
		}
		colsp->col_datas[colsp->col_norows] = befp;
		colsp->col_norows++;

		befp->b_data = fdata; 
		befp->b_cnt = entries++;
		befp++;
	}
	fclose(fp);
	entries--;

	for (i=0; i< matsize;i++) {
		colsp = allcols[i];
		qsort((int *) colsp->col_datas, colsp->col_norows,
			(size_t) sizeof(struct d_before *), compare);
	}
	entries = 0;
	for (i=0; i< matsize;i++) {
		colsp = allcols[i];
		entries += colsp->col_norows;
	}
	printf("ENTRIES=%d MATSIZE=%d\n", entries, matsize);

	/*
	 * Now I want to make sparse_t matrix
	 */

	sp = sparse(matsize, matsize, entries,  fname, SP_SINGLE);

	icol = sp->sp_colx;
	irow = sp->sp_rowx;
	dp = (float *) sp->sp_data;

	cnt = 0;
	for (i=0; i<matsize;i++) {
		colsp = allcols[i];
		*icol++ = cnt;
		for (j=0;j<colsp->col_norows;j++) {
			befp = colsp->col_datas[j];
			*irow++ = (befp->b_row - 1); 	/* Zero index */
			*dp++   = (befp->b_data);
			cnt++;	/* XXX */
		}
		free(colsp->col_datas);
	}
	*icol = cnt;	/* Very Easy to ommit this line */
	free(allcols[0]);
	free(allcols);

	free(root[0]);
	free(root);

	return  sp;
}


struct sparse_t *
sp_sadd(const struct sparse_t *one, const struct sparse_t *two)
{
	struct d_cols {
		int	 col_colno;
		int	 col_norows;
		struct detail *col_datas;	/* For sorting */
		int	 col_alloced;
	};
	int	rown, coln;
	struct d_cols **allcols, *colsp;
	int	i, j, k;
	int	start, end;
	struct detail *detp;
	int     (*compare)() = sp_rowcmp;
	struct sparse_t *sp;
	int	totalcnt, cnt;
	int	*ip, *iq;
	float	*dp, *dblp;
	char	name[128];
	int	row, prevrow, onoff, mat;


	if (!IS_SPARSE(one) || !IS_SPARSE(two))
		fatal("%s(): one of inputs is not sparse_t", __F);
	if (one->sp_type != SP_SINGLE || two->sp_type != SP_SINGLE)
		fatal("%s(): One of them(%s, %s) is not SP_SINGLE",
			__F, one->sp_name, two->sp_name);

	coln = (one->sp_coln >= two->sp_coln) ?
		one->sp_coln : two->sp_coln;
	rown = (one->sp_rown >= two->sp_rown) ?
		one->sp_rown : two->sp_rown;

	allcols = (struct d_cols **)
		malloc((size_t) (sizeof(struct d_cols *) * coln));
	if (allcols == NULL)
		memerr("%s(): allocing **allcols", __F);

	colsp = (struct d_cols *)
		malloc((size_t) (sizeof(struct d_cols) * coln));
	if (colsp == NULL)
		memerr("%s(): allocing *colsp", __F);

	for (i=0; i < coln; i++) {
		allcols[i] = colsp++;
		allcols[i]->col_colno = (i+1);
		allcols[i]->col_norows = 0;
		allcols[i]->col_alloced = 20;
		allcols[i]->col_datas = (struct detail *)
			malloc((size_t)(sizeof(struct detail) * 20));
	}


	for (i=0; i < coln; i++) {
		colsp = allcols[i];
		detp = colsp->col_datas;	/* information about location */
		colsp->col_norows = 0;		/* XXX */
		cnt = 0;
		if (i < one->sp_coln) {
			start = one->sp_colx[i];
			end   = one->sp_colx[i+1];

			for (j = start; j < end;j++) { /* needs more mem */
				if (colsp->col_norows >= colsp->col_alloced) {
					colsp->col_datas = (struct detail *)
					    realloc((void *) colsp->col_datas,
					    (size_t)(colsp->col_alloced + 20));
					colsp->col_norows += 20;
				}
				detp[cnt].d_whatrow = one->sp_rowx[j];
				detp[cnt].d_whichmat = 1;
				cnt++;
				colsp->col_norows++;
			}
		}
		if (i < two->sp_coln) {
			start = two->sp_colx[i];
			end   = two->sp_colx[i+1];

			for (j = start; j < end;j++) { /* needs more mem */
				if (colsp->col_norows >= colsp->col_alloced) {
					colsp->col_datas = (struct detail *)
					    realloc((void *) colsp->col_datas,
					    (size_t)(colsp->col_alloced + 20));
					colsp->col_norows += 20;
				}
				detp[cnt].d_whatrow = two->sp_rowx[j];
				detp[cnt].d_whichmat = 2;
				cnt++;
				colsp->col_norows++;
			}
		}
		qsort((int *) colsp->col_datas, cnt,
			(size_t) sizeof(struct detail), compare);
	}
	/*
	 * Let's calculate the number of the net nonzero location
	 */

	totalcnt = 0;
	for (i=0;i < coln; i++) {
		colsp = allcols[i];
		detp = colsp->col_datas;

		prevrow = -1;
		for (j = 0; j<colsp->col_norows;j++) {
			row = detp[j].d_whatrow;
			if (row != prevrow)
				totalcnt++;
			prevrow = row;
		}
	}
	sprintf(name, "sp_sadd of %s and %s",
		one->sp_name, two->sp_name);
	sp = sparse(rown, coln, totalcnt, name, SP_SINGLE);


	ip = sp->sp_colx;
	iq = sp->sp_rowx;
	dp = (float *) sp->sp_data;


	dp--;
	iq--;
	for (i=0, cnt = 0; i < coln; i++) {
		*ip++ = cnt;
		colsp = allcols[i];
		detp = colsp->col_datas;

		for (j=0, onoff = 1;j < colsp->col_norows;
		    j++, prevrow = row, onoff = 0) {
			row = detp[j].d_whatrow;
			mat = detp[j].d_whichmat;

			switch (mat) {
			case 1:
				for (k=one->sp_colx[i];
				    k<one->sp_colx[i+1];k++) {
					if (one->sp_rowx[k] == row) {
						dblp = (float *) one->sp_data
							+ k;
						break;
					}
				}
				if (k == one->sp_colx[i+1])
					fatal("sp_sadd(): [1]");
				break;
			case 2:
				for (k=two->sp_colx[i];
				    k<two->sp_colx[i+1];k++) {
					if (two->sp_rowx[k] == row) {
						dblp = (float *) two->sp_data
							+ k;
						break;
					}
				}
				if (k == two->sp_colx[i+1])
					fatal("sp_sadd(): [2]");
				break;
			default:
				fatal("%s(): [3] for %s", __F, sp->sp_name);
			}


			if (onoff || prevrow != row) {
				dp++;
				iq++;
				cnt++;
			}
			*dp += *dblp;
			*iq = row;
		}
	}
	*ip = cnt;

	return sp;
}



int
sp_srowcolcmp(struct d_before **first, struct d_before **second)
{
	struct d_before *one, *two;

	one = *first;
	two = *second;

	if (one->b_col != two->b_col)
		fatal("%s(): one and two have same column number", __F);

	/* now same col */
	if (one->b_row < two->b_row)
		return -1;
	else if (one->b_row > two->b_row)
		return 1;
	fatal("%s():There are two entries whose (row, col)=(%d %d)!", __F,
		one->b_row, one->b_col);
	/* NOTREACHED */
	return 0;
}



/*
 * private function: only used at sp_scrosssp()
 */
int
s_getproduct(struct sparse_t *spa, int *acolx, int *adatax, int ano,
	struct sparse_t *spb, int *browx, int off, int bno, float *result)
{
	int	cnt[2], gotit;
	float	ret;

	if (ano == 0 || bno == 0)
		fatal("%s(): ano == 0 || bno == 0", __F);
	cnt[0] = cnt[1] = 0;

	ret = 0.0;
	gotit = 0;
	while (cnt[0] < ano && cnt[1] < bno) {
		switch (INTCMP(acolx[cnt[0]], browx[cnt[1]])) {
			case  1:
				cnt[1]++;	/* Hahaha.. you are a loser */
				break;
			case  -1:
				cnt[0]++;	/* I can not match you */
				break;
			case 0:			/* we have equal power */
				ret += *((float *)spa->sp_data
						+ adatax[cnt[0]])
					* *((float *)spb->sp_data +
						off + cnt[1]);
				cnt[0]++;
				cnt[1]++;
				gotit++;
				break;
		}
	}
	if (gotit == 0)
		return 0;
	*result = ret;
	return 1;
}
				

/*
 * should be called from sp_crosssp();
 * mulifly a sparse_t with another sparse_t
 */
struct sparse_t *
sp_scrosssp(struct sparse_t *spa, struct sparse_t *spb)
{
	struct sparse_t *retsp;
	int	i, j, cnt, alloced;
	int	*ncolx, *nrowx;	/* neo col, row, data */
	float	*ndatax, retdata;
	int	*acolx, *adatax, ano, *browx, bno;
	int	exist, size;
	char	name[128];

	if (spa->sp_roworder == NULL)
		sp_mkroworder(spa);

	/*
	 * Let's make space for retsp approximately: if space is short
	 * we will use realloc
	 */
	sprintf(name, "product of %s and %s", spa->sp_name, spb->sp_name);
	retsp = sparse(spa->sp_rown, spb->sp_coln, spb->sp_nonzero,
		name, SP_SINGLE);

	ncolx = retsp->sp_colx;
	nrowx = retsp->sp_rowx;
	ndatax = (float *) retsp->sp_data;
	alloced = retsp->sp_nonzero;

	/*
	 * Remember sparse_t is column based
	 */
	cnt = 0;
	for (i=0;i<spb->sp_coln; i++) {
		*ncolx++ = cnt;
		bno = spb->sp_colx[i+1] - spb->sp_colx[i];
		browx = &spb->sp_rowx[spb->sp_colx[i]];
		if (bno == 0)
			continue; /* no of i'th col of B is zero */

		for (j=0; j < spa->sp_rown; j++) {
			ano = spa->sp_rowindex[j+1] - spa->sp_rowindex[j];
			acolx = &spa->sp_colindex[spa->sp_rowindex[j]];
			adatax = &spa->sp_dataindex[spa->sp_rowindex[j]];
			if (ano == 0)
				continue; /* no of j'th row of B is zero */

			/* now cal the (j, i) entry of the result matrix */
			exist = s_getproduct(spa, acolx, adatax, ano,
				spb, browx, spb->sp_colx[i], bno, &retdata);
			if (exist) { /* there exists of (j, i) */
				if (cnt == alloced) {
					alloced += alloced;
					size = sizeof(float) * alloced;
					retsp->sp_data = (int *) /* XXX */
						realloc(retsp->sp_data, size);

					size = sizeof(int) * alloced;
					retsp->sp_rowx = (int *)
						realloc(retsp->sp_rowx, size);

					nrowx = retsp->sp_rowx + cnt;
					ndatax = (float *)retsp->sp_data + cnt;
				}
				*nrowx++ = j;
				*ndatax++ = retdata;
				cnt++;
			}
		}
	}
	*ncolx = cnt;

	/*
	 * rearrange
	 */
	retsp->sp_nonzero = cnt;

	ndatax = (float *) malloc((sizeof(float) * cnt));
	memcpy((void *) ndatax, (void *)retsp->sp_data, (sizeof(float) * cnt));
	free((void *) retsp->sp_data);
	retsp->sp_data = (int *) ndatax;

	ncolx = (int *) malloc((sizeof(int)*cnt));
	memcpy((void *) ncolx, (void *) retsp->sp_rowx, (sizeof(int) * cnt));
	free((void *) retsp->sp_rowx);
	retsp->sp_rowx = ncolx;

	return retsp;
}
