#include <stdio.h>
#include <sys/types.h>
#include "common.h"
#include "vector.h"

struct vector_t *
vec_sumofsq(struct vector_t *xp, struct vector_t *ap, struct vector_t *bp)
{
	int	i;
	float	*sap, *sbp;
	double	*ddata, *dap, *dbp;
	struct simplex_t *cap, *cbp;
	struct duplex_t *zap, *zbp;

	if (vec_isdouplets(ap, bp))
		fatal("%s(): NOT istriplets", __F);
	if (xp->vt_type != SP_DOUBLE || xp->vt_size != bp->vt_size)
		fatal("%s(): ERROR!!", __F);

	vec_null(xp);
	ddata = dvec(xp);
	switch (ap->vt_type) {
		case SP_SINGLE:
			svec(xp); sap = svec(ap); sbp = svec(bp);
			for (i=0;i<ap->vt_size;i++)
				ddata[i] = sap[i]*sap[i] + sbp[i]*sbp[i];
			break;
		case SP_DOUBLE:
			dap = dvec(ap); dbp = dvec(bp);
			for (i=0;i<ap->vt_size;i++)
				ddata[i] = dap[i]*dap[i] + dbp[i]*dbp[i];
			break;
		case SP_SIMPLEX:
			cap = cvec(ap); cbp = cvec(bp);
			for (i=0;i<ap->vt_size;i++)
				ddata[i] = cp_square(&cap[i])
					+ cp_square(&cbp[i]);
			break;
		case SP_DUPLEX:
			zap = zvec(ap); zbp = zvec(bp);
			for (i=0;i<ap->vt_size;i++)
				ddata[i] = zp_square(&zap[i])
					+ zp_square(&zbp[i]);
			break;
	}
	return xp;
}

struct vector_t *
vec_sqofsum(struct vector_t *xp, struct vector_t *ap, struct vector_t *bp)
{
	int	i;
	float	*sap, *sbp;
	double	*ddata, *dap, *dbp;
	struct simplex_t *cap, *cbp, ctmp;
	struct duplex_t *zap, *zbp, ztmp;

	if (vec_isdouplets(ap, bp))
		fatal("%s(): NOT istriplets", __F);
	if (xp->vt_type != SP_DOUBLE || xp->vt_size != bp->vt_size)
		fatal("%s(): ERROR!!", __F);

	vec_null(xp);
	ddata = dvec(xp);
	switch (ap->vt_type) {
		case SP_SINGLE:
			svec(xp); sap = svec(ap); sbp = svec(bp);
			for (i=0;i<ap->vt_size;i++) {
				ddata[i] = sap[i] + sbp[i];
				ddata[i] *= ddata[i];
			}
			break;
		case SP_DOUBLE:
			dap = dvec(ap); dbp = dvec(bp);
			for (i=0;i<ap->vt_size;i++) {
				ddata[i] = dap[i] + dbp[i];
				ddata[i] *= ddata[i];
			}
			break;
		case SP_SIMPLEX:
			cap = cvec(ap); cbp = cvec(bp);
			for (i=0;i<ap->vt_size;i++) {
				cp_plus(&ctmp, &cap[i], &cbp[i]);
				ddata[i] = cp_square(&ctmp);
			}
			break;
		case SP_DUPLEX:
			zap = zvec(ap); zbp = zvec(bp);
			for (i=0;i<ap->vt_size;i++) {
				zp_plus(&ztmp, &zap[i], &zbp[i]);
				ddata[i] = zp_square(&ztmp);
			}
			break;
	}
	return xp;
}
