#include <stdio.h>
#include "common.h"
#include "options.h"

void
opt_setup(void)
{
	opt_add(OPT_STRING, "gridfile",
		"The name of the input grid file", "grid", NULL);

	opt_add(OPT_INTEGER, "generate new mesh",
		"Do you want to generate new mesh (grid) file?",
		"0", "I[0:1]1");

	opt_add(OPT_TIMESTEP, "time integration parameters",
		"TimePrm::scan(Is) syntax",
		NULL, "dt=0.5  t in [0,64]");

	opt_add(OPT_STRING, "BoInd", "boundary indicator",
		"nb=4 names=RSIDE TOP LSIDE BOTTOM 1=(1), 2=(2), 3=(3), 4=(4)",
		NULL);

	opt_add(OPT_STRING, "printform", "format specifier in fprintf()",
		"%-.9lg ", "S");

	opt_add(OPT_INTEGER, "save_collected",
		"save collected stiffness matrix K?",
		"0", "I[0:1]1");

	opt_add(OPT_INTEGER, "load_collected",
		"load collected stiffness matrix K?",
		"0", "I[0:1]1");

	opt_add(OPT_INTEGER, "export_mat",
		"dump matrix with the sparse format ?",
		"0", "I[0:1]1");

	opt_add(OPT_INTEGER, "direct_method",
		"use direct method to solve linear equation ?",
		"1", "I[0:1]1");

	opt_add(OPT_INTEGER, "CheckSolution",
		"Check the accuracy of solution by calculating residuals",
		"1", "I[0:1]1");

	opt_add(OPT_REAL, "AD", "Dill's parameter A_Dill (unit = 1/um)",
		"0.54", "R[0.0:1.0]1");

	opt_add(OPT_REAL, "BD", "Dill's parameter B_Dill (unit = 1/um)",
		"0.06", "R[0.0:1.0]1");

	opt_add(OPT_REAL, "CD", "Dill's parameter C_Dill (unit = cm^2 sec/mJ)",
		"0.014", "R[0.0:1.0]1");

	opt_add(OPT_REAL, "m0", "Initial concentration of photoactive compound",
		"1.0", "R[0.0:1000.0]1");

	opt_add(OPT_REAL, "nr", "real part of refractive index of PR",
		"1.68", "R[0.0:10.0]1");

	opt_add(OPT_REAL, "n_substr",
		"real part of refractive index of substrate",
		"2.0", "R1");

	opt_add(OPT_REAL, "tr", "thickness of resist",
		"1.2", "R[0.0:100.0]1");

	opt_add(OPT_REAL, "lambda", "wavelength of incident light",
		"0.4358", "R[0.0:1000.0]1");

	opt_add(OPT_REAL, "NA", "optical Numerical aperture parameter",
		"0.6", "R[0.0:100.0]1");

	opt_add(OPT_REAL, "sigma", "optical coherence paramter",
		"0.5", "R[0.0:100.0]1");

	opt_add(OPT_REAL, "b_m", "mask period",
		"1.2", "R[0.0:100.0]1");

	opt_add(OPT_REAL, "a_m", "aperture width",
		"0.6", "R[0.0:100.0]1");

	opt_add(OPT_REAL, "Zi", "focal point position (z-direction)",
		"1.2", "R[0.0:100.0]1");

	opt_add(OPT_INTEGER, "MaxOrder", "maximum order of Fourier expansion",
		"64", "I[1:100]1");

	opt_add(OPT_REAL, "source sigmas", "normalized source position",
		"1 0.053  2 0.1  3 0.33;", NULL);

	opt_add(OPT_STRING, "output initial","initial of output file name",
		"expfem", NULL);

	opt_add(OPT_STRING, "input cnt", "input cnt format file name",
		"elbows.cnt", NULL);

	opt_add(OPT_REAL, "aniso_alpha", "alpha of a = alpha - j beta",
		"1.0", "R");

	opt_add(OPT_REAL, "aniso_beta", "beta of a = alpha - j beta",
		"3.0", "R");
}

void
opt_init(const char *filename)
{
	opt_setup();		/* litho::define */
	opt_read(filename);	/* change menu according to filename */
#if DEBUG
	opt_print();
#endif
	opt_check();		/* range check */
	return;
}
