#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <strings.h>
#include <time.h>
#include <math.h>

#include "common.h"
#include "options.h"
#include "vector.h"
#include "matrix.h"
#include "sparse.h"
#include "finite.h"

#define RSIDE      1
#define TOP        2
#define LSIDE      3
#define BOTTOM     4


struct duplex_t f_a(int);
struct duplex_t f_b(int);
struct duplex_t f_c(int);

struct duplex_t n_f(int subdo, int elemno, int intgno);
void	makeFemMatrix();
void	impose_bc(void);
void	imposeCondBC(int node);
void	updateData(int iSrc);
void	rightput(const char *filename);
void	saveR(void);

/*
 * File : prepost.c
 *  preprocessor and postprocessor functions
 */

#define	MAXSRC	7
#define  MAXPRDNODE 200

struct timestep *chronos;
struct vector_t *sum_src;

struct vector_t *vsr, *vsi;	/*double & imaginary part of Vs */
struct vector_t *m_field;

struct vector_t *feUis_re;
struct vector_t *feUis_im;	/*fieldFE object of Uis (x,z) */
struct vector_t *feLaplUis_re;
struct vector_t *feLaplUis_im;


				/* litho system specific parameters */

int	NoSrc;			/* number of light sources */
double	AD, BD, CD;
double	m0, nr;
double	tr, lambda;		/* thickness of resist, wavelength */
double	NA;
double	sigma;			/* coher. param. */
double	a_m, b_m;		/* aperture width, mask period */
double	Zi;  			/* focal point position */

				/* derived global parameters  */

double	wave_k;			/* wavenumber k */
double	alphaS;			/* alphaS = k NA sigmaS */
char	format[80];

double	n_substr;


double	 aniso_alpha;		/* alpha of a = alpha - j beta */
double	 aniso_beta;		/* beta  of a = alpha - j beta */
struct duplex_t aniso_a_X;	/* derived global var. for PML ABC  */
struct duplex_t aniso_b_X;
struct duplex_t aniso_a_Y;
struct duplex_t aniso_b_Y;
struct duplex_t aniso_c;

/* global variables for the complex arithmetic of OPTICAL system */

int	 MaxOrder;
int	 m_max;
double	x_bar;			/* distance unit */
double	b_mask;			/* external var. of b_m */
double	glA, glB; 		/* global paramter A and B (lambda*A/(4*M_PI))*/

char	*inputCnt;


void
opt_scan(void)
{
	double	sigma_val;

	chronos = opt_gettimestep("time integration parameters");

	AD  = opt_getreal("AD");
	BD  = opt_getreal("BD");
	CD  = opt_getreal("CD");
	m0  = opt_getreal("m0");
	nr  = opt_getreal("nr");

	tr	= opt_getreal("tr");
	n_substr= opt_getreal("n_substr");
	lambda  = opt_getreal("lambda");
	NA	= opt_getreal("NA");
	sigma   = opt_getreal("sigma");
	a_m	= opt_getreal("a_m");
	b_m	= opt_getreal("b_m");
	Zi	= opt_getreal("Zi");
	inputCnt= opt_getstr("input cnt");

	MaxOrder = opt_getint("MaxOrder");
	sigma_val = opt_getreal("source sigmas");

	m_max = MaxOrder;
	NoSrc = 1;			/* XXX */

	aniso_alpha  = opt_getreal("aniso_alpha");
	aniso_beta   = opt_getreal("aniso_beta");
}

void
setParam(void)
{
	aniso_c = duplexv(aniso_alpha, -1.0*aniso_beta); /*c = alpha - j beta */
	zp_inverse(&aniso_a_X, &aniso_c);	/* x-normal PML condition */
	aniso_b_X = aniso_c;			/* x-normal PML condition */

	aniso_a_Y = aniso_c;			/* y-normal PML condition */
	zp_inverse(&aniso_b_Y, &aniso_c);	/* y_normal PML condition */

	zp_print(&aniso_a_X);
	zp_print(&aniso_b_X);
	zp_print(&aniso_a_Y);
	zp_print(&aniso_b_Y);
}

void
Initialize(void)
{
	const int nno = meshes->m_nodesno;
	int	ntdf;

	vsr		= vector(nno, "VSR", SP_DOUBLE);
	vsi		= vector(nno, "VSI", SP_DOUBLE);
	m_field		= vector(nno, "m_field", SP_DOUBLE);
	feUis_re	= vector(nno, "feUis_re", SP_DOUBLE);
	feUis_im	= vector(nno, "feUis_im", SP_DOUBLE);
	feLaplUis_re	= vector(nno, "feLaplUis_re", SP_DOUBLE);
	feLaplUis_im	= vector(nno, "feLaplUis_im", SP_DOUBLE);

	ntdf = nno;

	/* External Matrix and Vectors initialization  */

	sum_src = vector(ntdf, "sum_src", SP_DOUBLE);
	vec_null(sum_src);

	return;
}

void
setIC(void)
{
	double	x, y;
	struct node_t *np;
	int	i;

	double	*m_fi;

	vec_null(vsr);
	vec_null(vsi);

	m_fi = (double *) m_field->vt_data;
	for (i=0; i< meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		x = np->n_coor[0];
		y = np->n_coor[1];
		if (y >= 0.0 && y <= tr)
			m_fi[i] = m0;
		else
			m_fi[i] = 0.0;
	}
}


#define  SMALL    0.00001
#define  MAXDIVX    50
#define  MAXDIVY    50

void
setTimeIndeps(void)
{
	FILE  *fp;

	double	 bc;

	const int nno = meshes->m_nodesno;
	struct duplex_t Uis, LaplUis;
	struct duplex_t temp;
	double	 e_arg, x, z;
	struct node_t *np;

	FILE    *fpi;
	float   cnt_xmin, cnt_xmax, cnt_ymin, cnt_ymax;
	float   fdivx, fdivy, cnt_dx, cnt_dy, cnt_dz;

	float   xpos, ypos;
	float   Ein[MAXDIVX+1][MAXDIVY+1];
	float   dxxEin[MAXDIVX+1]; 	/* d/dx (d/dx Ein (y=const)) */
	int	divx, divy, i, j;

	double xmin, xmax;

	int  cnt_i, cnt_j;
	double cnt_E_left, cnt_E_right;
	double cnt_dxxE_left, cnt_dxxE_right;
	double Uisx, laplUisx;

	double cnt_x_left, cnt_x_right, w1, w2, kz2;
	double dxx;

	wave_k =  2.0 * M_PI / lambda;
	x_bar  =  lambda/NA; 			/* distance unit */
	bc     =  b_m / x_bar; 			/* curled b (mask period) */
	b_mask =  b_m;
	glA    = -lambda*AD/ (4.0*M_PI);
	glB    = -lambda*BD/ (4.0*M_PI);

	fpi = fopen(inputCnt, "r");
	if (fpi == NULL) {
		fprintf (stderr,"file not found\n");
		exit(1);
	}

	fscanf (fpi,"%f %f %f %f\n",
		&cnt_ymin, &cnt_ymax, &cnt_xmin, &cnt_xmax);
	fscanf (fpi,"%f %f\n", &fdivx, &fdivy);
	divx = (int) fdivx;
	divy = (int) fdivy;

	if (divx > MAXDIVX || divy > MAXDIVY)  {
		fprintf (stderr,"Error: divx or divy too large (50).\n");
		exit(1);
	}

	printf ("%f %f  div : %d %d\n", cnt_xmin, cnt_xmax, divx, divy);

	if (fabs(b_m - (cnt_xmax-cnt_xmin)) > SMALL) {
		fprintf (stderr,
			"Warning : width of cnt and grid is different.\n");
	}

	for (j=0; j < divy; j++) {
		for (i=0; i < divx; i++) {
			fscanf (fpi, "%f ", &Ein[i][j]);
		}
		Ein[divx][j] = 0.0;
	}

	Ein[divx][divy] = 0.0;
	fclose(fpi); 		/* end of reading cnt file.... */

	cnt_dx = (cnt_xmax - cnt_xmin) / fdivx;
	cnt_dy = (cnt_ymax - cnt_ymin) / fdivy;
	cnt_dz = tr / divx;	/* assume divz = divx */

	xmin = -(cnt_xmax - cnt_xmin) / 2.0;
	xmax = xmin + (cnt_xmax - cnt_xmin);

	fprintf(stderr, "xmin %f xmax %f\n", xmin, xmax);

	fp = fopen ("00cnt.dat", "w");
	for (j=0; j < divy; j++) {
		for (i=0; i < divx; i++) {
			xpos = xmin + cnt_dx * i;
			ypos = cnt_dy * j;
			fprintf (fp, "%f\t%f\t%f\n", xpos, ypos, Ein[i][j]);
		}
	}
	fclose (fp);

	dxx   = 1.0 / (cnt_dx * cnt_dx);
	cnt_j = 0; 	/* for one slice  */

	for (cnt_i=1; cnt_i < divx-1; cnt_i++) {
		dxxEin[cnt_i] = (Ein[cnt_i+1][cnt_j] - 2.0*Ein[cnt_i][cnt_j] +
				Ein[cnt_i-1][cnt_j] ) * dxx;
	}

	dxxEin[0] = dxxEin[divx-1] = dxxEin[divx] = 0.0; /* APPROXIMATION ... */
	kz2 = wave_k * wave_k; 	/* APPROX. of k_z */

	for (i = 0; i < nno; i++) {
		np = meshes->m_nodes[i];
		x = np->n_coor[0];
		z = np->n_coor[1];

		if (x <= xmin || x >= xmax) {
			Uisx = 0.0;  laplUisx =  0.0;}
		else {
			cnt_i = (x - xmin) / cnt_dx; 	/* ?? float -> int */
			cnt_E_left  = Ein[cnt_i][cnt_j];
			cnt_E_right = Ein[cnt_i+1][cnt_j];
			cnt_dxxE_left  = dxxEin[cnt_i] - kz2 * cnt_E_left;
			cnt_dxxE_right = dxxEin[cnt_i+1] - kz2 * cnt_E_right;
			cnt_x_left  = xmin + cnt_dx * cnt_i;
			cnt_x_right = xmin + cnt_dx * (cnt_i + 1);
			w1 = (cnt_x_right - x) / cnt_dx;
			w2 = 1.0 - w1;
			Uisx     = w1 * cnt_E_left + w2 * cnt_E_right;
			laplUisx = w1 * cnt_dxxE_left + w2 * cnt_dxxE_right;
		}
		e_arg = wave_k * z;

		/*
		 * Dirty complex calculation
		 */
		temp = duplexv(cos(e_arg), sin(e_arg));
		zp_crossd(&Uis, &temp, Uisx);

		temp    = duplexv(cos(e_arg), sin(e_arg));
		zp_crossd(&LaplUis, &temp, laplUisx);

		dvec(feUis_re)[i] = Uis.real;
		dvec(feUis_im)[i] = Uis.imag;
		dvec(feLaplUis_re)[i] = LaplUis.real;
		dvec(feLaplUis_im)[i] = LaplUis.imag;
	}

#if 0
	fp = fopen("out", "w");
	dp = dvec(feLaplUis_re);
	for (i=0;i<meshes->m_nodesno;i++) {
		np = meshes->m_nodes[i];
		fprintf(fp, "%15.6f\n", dp[i]);
	}
	fclose(fp);
	put_gridandvector(feUis_re, "uis_re");
	put_gridandvector(feUis_im, "uis_im");
	put_gridandvector(feLaplUis_re, "uislp_re");
	put_gridandvector(feLaplUis_im, "uislp_im");

	feUis_re.values().print ("FILE=uis_re");
	feUis_im.values().print ("FILE=uis_im");
	feLaplUis_re.values().print ("FILE=uislp_re");
	feLaplUis_im.values().print ("FILE=uislp_im");
#endif
}


void updateData (int iSrc)
{
	const int nno = meshes->m_nodesno;
	const int ndof = meshes->m_nodesno;	/* DOF == 1 */
	struct duplex_t c_vs_at_this_node;
	double		abs_cvs, farg;
	int	i;
	struct duplex_t *duxp, tmpa, tmpb;

	duxp = zvec(meshes->m_bvector);
	for (i=0; i<ndof; i++) {
		dvec(vsr)[i] = duxp[i].real;
		dvec(vsi)[i] = duxp[i].imag;
	}

	for (i=0; i<nno; i++) {	/* for all nodal points  */
		tmpa = duplexv(dvec(vsr)[i], dvec(vsi)[i]);
		tmpb = duplexv(dvec(feUis_re)[i],dvec(feUis_im)[i]);

		zp_plus(&c_vs_at_this_node, &tmpa, &tmpb);
		abs_cvs = zp_absolute(&c_vs_at_this_node);

		dvec(sum_src)[i] += (abs_cvs * abs_cvs);
		farg = -CD * dvec(sum_src)[i] * chronos->ts_step;
		dvec(m_field)[i] *= exp(farg);
	}
	put_gridandvector(m_field, "m_field.mtv");
	vec_null(sum_src); 			/* clear (reset) sum */
	return;
}


void
problem(void)
{
	int	i;
	struct vector_t *xp;

	opt_init("test3.i");
	opt_scan();		/* litho::scan */
	setParam();		/* litho::setParam--- aniso--- */
	Initialize();		/* litho::Initialze */
	setTimeIndeps();	/* litho::setTimeIndeps */
	setIC();		/* litho::setIC */

	verbose = 1;
	xp = vec_puredup(meshes->m_bvector);

	chronos = opt_gettimestep("time integration parameters");
	ts_info(chronos);
	while (ts_next(chronos)) {
		for (i=0; i < NoSrc; i++) {
			makeFemMatrix();
			/*
			 * sp_print(meshes->m_kmatrix);
			 * rightput("beforebc");
			 * XXX
			 */
			impose_bc();
			sp_symcheck(meshes->m_kmatrix);
			verbose = 1;
			solveall();
		}
		updateData(0);
		saveR();
	}
	return;
}


void
makeFemMatrix(void)
{
	int	i;
	struct matrix_t	*retmatf, *bmatf;
	struct proto_t	*prevproto = NULL;

	sp_null(meshes->m_kmatrix);
	vec_null(meshes->m_bvector);

	currp = NULL;
	for (i=0; i< meshes->m_ementsno; i++) {
		/*
		 * currp is only set at this location. You must not set currp
		 * at other places.
		 */
		currp = meshes->m_ements[i]->e_proto;
		if (prevproto != currp) {
			if (prevproto != NULL)
				m_mfree(2, retmatf, bmatf);/* no longer used */
			create_smalls(currp, &retmatf, &bmatf);
		}

		integration(i, retmatf, bmatf);
		bassemble(i, bmatf, meshes->m_bvector);
		Kassemble(i, retmatf, meshes->m_kmatrix);
		prevproto = currp;
	}
	m_mfree(2, retmatf, bmatf);
	return;
}

void
integration(const int elemno, struct matrix_t *retmatf, struct matrix_t *bmatf)
{
	struct ement_t *ep;
	int	i, j, k;
	struct duplex_t **retmat, **bmat;
	double	**unit, **deriv, *weights, *jacobs;
	double	x, y;

	double	k2  = wave_k * wave_k;
	struct duplex_t ncmpl, Uis, LaplUis, f_src;
	struct duplex_t axx, ayy, beta;
	struct duplex_t dxtmp, dx1, dx2, dx3;
	int	subdomain;
	struct integral_t *intg;

	proto_einit(elemno);
	intg = currp->p_intg;
	m_mnull(2, retmatf, bmatf);

	retmat = zmat(retmatf);
	bmat   = zmat(bmatf);
	unit   = dmat(intg->intg_trans);
	jacobs = intg->intg_jacob;
	weights= intg->intg_weights;

	ep = meshes->m_ements[elemno];
	subdomain = ep->e_sdomain;
	for (k=0; k<currp->p_intgno; k++) {	/* four integration points */
		deriv = dmat(intg->intg_deriv[k]);
		ncmpl = n_f(subdomain, elemno, k);

		x = getevalpoint(elemno, k, 0);
		y = getevalpoint(elemno, k, 1);

		Uis    =duplexv(dgetat(elemno, k, feUis_re),
				dgetat(elemno, k, feUis_im));
		LaplUis=duplexv(dgetat(elemno, k, feLaplUis_re),
				dgetat(elemno, k, feLaplUis_im));

		zp_3cross(&dxtmp, &ncmpl, &ncmpl, &Uis);
		zp_crossd(&dxtmp, &dxtmp, k2);
		zp_plus(&f_src, &LaplUis, &dxtmp);

		dx1 = f_b(subdomain);
		dx2 = f_a(subdomain);
		zp_inverse(&axx, &dx1);
		zp_inverse(&ayy, &dx2);

		dx3 = f_c(subdomain);

		zp_3cross(&dxtmp, &ncmpl, &ncmpl, &dx3);
		zp_crossd(&beta, &dxtmp, -1.0 * k2);

		for (i=0; i< currp->p_nodsno;i++) {	/* XXX */
			for (j=0; j< currp->p_nodsno;j++) {
				zp_crossd(&dx1, &axx, deriv[0][i]*deriv[0][j]);
				zp_crossd(&dx2, &ayy, deriv[1][i]*deriv[1][j]);
				zp_crossd(&dx3, &beta, unit[k][i]*unit[k][j]);
				zp_3plus(&dxtmp, &dx1, &dx2, &dx3);
				zp_crossd(&dxtmp, &dxtmp, jacobs[k]*weights[k]);

				zp_pplus(&retmat[i][j], &dxtmp);
			}
			zp_pplus(&bmat[i][0], zp_crossd(&dx1,
				&f_src, unit[k][i]*jacobs[k]*weights[k]));
		}
	}
	if ((elemno % 100) == 0)
		printf("%d\n", elemno);
	return;
}


struct duplex_t
n_f(int subdo, int elemno, int intgno)
{
	struct duplex_t ret;
	double	tmp;

	if (subdo <= 10)
		ret = duplexv(n_substr, 0.0);    // substrate
	else if (subdo >= 16)
		ret = duplexv(1.0, 0.0);     // vacuum
	else {
		tmp = dgetat(elemno, intgno, m_field);
		ret = duplexv(nr, glA * tmp + glB) ;
	}
	return ret;
}


void
impose_bc(void)
{
	int	nno = meshes->m_nodesno;
	int	i;

	zdirichlet(TOP, duplex_null);

	for (i=0; i < nno; i++) {
		if (!isbdset(i))
			continue;
		if (isbdnoset(i, BOTTOM))
			imposeCondBC(i);
	}
	return;
}

void
imposeCondBC(int no)
{
	struct duplex_t dux;

	dux = duplexv(dvec(feUis_re)[no] * (-1.0),
		dvec(feUis_im)[no] * (-1.0));
	di_zimpose(no, dux);
	return;
}


void
saveR(void)
{
	return;
}

struct duplex_t
f_a(int subdo)
{
	struct duplex_t ret;

	if (subdo <= 5 || subdo >= 21)
		ret = aniso_a_Y;
	else if ( subdo==6 || subdo==10 || subdo==11 || subdo==15 || 
	    subdo==16 || subdo==20)
		ret = aniso_a_X;
	else
		ret = duplexv(1.0, 0.0);
	return ret;
}

struct duplex_t
f_b(int subdo)
{
	struct duplex_t ret;

	if (subdo <= 5 || subdo >= 21)
		ret = aniso_b_Y;
	else if ( subdo==6 || subdo==10 || subdo==11 || subdo==15 || 
	    subdo==16 || subdo==20)
		ret = aniso_b_X;
	else
		ret = duplexv(1.0, 0.0);

	return ret;
}

struct duplex_t
f_c(int subdo)
{
	struct duplex_t ret;

	if (subdo<=5 || subdo>=21)  ret = aniso_c;

	else if ( subdo==6 || subdo==10 || subdo==11 || subdo==15 || 
	     subdo==16 || subdo==20)
		ret = aniso_c;
	else
		ret = duplexv(1.0, 0.0);

	return ret;
}
