#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <math.h>

#include "common.h"
#include "options.h"
#include "vector.h"
#include "matrix.h"
#include "sparse.h"
#include "finite.h"

#define __L  __LINE__
#define SD_LEFT		1
#define SD_NEAR		2
#define SD_RIGHT	5
#define SD_FAR		6
#define SD_RESIST	4	/* XXX */
#define SD_SUBSTRATE	3	/* XXX */

#define F_LEFT	0x01
#define F_RIGHT	0x02
#define F_NEAR	0x04
#define F_FAR	0x08

#define X_CENTER  0.00
#define XL_MARGIN -0.900
#define XR_MARGIN +0.900
#define X_PMLWIDTH 0.10
#define X_LEN (2.0)

#define Y_CENTER  0.750
#define YL_MARGIN 0.100
#define YR_MARGIN 1.400
#define Y_PMLWIDTH 0.10
#define Y_LEN (YL_MARGIN+YR_MARGIN)

int	lightsrcno;
int	resist_on;
int	resist_off;

struct timestep *chronos;
struct vector_t *sum_src;
struct vector_t *m_field;

double	AD, BD, CD;
double	m0, nr;
double	lambda;

double	n_substr;
double	glA, glB; 

float	resist0;
float	resist1;

struct vector_t *pure_light;
struct vector_t *lapl_light;
struct vector_t *tot_light;
int	lightcnt;
int	lightnodeno;
double	wave_k;			/* wavenumber k */

struct vector_t *mux;
struct vector_t *muy;
struct vector_t *epsilon;

float	delta_x;
float	delta_z;
int	xnodesno, znodesno;

int	n_max, n_all; /* of fourier */
int	n_lightno;
double	Zi;
double	a_m, b_m;
double	x_bar;
double	NA;
struct matrix_t *cee_si;
struct vector_t *fourier;
struct matrix_t *gamma_sn;
struct matrix_t *argument;
struct vector_t *sigma_s;

int	whichsubdo(const struct ement_t *);
int	whichbound(const struct node_t *);
void	set_const3rd(const double);
void	putmsg(const char *msg, const int lineno);
void	impose_bcs(void);
void	initial(void);
void	opt_scan(void);
void	mk_femmatrix();
void	impose_bc(void);
void	imposeCondBC(int node);
void	update(const int saveit);
void	vec_putresist(const char *fname, const struct vector_t *vp);
int	insideunit(const double x);
void	set_deltas(void);
void	resist(void);
void	get_lightatfile2nd(const char *fname);
void	get_lightatfile(const char *fname);
void	mk_laplace(void);
void	mk_light(const int light);

struct vector_t *get_sigmas(const char *fname);
struct duplex_t exponent(const double x);
struct duplex_t n_f(int subdo, int elemno, int intgno);


void
problem(void)
{
	int	i, j;

	n_max = 10;
	n_all = n_max * 2 + 1;
	lightsrcno = find_nearestnode(1.0, 0.75);
	sigma_s  = get_sigmas("sigma.0");

	set_deltas();
	opt_init("test3.i");
	opt_scan();		/* litho::scan */
#if DEBUG
	opt_print();
#endif
	wave_k =  2.0 * M_PI / lambda;
	x_bar     = lambda/NA;

	set_subdomain(whichsubdo);
	set_boundary(whichbound);
	set_const3rd(7.0);

	pure_light = vector(meshes->m_nodesno, NULL, SP_DUPLEX);
	lapl_light = vector(meshes->m_nodesno, NULL, SP_DUPLEX);
	tot_light  = vector(meshes->m_nodesno, NULL, SP_DUPLEX);
	lightnodeno = find_nearestnode(1.0, 1.0);
	resist();

	initial();		/* litho::Initialze */
#if 0
	if (0)
		get_lightatfile2nd("twolines");
	else
		get_lightatfile("thison.all.547");
#endif


	chronos = opt_gettimestep("time integration parameters");
	ts_info(chronos);

	meshes->m_kmatrix = create_kmatrix();
	meshes->m_bvector = create_bvector();
	for (i=0; i<51; i++) {
		for (j=0; j < n_lightno; j++) {
			printf("%s(): %2dth\n", __F, i);
			mk_light(j);
			mk_femmatrix();
			impose_bc();
			solveall();
			update(j == (n_lightno-1));
		}
		if (i != 0 && ((i%5) == 0))
			vec_putresist("a", m_field);
	}
	return;
}

int
fdiff(const float one, const float two)
{
	if (fabs(one-two) > 0.00001)
		return 1;
	return 0;
}

void
set_deltas(void)
{
	int	i;
	struct node_t *np;
	float	oldx, x, oldz, z;
	int	xon, zon;

	xon = 0; zon = 0;
	for (i=0; i<meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		x = np->n_coor[0];
		z = np->n_coor[1];
		if (i == 0) {
			oldx = x;
			oldz = z;
			continue;
		}
		if ((xon == 0) && fdiff(x, oldx)) {
			delta_x = fabs(x-oldx);
			xon = 1;
		}
		if ((zon == 0) && fdiff(z, oldz)) {
			delta_z = fabs(z-oldz);
			xnodesno = i;
			zon = 1;
		}
		if (xon && zon)
			break;
	}
	znodesno = meshes->m_nodesno/xnodesno;
	if (meshes->m_nodesno % xnodesno)
		fatal("%s(): ", __F);
	printf("%s(): xnodesno = %3d\n", __F, xnodesno);
	printf("%s(): znodesno = %3d\n", __F, znodesno);
	return;
}

int
oob(const int nodeno)
{
	if (nodeno < 0 || nodeno >= meshes->m_nodesno)
		return 1;
	return 0;
}

int
back(const int nodeno)
{
	int	ret;
	ret = nodeno - xnodesno;
	return oob(ret) ? nodeno : ret;
}
		
int
fore(const int nodeno)
{
	int	ret;
	ret = nodeno + xnodesno;
	return oob(ret) ? nodeno : ret;
}

int
left(const int nodeno)
{
	if ((nodeno % xnodesno) == 0)
		return nodeno;
	return nodeno - 1;
}

int
right(const int nodeno)
{
	if (((nodeno+1) % xnodesno) == 0)
		return nodeno;
	return nodeno + 1;
}

struct value_t {
	int	v_index;
	float	v_x;
	struct duplex_t	v_pure;
} *vroot;

int
get_index(const float xcoor)
{
	struct value_t *p;

	for (p=vroot;; p++) {
		if (p->v_x >= xcoor)
			return p->v_index;
	}
}
		


void
mk_laplace(void)
{
	struct duplex_t *photos, *lphotos, dx1, dx2, dx3;
	float	dxx, dzz;
	int	i;

	photos = zvec(pure_light);
	lphotos = zvec(lapl_light);

	dxx = 1.0/(delta_x * delta_x);
	dzz = 1.0/(delta_z * delta_z);
	printf("%s(): dxx = %7.5f\n", __F, dxx);
	printf("%s(): dzz = %7.5f\n", __F, dzz);
	for (i=0; i<meshes->m_nodesno; i++) {
		/* z please do not change dx2*/
		zp_crossd(&dx2, &photos[i], 2.0);

		zp_plus(&dx1, &photos[fore(i)], &photos[back(i)]);
		zp_minus(&dx3, &dx1, &dx2);
		zp_crossdme(&dx3, dzz);
		zp_pplus(&lphotos[i], &dx3);

		/* x */
		zp_plus(&dx1, &photos[right(i)], &photos[left(i)]);
		zp_minus(&dx3, &dx1, &dx2);
		zp_crossdme(&dx3, dxx);
		zp_pplus(&lphotos[i], &dx3);

	}
	putmsg(__F, __L);

	return;
}


void
mk_light(const int light)
{
	struct duplex_t *photos, cee, dx0, dx1, dx2, sum;
	struct duplex_t *ceep;
	double 	*gammap, *argp;
	double	db0, db1;
	float	x, x_wave, z;
	int	i, j;

	putmsg(__F, __L);
	vec_null(pure_light); /* XXX */
	vec_null(lapl_light); /* XXX */
	photos = zvec(pure_light);

	ceep   = zmat(cee_si)[light];
	gammap = dmat(gamma_sn)[light];
	argp   = dmat(argument)[light];

	for (j=0; j<meshes->m_nodesno; j++) {
		x = meshes->m_nodes[j]->n_coor[0];
		z = meshes->m_nodes[j]->n_coor[1];
		x_wave = x/x_bar;
		sum = duplex_null;
		for (i=0; i < n_all; i++) {
			cee = ceep[i];
			db0 = -1.0 * wave_k * gammap[i];
			dx0 = exponent(db0*z);

			db1 = 2.0 * M_PI * argp[i];
			dx1 = exponent(db1*x_wave);

			zp_3cross(&dx2, &cee, &dx0, &dx1);
			zp_pplus(&sum, &dx2);
		}
		zp_assign(&photos[j], &sum);
	}
	mk_laplace();

	return;
}


void
get_lightatfile(const char *fname)
{
	FILE	*fp;
	struct duplex_t *photos, *lphotos, dx1, dx2, dx3;
	float	dxx, dzz;
	float	ft0, ft1, ft2;
	int	cnt, readn;
	char	line[256];
	int	i;
	double	max, maxt;

	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("%s(): can not open", __F);

	cnt = 0;
	while(fgets(line, 256, fp))
		cnt++;
	fclose(fp);

	vroot = (struct value_t *) malloc(sizeof(struct value_t)*(cnt+1));
	if (vroot == NULL)
		fatal("%s(): vroot", __F);
	for (i=0; i<(cnt+1); i++)
		vroot[i].v_index = i;

	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("%s(): can not open", __F);

	cnt = 0;
	max = -1.0e+10;
	while(fgets(line, 256, fp)) {
		readn = sscanf(line, "%f %f %f",
			&ft0, &ft1, &ft2);
		if (readn != 3)
			fatal("%s(): sscanf()", __F);
		vroot[cnt].v_x = ft0;
		vroot[cnt].v_pure.real = ft1;
		vroot[cnt].v_pure.imag = ft2;
		maxt = zp_absolute(&vroot[cnt].v_pure);
	
		/*
		 * XXX
		 */
#if 1
		vroot[cnt].v_pure.real = maxt;
		vroot[cnt].v_pure.imag = maxt;
#endif
		if (maxt > max) 
			max = maxt;
		cnt++;
	}
	fclose(fp);
	putmsg(__F, __L);

	vroot[cnt].v_x = ft0 + 0.02;
	vroot[cnt].v_pure.real = ft1;
	vroot[cnt].v_pure.imag = ft2;

	/*
	 * normal lize
	 */
	maxt = 1.0/max;
	for (i=0; i < cnt; i++) {
		zp_crossdme(&vroot[i].v_pure, maxt);
	}
	/*
	 * end of reading
	 * Let's calculate the Laplace
	 */
	lightcnt = cnt;

	photos = zvec(pure_light);

	for (i=0; i<meshes->m_nodesno; i++) {
		ft0 = meshes->m_nodes[i]->n_coor[0];
		ft1 = meshes->m_nodes[i]->n_coor[1];
		dx1 = duplexv(cos(wave_k * ft1), sin(wave_k * ft1));
		dx2 = vroot[get_index(ft0)].v_pure;
		zp_cross(&dx3, &dx1, &dx2);
		
		photos[i] = dx3;
	}
	putmsg(__F, __L);

	lphotos = zvec(lapl_light);

	dxx = 1.0/(delta_x * delta_x);
	dzz = 1.0/(delta_z * delta_z);
	printf("%s(): dxx = %7.5f\n", __F, dxx);
	printf("%s(): dzz = %7.5f\n", __F, dzz);
	for (i=0; i<meshes->m_nodesno; i++) {
		/* z please do not change dx2*/
		zp_crossd(&dx2, &photos[i], 2.0);

		zp_plus(&dx1, &photos[fore(i)], &photos[back(i)]);
		zp_minus(&dx3, &dx1, &dx2);
		zp_crossdme(&dx3, dzz);
		zp_pplus(&lphotos[i], &dx3);

		/* x */
		zp_plus(&dx1, &photos[right(i)], &photos[left(i)]);
		zp_minus(&dx3, &dx1, &dx2);
		zp_crossdme(&dx3, dxx);
		zp_pplus(&lphotos[i], &dx3);

	}
	putmsg(__F, __L);

	return;
}

struct value2nd_t {
	int	v_index;
	float	v_x;
	float	v_val;
} *vroot2nd;

float
get_index2nd(const float xcoor, const int cnt)
{
	int	i;
	float	w1, w2, len;
	struct value2nd_t *vp;
#if 0
	struct value2nd_t *p, *base = vroot2nd;
	int lim, cmp;

	for (lim = cnt; lim != 0; lim >>= 1) {
		p = base + (lim >> 1);
		cmp = p->v_x < xcoor ? 1 : p->v_x > xcoor ? -1 : 0;
		if (cmp == 0)
			break;
		if (cmp > 0) {  /* key > p: move right */
			base = (p + 1);
			lim--;
		} /* else move left */
	}
	i = p->v_index;

#else
	for (vp = vroot2nd,i=0;i<cnt; vp++, i++) {
		if (vp->v_x > xcoor)
			break;
	}
#endif
	if (i == cnt || i == 0)
		return vroot2nd[cnt].v_val;
	len = vroot2nd[i].v_x - vroot2nd[i-1].v_x;
	w1 = (xcoor-vroot2nd[i-1].v_x)/len;
	w2 = (vroot2nd[i].v_x - xcoor)/len;

	len = w1 * vroot2nd[i-1].v_val + w2 * vroot2nd[i].v_val;
	return len;
}

void
get_lightatfile2nd(const char *fname)
{
	FILE	*fp;
	struct duplex_t *photos, *lphotos, dx1, dx2, dx3;
	float	dxx, dzz;
	float	ft0, ft1, ft2;
	int	cnt, readn;
	char	line[256];
	int	i;
	double	max, maxt;

	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("%s(): can not open", __F);

	cnt = 0;
	while(fgets(line, 256, fp))
		cnt++;
	fclose(fp);

	vroot2nd = (struct value2nd_t *)
		malloc(sizeof(struct value2nd_t)*(cnt+1));
	if (vroot2nd == NULL)
		fatal("%s(): vroot2nd", __F);
	for (i=0; i<(cnt+1); i++)
		vroot2nd[i].v_index = i;

	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("%s(): can not open", __F);

	cnt = 0;
	max = -1.0e+10;
	while(fgets(line, 256, fp)) {
		readn = sscanf(line, "%f %f", &ft0, &ft1);
		if (readn != 2)
			fatal("%s(): sscanf()", __F);
		vroot2nd[cnt].v_x = ft0;
		vroot2nd[cnt].v_val = ft1;
		maxt = ft1;

		if (maxt > max) 
			max = maxt;
		cnt++;
	}
	fclose(fp);
	putmsg(__F, __L);

	vroot2nd[cnt].v_x = ft0 + 0.02;
	vroot2nd[cnt].v_val = ft1;

	/*
	 * normal lize
	 */
#if 0
	maxt = 1.0/max;
	for (i=0; i < cnt; i++) {
		zp_crossdme(&vroot2nd[i].v_pure, maxt);
	}
#endif
	/*
	 * end of reading
	 * Let's calculate the Laplace
	 */
	lightcnt = cnt;

	photos = zvec(pure_light);

	for (i=0; i<meshes->m_nodesno; i++) {
		ft0 = meshes->m_nodes[i]->n_coor[0];
		ft1 = meshes->m_nodes[i]->n_coor[1];
		dx1 = duplexv(cos(wave_k * ft1), sin(wave_k * ft1));
		ft2 = get_index2nd(ft0, lightcnt);
		zp_crossd(&dx3, &dx1, ft2);
		
		photos[i] = dx3;
	}
	putmsg(__F, __L);

	lphotos = zvec(lapl_light);

	dxx = 1.0/(delta_x * delta_x);
	dzz = 1.0/(delta_z * delta_z);
	printf("%s(): dxx = %7.5f\n", __F, dxx);
	printf("%s(): dzz = %7.5f\n", __F, dzz);
	for (i=0; i<meshes->m_nodesno; i++) {
		/* z please do not change dx2*/
		zp_crossd(&dx2, &photos[i], 2.0);

		zp_plus(&dx1, &photos[fore(i)], &photos[back(i)]);
		zp_minus(&dx3, &dx1, &dx2);
		zp_crossdme(&dx3, dzz);
		zp_pplus(&lphotos[i], &dx3);

		/* x */
		zp_plus(&dx1, &photos[right(i)], &photos[left(i)]);
		zp_minus(&dx3, &dx1, &dx2);
		zp_crossdme(&dx3, dxx);
		zp_pplus(&lphotos[i], &dx3);

	}
	putmsg(__F, __L);

	return;
}

void
resist(void)
{
	int	i, onoff;
	float	z;

	onoff = 0;
	for (i=0; i < meshes->m_ementsno; i++) {
		z = get_centercoor(i, 1);
		if ((onoff == 0) && z > resist0) {
			resist_on = i;
			onoff = 1;
		}
		if ((onoff == 1) && z > resist1) {
			resist_off = i;
			break;
		}
	}
	return;
}


void
opt_scan(void)
{
	printf("start  %s():", __F);
	chronos = opt_gettimestep("time integration parameters");

	AD  = opt_getreal("AD");
	BD  = opt_getreal("BD");
	CD  = opt_getreal("CD");
	m0  = opt_getreal("m0");
	nr  = opt_getreal("nr");
	NA  = opt_getreal("NA");
	a_m  = opt_getreal("a_m");
	b_m  = opt_getreal("b_m");
	Zi  = opt_getreal("Zi");

	n_substr= opt_getreal("n_substr");
	lambda  = opt_getreal("lambda");

	printf("a_m = %7.5f\n", a_m);
	printf("b_m = %7.5f\n", b_m);
	printf("end    %s():", __F);
}

/*
@ARTICLE{urbach,
        author="H.Paul Urbach and Douglas A. Bernard",
        title="``{Modeling latent-image formation in photolithography,
                using the Helmholtz equation}''",
        journal="Optical society of America",
        volume="6",
        number="9",
        pages="1343--1356",
        year="1989"
}
 */


void
initial(void)
{
	double	z;
	int	i, j, net_n;
	double	*dp, *fourierp, **argump, *sigmap, **gammap;
	struct duplex_t **ceep;
	double	tmp0, tmp1;
	double	b_bar;
	struct duplex_t dx0;

	b_bar = b_m/x_bar;
	glA = -1.0 * lambda * AD / (4.0*M_PI);
	glB = -1.0 * lambda * BD / (4.0*M_PI);

	m_field	= vector(meshes->m_nodesno, "m_field", SP_DOUBLE);
	sum_src = vector(meshes->m_nodesno, "sum_src", SP_DOUBLE);
	dp = dvec(m_field);

	for (i=0; i< meshes->m_nodesno; i++) {
		z = meshes->m_nodes[i]->n_coor[1];
		if (z >= resist0 && z <= resist1)
			dp[i] = m0;
		else
			dp[i] = 0.0;
	}

	gamma_sn = matrix(n_lightno, n_all, NULL, SP_DOUBLE);
	argument = matrix(n_lightno, n_all, NULL, SP_DOUBLE);
	cee_si   = matrix(n_lightno, n_all, NULL, SP_DUPLEX);
	fourier  = vector(n_all, NULL, SP_DOUBLE);

	fourierp = dvec(fourier);
	for (i=0; i < n_all; i++) {
		net_n = i - n_max;
		tmp0 = M_PI * (double) net_n;
		if (net_n == 0)
			fourierp[i] = a_m/b_m;
		else
			fourierp[i] = sin(tmp0 * a_m/b_m)/tmp0;
	}

	argump = dmat(argument);
	gammap = dmat(gamma_sn);
	ceep   = zmat(cee_si);
	sigmap = dvec(sigma_s);
	for (i=0; i < n_lightno; i++) {
		for (j=0; j < n_all; j++) { 
			net_n = j - n_max;
			tmp0 = sigmap[i] + ((double) net_n)/b_bar;
			argump[i][j] = tmp0;
			if (insideunit(argump[i][j])) {
				tmp1 = NA * tmp0;
				gammap[i][j] = sqrt(1.0-(tmp1*tmp1));
				tmp0 = wave_k * gammap[i][j] * Zi;

				dx0 = exponent(tmp0);
				zp_crossd(&ceep[i][j], &dx0, fourierp[j]);
			} else {
				gammap[i][j] = 0.0;
				ceep[i][j]   = duplex_null;
			}
	
		}
	}
	return;
}


void
update(const int lastone)
{
	double abs_cvs, farg;
	int	i;
	struct duplex_t *all;

	vec_plus(tot_light, pure_light, meshes->m_bvector);
	all = zvec(tot_light);

	for (i=0; i < meshes->m_nodesno; i++) {
		abs_cvs = zp_absolute(&all[i]);
		dvec(sum_src)[i] += (abs_cvs * abs_cvs);
	}
	if (lastone) {
		printf("%s(): lastone!!\n", __F);
		for (i=0; i < meshes->m_nodesno; i++) {
			farg = -CD * dvec(sum_src)[i] * chronos->ts_step;
			dvec(m_field)[i] *= exp(farg);
		}
		vec_null(sum_src);
	}
	return;
}

void
mk_femmatrix(void)
{
	int	i;
	struct matrix_t	*retmatf, *bmatf;
	struct proto_t	*prevproto = NULL;

	/*
	 * Make null
	 * integration
	 */
	sp_null(meshes->m_kmatrix);
	vec_null(meshes->m_bvector);

	currp = NULL;
	for (i=0; i< meshes->m_ementsno; i++) {
		/*
		 * currp is only set at this location. You must not set currp
		 * at other places.
		 */
		currp = meshes->m_ements[i]->e_proto;
		if (prevproto != currp) {
			if (prevproto != NULL)
				m_mfree(2, retmatf, bmatf);/* no longer used */
			create_smalls(currp, &retmatf, &bmatf);
		}

		integration(i, retmatf, bmatf);
		bassemble(i, bmatf, meshes->m_bvector);
		Kassemble(i, retmatf, meshes->m_kmatrix);
		prevproto = currp;
	}
	m_mfree(2, retmatf, bmatf);
	return;
}

void
integration(const int elemno, struct matrix_t *retmatf, struct matrix_t *bmatf)
{
	struct ement_t *ep;
	struct proto_t *proto;
	struct integral_t *intg;
	int	i, j, k;
	nodno_t	*nnp;

	struct duplex_t **retmat, **bmat;
	double	**unit, **deriv, *weights, *jacob;

	double	k2  = wave_k * wave_k;
	struct duplex_t ncmpl, Uis, LaplUis, f_src;
	struct duplex_t axx, ayy, beta;
	struct duplex_t dxtmp, dx1, dx2, dx3;
	int	subdomain;

	m_mnull(2, retmatf, bmatf);
	proto = currp;
	proto_einit(elemno);
	intg = proto->p_intg;

	retmat = zmat(retmatf);
	bmat = zmat(bmatf);

	jacob  = intg->intg_jacob;
	weights= intg->intg_weights;
	unit   = dmat(intg->intg_trans);


	ep = meshes->m_ements[elemno];
	nnp = ep->e_nodes;
	subdomain = ep->e_sdomain;
	for (k=0; k<proto->p_intgno; k++) {	/* four integration points */
		ncmpl = n_f(subdomain, elemno, k);
		deriv = dmat(intg->intg_deriv[k]);

		Uis     = zgetat(elemno, k, pure_light);
		LaplUis = zgetat(elemno, k, lapl_light);

		zp_3cross(&dxtmp, &ncmpl, &ncmpl, &Uis);
		zp_crossdme(&dxtmp, k2);
		zp_plus(&f_src, &LaplUis, &dxtmp);

		dx1 = zgetat(elemno, k, muy);
		dx2 = zgetat(elemno, k, mux);
		dx3 = zgetat(elemno, k, epsilon);
		if (zp_isnull(&dx1))
			fatal("%s(): [1] %d elm %dth p", __F, elemno, k);
		if (zp_isnull(&dx2))
			fatal("%s(): [2] %d elm %dth p", __F, elemno, k);
		zp_inverse(&axx, &dx1);
		zp_inverse(&ayy, &dx2);

		zp_3cross(&dxtmp, &ncmpl, &ncmpl, &dx3);
		zp_crossd(&beta, &dxtmp, -1.0 * k2);

		for (i=0; i< proto->p_nodsno;i++) {	/* XXX */
			for (j=0; j< proto->p_nodsno;j++) {
				zp_crossd(&dx1, &axx, deriv[0][i]*deriv[0][j]);
				zp_crossd(&dx2, &ayy, deriv[1][i]*deriv[1][j]);
				zp_crossd(&dx3, &beta, unit[k][i]*unit[k][j]);
				zp_3plus(&dxtmp, &dx1, &dx2, &dx3);
				zp_crossd(&dxtmp, &dxtmp, jacob[k]*weights[k]);

				zp_pplus(&retmat[i][j], &dxtmp);
			}
#if 1
			zp_pplus(&bmat[i][0], zp_crossd(&dx1, &f_src,
				unit[k][i]*jacob[k]*weights[k]));
#else

			if (nnp[i] == lightsrcno) {
				zp_pplus(&bmat[i][0], &duplex_unit);
				printf("%5d\n", elemno);
			}
#endif

		}
	}
	if ((elemno % 1000) == 0)
		printf("%d\n", elemno);
	return;
}


struct duplex_t
n_f(int subdo, int elemno, int intgno)
{
	struct duplex_t ret;
	double	tmp;
	if (n_substr == 0.0 || nr == 0.0)
		fatal("%s(): ", __F);

	if (subdo == SD_SUBSTRATE || subdo == SD_NEAR)
		ret = duplexv(n_substr, 0.0);    // substrate
	else if (subdo == SD_RESIST) {
		tmp = dgetat(elemno, intgno, m_field);
		ret = duplexv(nr, glA * tmp + glB) ;
	} else
		ret = duplexv(1.0, 0.0);     // vacuum
	return ret;
}


void
imposeCondBC(int no)
{
	struct duplex_t dux;

	dux = zvec(pure_light)[no];
	zp_crossdme(&dux, -1.0);
	di_zimpose(no, dux);
	return;
}




int
whichsubdo(const struct ement_t *ep)
{
	struct proto_t *proto;
	struct node_t *np;
	nodno_t *nnp;
	int	noofnodes;
	float	means[3];
	int	i, j, flag;
	double	x, y;

	proto = ep->e_proto;
	nnp = ep->e_nodes;
	noofnodes = ep->e_nodesno;

	for (i=0;i<meshes->m_spacedim;i++)
		means[i] = 0.0; /* initialization */

	for (i=0;i<proto->p_nodsno;i++) {
		np = meshes->m_nodes[nnp[i]]; /* for surrounding nodes*/
		for (j=0; j < meshes->m_spacedim; j++)
			means[j] += np->n_coor[j];
	}

	for (i=0;i < meshes->m_spacedim;i++)
		means[i] /= (float) noofnodes; /* initialization */

	flag = 0;
	x = means[0];
	y = means[1];

	if (x <= XL_MARGIN)
		return SD_LEFT;
		/* flag |= F_LEFT; */
	if (x >= XR_MARGIN)
		return SD_RIGHT;
		/* flag |= F_RIGHT; */

	if (y <= YL_MARGIN)
		return SD_NEAR;
		/* flag |= F_NEAR; */
	if (y >= YR_MARGIN)
		return SD_FAR;
		/* flag |= F_FAR; */

	switch (flag) {
		case F_LEFT:	return SD_LEFT; break;
		case F_RIGHT:	return SD_RIGHT; break;
		case F_NEAR :	return SD_NEAR; break;
		case F_FAR:	return SD_FAR; break;
		case F_LEFT | F_NEAR:
			if (x < y)
				return SD_LEFT;
			else
				return SD_NEAR;
			break;
		case F_RIGHT | F_NEAR:
			if ((X_LEN-x) < y)
				return SD_RIGHT;
			else
				return SD_NEAR;
			break;
		case F_LEFT | F_FAR:
			if (x > (Y_LEN-y))
				return SD_FAR;
			else
				return SD_LEFT;
			break;
		case F_RIGHT | F_FAR:
			if ((X_LEN-x) > (Y_LEN-y))
				return SD_FAR;
			else
				return SD_RIGHT;
			break;
	}
	if (y < 0.5) {
		resist0 = 0.5;
		return SD_SUBSTRATE;
	}
	if (y < 1.0) {
		resist1 = 1.0;
		return SD_RESIST;
	}


	return 0;
}

int
whichbound(const struct node_t *np)
{
	float	x, y;

	x = np->n_coor[0];
	y = np->n_coor[1];

	return -1;
}



void
impose_bc(void)
{
	zdirichlet(0, duplex_null);
	zdirichlet(1, duplex_null);
	zdirichlet(2, duplex_null);
	zdirichlet(3, duplex_null);
	return;
}


/*
 * Following does not multiply with cosine
 */

double
get_beta(const double order, const double delta, const double totlen)
{
	double	retval;
	double	factor;

	factor = M_LN10 * order / (totlen * wave_k); 	/* XXX */

	retval = (delta/totlen);
	retval = pow(retval, 1.0);
	retval *= (1.0 + 1.0) * factor;
	retval *= -1.0;

	return retval;
}

void
set_const3rd(const double order)
{
	int	i, j, subdomain;
	struct node_t *np;
	nodno_t	*nnp;

	double	xcoor, ycoor, delta;
	double	sq, howmany;
	struct duplex_t ztmp0, ztmp1;
	struct duplex_t *muxp, *muyp, *epsp;

	printf("start of %s\n", __F);

	mux = vector(meshes->m_nodesno, "mux", SP_DUPLEX);
	muy = vector(meshes->m_nodesno, "muy", SP_DUPLEX);
	epsilon = vector(meshes->m_nodesno, "epsilon", SP_DUPLEX);

	muxp = zvec(mux);
	muyp = zvec(muy);
	epsp = zvec(epsilon);

	for (i=0; i < meshes->m_nodesno;i++) {
		np = meshes->m_nodes[i];
		xcoor = np->n_coor[0];
		ycoor = np->n_coor[1];
		nnp = np->n_ements;

		for (j=0; j < np->n_ementsno; j++) {
			subdomain = meshes->m_ements[nnp[j]]->e_sdomain;
			switch (subdomain) {
				case SD_LEFT:
				case SD_RIGHT:
					if (subdomain == SD_LEFT)
						delta = fabs(xcoor-XL_MARGIN);
					else
						delta = fabs(xcoor-XR_MARGIN);

					sq = get_beta(order, delta, X_PMLWIDTH);

					ztmp0 = duplexv(1.0, sq);
					zp_inverse(&ztmp1, &ztmp0);

					zp_pplus(&muxp[i], &ztmp1);
					zp_pplus(&muyp[i], &ztmp0);
					zp_pplus(&epsp[i], &ztmp0);
					break;
				case SD_NEAR:
				case SD_FAR:
					if (subdomain == SD_NEAR)
						delta = fabs(ycoor-YL_MARGIN);
					else
						delta = fabs(ycoor-YR_MARGIN);

					sq = get_beta(order, delta, Y_PMLWIDTH);

					ztmp0 = duplexv(1.0, sq);
					zp_inverse(&ztmp1, &ztmp0);

					zp_pplus(&muxp[i], &ztmp0);
					zp_pplus(&muyp[i], &ztmp1);
					zp_pplus(&epsp[i], &ztmp0);
					break;
				case SD_SUBSTRATE:
					ztmp0 = duplexv(1.0, 0.0); /* XXX */
					zp_pplus(&muxp[i], &ztmp0);
					zp_pplus(&muyp[i], &ztmp0);
					zp_pplus(&epsp[i], &ztmp0);
					break;
				case SD_RESIST:
					ztmp0 = duplexv(1.0, 0.0);
					zp_pplus(&muxp[i], &ztmp0);
					zp_pplus(&muyp[i], &ztmp0);
					zp_pplus(&epsp[i], &ztmp0);
					break;
				default:
					ztmp0 = duplexv(1.0, 0.0);
					zp_pplus(&muxp[i], &ztmp0);
					zp_pplus(&muyp[i], &ztmp0);
					zp_pplus(&epsp[i], &ztmp0);
					break;
			}
		}
		howmany = 1.0/(double) j;

		zp_crossdme(&muxp[i], howmany);
		zp_crossdme(&muyp[i], howmany);
		zp_crossdme(&epsp[i], howmany);
	}

	printf("end of %s\n", __F);
	return;
}

void
putmsg(const char *msg, const int lineno)
{
	static int no;
	fprintf(stderr, "[%2d]: %s(): %5dth \n", no++, msg, lineno);
	return;
}

void
vec_putresist(const char *filename, const struct vector_t *vp)
{
	static float offset;
	static int filecnt = -1; /* XXX */
	static int firstcalled = 0;
	float	resistwidth, yoffset;
	FILE	*fp;
	int	i, j, whethernewone;
	nodno_t	*nnp;
	double	*ddata;
	struct node_t *np;
	struct ement_t *ep;
	char	*headers[]= {"$ DATA=CONTCURVE",
		"% contstyle=2  # 1=contour lines, 2=colors, 3=mesh",
		"% nsteps=50",
		"% cmin=0",
		"#% contours= 0.1 0.3 0.35  # each line value is specified",
		"#% comment=\"A comment in the upper right corner\"",
		"#% eyepos.x=0.5 eyepos.y=-1.5 eyepos.z=0.5 # another 3D view",
		NULL};
	char	**cpp, *p, fname[128];

	resistwidth = resist1 - resist0;
	offset = (resistwidth + 0.05) * (float) firstcalled;
	whethernewone = ((firstcalled % 5) == 0);

	if (whethernewone)
		filecnt++;

	if (vp->vt_size != meshes->m_nodesno)
		fatal("%s(): [0]", __F);

	p = rindex(filename, '.');
	if (p != NULL)
		*p = 0;
	sprintf(fname, "%s.%02d.grids.mtv", filename, filecnt);

	if (whethernewone)
		fp = fopen(fname, "w");
	else
		fp = fopen(fname, "a");

	if (fp == NULL)
		fatal("%s(\"%s\"): can not fopen", __F, fname);

	if (whethernewone) {
		cpp = headers;

		while (*cpp)
			fprintf(fp, "%s\n", *cpp++);
	}

	fprintf(fp, "\n");

	ddata =  dvec(vp);

	for (i= resist_on;i < resist_off;i++) {
		ep = meshes->m_ements[i];
		nnp = ep->e_nodes;
		for (j=0;j < ep->e_nodesno; j++) {
			np = meshes->m_nodes[nnp[j]];
			yoffset = (np->n_coor[1] - resist0);
			yoffset += offset;

			fprintf(fp, "%10.3e %10.3e %10.3e\n",
				np->n_coor[0], yoffset,
				ddata[nnp[j]]);
		}
		fprintf(fp, "\n\n");
	}
	fclose(fp);
	firstcalled++; /* XXX */
	return;
}

int
insideunit(const double x)
{
	if (x <= 1 && x >= -1)
		return 1;
	return 0;
}

struct duplex_t
exponent(const double x)
{
	struct duplex_t ret;

	ret.real = cos(x);
	ret.imag = sin(x);

	return ret;
}

struct vector_t *
get_sigmas(const char *fname)
{
	FILE	*fp;
	char	lines[256];
	int	readn, cnt;
	float	flt;
	double	*dp;
	struct vector_t *ret;

	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("%s(): [0]", __F);

	cnt = 0;
	while (fgets(lines, 256, fp)) {
		if (*lines != '#')
			cnt++;
	}
	fclose(fp);
	n_lightno = cnt;

	ret = vector(cnt, "sigmas", SP_DOUBLE);
	dp = dvec(ret);

	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("%s(): [0]", __F);

	cnt = 0;
	while (fgets(lines, 256, fp)) {
		if (*lines == '#')
			continue;
		readn = sscanf(lines, "%f", &flt);
		if (readn != 1)
			fatal("%s(): sscanf()", __F);
		printf("%s(): [%2d]: %7.3f\n", __F, cnt, flt);
		dp[cnt++] = flt;
	}
	if (cnt != n_lightno)
		fatal("%s(): [1]", __F);
	fclose(fp);
	return ret;
}
