#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <ctype.h>
#include "common.h"
#include "finite.h"

/*
 * In order to read mesh in diffpack format
 */
enum stage_t {
	BEFORE_START = 0,
	READING_NODE,
	/*
	BEFORE_ELE,
	 */
	READING_ELE,
	END
};

struct bound_t **m_bound;
int	m_bdindno;

static int	getbdint(char *line);
#if 0
static struct	bound_t *getbound(char *);
#endif
static int	hackline(char *, int);
static int	getint(int, char *, char);
static void	checkmesh(void);
static int	eachnode(char *line);
static int	eachelem(char *line);

/*
 * read the nodes and emement information that is stored at the char *fname
 * file.  The file should have the format of the .grid format of the well-known
 * diffpack package.  Therefore if you want to use the grids information which
 * was produced by piffpack, you can use this function.
 */
int
dp_getmesh(const char *fname)
{
	FILE	*fp;
	char	line[BUFSIZ];
	int	stage=0;
	static	int	isbdind=0;
	int	noderead=0;
	int	elemread=0;

	if (meshes == NULL)
		fatal("%s(%s): You should called mesh_init before!!",
			__F, fname);

	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("%s():input mesh fatal error : fopen(\"%s\")!!",
			__F, fname);

	while (fgets(line, BUFSIZ-1, fp)) {
		if (!islineok(line))
			continue;	/* nothing in this line */
		if (*line == '#') {	/* change status */
			switch(stage) {
				case BEFORE_START:	/* mesh check */
					checkmesh();
					/* malloc space of nodes */
					meshes->m_nodes = mesh_nodesbuf
						(meshes->m_nodesno);
					break;
				case READING_NODE:	/* nodes check */
					if (noderead != meshes->m_nodesno)
						fatal("You gave me insuff. no");
					/* malloc space of elements */
					meshes->m_ements = mesh_ementsbuf(
						meshes->m_ementsno);
					break;
				/*
				case BEFORE_ELE:
					break;
				 */
				case READING_ELE:
					break;
				case END:
					if (elemread != meshes->m_ementsno)
						fatal("You gave me insuff.elm");
					return 0;
					break;
				default:
					return 0;
			}
			stage++;
			continue;	/* to skip this line */
		}
		switch (stage) {
			case BEFORE_START:
				isbdind = hackline(line, isbdind);
				break;
			case READING_NODE:	/* nodes check */
				noderead = eachnode(line);
				break;
			/*
			case BEFORE_ELE:
				break;
			 */
			case READING_ELE:
				elemread = eachelem(line);
				if (elemread == meshes->m_ementsno)
					stage++;
				break;
			default:
				mk_nearements();
				return 0;
		}
	}
	mk_nearements();
	return 0;
}


enum bd_type {
	M_SPACE,
	M_ENO,
	M_NODENO,
	M_SAMETYPE,
	M_MAXNONO,
	M_ONESD,
	M_LATTICE,
	M_BOUN
};

struct hack {
	int	h_no;
	int	h_type;
	char	*h_name;
};
struct hack dirty[9] = {
	{ 0,	M_SPACE,	"Number of space dim"},
	{ 1,	M_ENO,		"Number of elements"},
	{ 2,	M_NODENO,	"Number of nodes"},
#if COMMENTED_OUT
	{ 3,	M_SAMETYPE,	"same type"},
	{ 4,	M_MAXNONO,	"Max number of nodes"},
	{ 5,	M_ONESD,	"one subdomain"},
	{ 6,	M_LATTICE,	"Lattice"},
#endif
	{ 7,	M_BOUN,		"Boundary ind"},
	{ 8,	0, NULL},
};

static int
hackline(char *line, int bound)
{
	struct hack *hp;
	int	retval = 0;
	struct bound_t	**bdpp;/* *bp; */
	static int	bdcnt = -1;

	if (bound) {
		if (m_bdindno < 0) /* || m_bound == NULL) */
			fatal("What's wrong!! in hackline()");
#if 0
		bp = getbound(line);
		bp->b_index = bdcnt;
		m_bound[bdcnt] = bp;
#endif
		bdcnt++;
		if (bdcnt == m_bdindno)
			return 0;	/* no more boundary indicator */
		return 1;
	}

	hp = (struct hack *) &dirty[0];
	for (; hp->h_name && !strstr(line, hp->h_name);)
		 hp++;

	if (hp->h_name == NULL)
		return 0;
		
	switch (hp->h_type) {
		case M_SPACE:
			meshes->m_spacedim = getint(hp->h_no, line, '=');
			break;
		case M_ENO:
			meshes->m_ementsno = getint(hp->h_no, line, '=');
			break;
		case M_NODENO:
			meshes->m_nodesno = getint(hp->h_no, line, '=');
			break;
#if COMMENTED_OUT
		case M_SAMETYPE:
			meshes->m_sametype = getint(hp->h_no, line, ':');
			break;
		case M_MAXNONO:
			meshes->m_maxnode = getint(hp->h_no, line, ':');
			break;
		case M_ONESD:
			meshes->m_onesub = getint(hp->h_no, line, ':');
			break;
		case M_LATTICE:
			meshes->m_isperiodic = getint(hp->h_no, line, '?');
			break;
#endif
		case M_BOUN:
			retval = 1;
			m_bdindno = getbdint(line);
			bdpp = (struct bound_t **)
				malloc((size_t)(sizeof(struct bound_t *)
					* m_bdindno));
			if (bdpp == NULL)
				memerr("error in alloc() the m_bound");
			m_bound = bdpp;
			bdcnt++;
			break;
		default:
			fatal("This is impossible!!");
	}
	return retval;
}

static int
getint(num, line, indicator)
	int	num;
	char	*line;
	char	indicator;
{
	char	*p;
	int	no;

	p = index(line, indicator);
	if (p == NULL)
		fatal("%s() error[0] in reading %s.", __F, dirty[num].h_name);
	
	for (p++; isspace(*p); p++) /* skip white space */
		/* void */ ;

	if (p == NULL)
		fatal("%s() error[1] in reading %s.", __F, dirty[num].h_name);
	no = atoi(p);
	if (no < 0)
		fatal("%s() error[2] in reading %s.", __F, dirty[num].h_name);
	return no;
}

static int
getbdint(line)
	char *line;
{
	char	*p;
	int	num;

	for (p=line; isspace(*p); p++) /* skip white space */
		/* void */;

	if (p == NULL)
		fatal("%s(): This is not possible!!!", __F);

	num = atoi(p);
	if (num < 0 || num > 31)
		fatal("%s(): negative boundary indicator number??", __F);
	return num;
}

#define PERM(c) (isalnum(c) || index("+-=_", c))

#if COMMENT
static struct bound_t *
getbound(char *line)
{
	char *p, *ret, *tmp;
	struct bound_t	*bt;

	for (p=line; isspace(*p); p++) /* skip the white space */
		/* void */ ;

	if (p == NULL)
		fatal("error in the indicator name");
	tmp = p;
	for (; PERM(*p); p++) /* skip the u space */
		/* void */ ;

	*p = 0;	/* terminate */

	ret = strdup(tmp);
	if (ret == NULL)
		memerr("%s(): strduping %s", __F,  tmp);

	bt = (struct bound_t *) malloc((size_t)(sizeof(struct bound_t)));
	if (bt == NULL)
		memerr("%s(): err at malloc bound_t", __F);
	
	bt->b_name = ret;
	return bt;
}
#endif

#undef PERM


static void
checkmesh(void)
{
	struct mesh *mp;

	mp = meshes;
	if (mp->m_spacedim < 0 || mp->m_ementsno < 0
		|| mp->m_nodesno < 0
#if COMMENTED_OUT
		|| mp->m_sametype < 0
		|| mp->m_onesub < 0
			/* || mp->m_isperiodic < 0 */
		|| mp->m_bdindno < 0
		|| mp->m_bound ==NULL
#endif
		|| mp->m_nodes != NULL
		|| mp->m_ements != NULL) {
		fatal("Uh...");
	}
}


static int
eachnode(char *line)
{
	static int nodcnt = 0;
	struct node_t *np;
	float	tfloat;
	int	nodeno;
	int	indno;	/* indicator number */
	int	ind;
	char	*p, *q;
	int	tmp, i;
	int	totind;
	int	spacedim;

	if (nodcnt >= meshes->m_nodesno)
		return nodcnt;
	totind = m_bdindno;

	for (p = line; isspace(*p); p++)	/* skip white spaces */
		/* void */;
	nodeno = atoi(p);	/* if failed, return atoi() returns zero */
	if (nodeno == 0)
		fatal("%s(): Nodeno is not good!!", __F);

	nodeno--;	/* index starts with 0 */
	np = meshes->m_nodes[nodeno];
	np->n_no = nodeno;

	p = skipto(p, '(');
	p++;

	
	spacedim = meshes->m_spacedim;
	for (i = 0; i < 3; i++)	/* clear its coors */
		np->n_coor[i] = 0.0;

	for (i = 0; i < spacedim; i++) {
		q = strpbrk(p, ",)");	/* any of theses , and ) */
		if (q == NULL)
			fatal("%s(): reading coordinates", __F);
		*q = 0;
		tmp = sscanf(p, "%e", &tfloat);
		np->n_coor[i] = tfloat;
		if (tmp - 1)
			fatal("%s(): reading coordinates of node %d",
				__F, nodeno);
		p = ++q;
	}

	/*
	 * End of reading of coordinates
	 */

	p = skipto(p, '[');
	if (p == NULL)
		fatal("%s(): Huhuhu... wrong format of boundary indi..", __F);
	p++;

	tmp = sscanf(p, "%d", &indno);
	if (tmp - 1)
		fatal("%s(): Huhuhu... wrong format of boundary indi..", __F);
#if 0
	np->n_bdindno = indno;

	if (indno > totind)
#endif
	if (indno > 31)
		fatal("%s(): greater than the given indi..", __F);

	np->n_bdflag = 0;

	for (i = 0; i< indno; i++) {
		for (; !isspace(*p) && *p; p++) /* skip white spaces */
			/* void */ ;
		for (; !isdigit(*p) && *p; p++)	/* skip to a number */
			/* void */ ;
		tmp = sscanf(p, "%d", &ind);
		if (tmp - 1)
			fatal("Wrong format in bd indi..");
		ind--;
		if (ind < 0 || ind >= 31)
			fatal("%s(): less than 0 or greater than the max ind",
				__F);
		setbit(np->n_bdflag, ind);
	}
	if (i != indno)
		fatal("%s(): node indicater is not sufficient!!", __F);
	nodcnt++;
	return nodcnt;
}

static int
eachelem(char *line)
{
	static int elemcnt=0;
	int	i;
	struct ement_t *ep;
	char	*from;
	char	*p;
	int	eno;
	int	subdomain;
	int	tmp;
	int	gnode;
	nodno_t	*glob;

	if (elemcnt >= meshes->m_ementsno)
		return elemcnt;

	for (p=line; isspace(*p) && *p; p++) /* skip white spaces */
		/* void */ ;

	if (*p == NULL)
		fatal("IMPOSSIBLE[0] in eachelem()");
	tmp = sscanf(p, "%d", &eno);
	if ((tmp - 1) || eno > meshes->m_ementsno)
		fatal("%s(): Wrong element number in eachelem()", __F);

	eno--;	/* index starts at 0 */
	ep = meshes->m_ements[eno];
	ep->e_no = eno;

	for (;isdigit(*p); p++)	/* skip element no */
		/* void */;

	p = skipwhites(p);
	from = p;
	for (; TOKEN(*p) && *p; p++)	/* skip element type string */
		/* void */ ;
	if (*p == NULL)
		fatal("%s(): fatal error getting prototypes in ele %d",
			__F, (eno+1));
	*p = 0;
	p++;
	ep->e_proto = getproto(from);

	p = skipwhites(p);

	tmp = sscanf(p, "%d", &subdomain);
	if (tmp - 1)
		fatal("%s(): Uhhh...in reading subdomain in eachelem", __F);

	ep->e_sdomain = subdomain;
#if XXX
	subdomain--;	/* index starts in 0 */
#endif

	for (; TOKEN(*p) && *p; p++)	/* subdomain */
		/* void */ ;

	glob = (nodno_t *)
		malloc((size_t)(sizeof(nodno_t) * ep->e_proto->p_nodsno));
	if (glob == NULL)
		memerr("%s(): mallocing glob correspondants", __F);

	ep->e_nodesno = ep->e_proto->p_nodsno;
	ep->e_nodes = glob;
	for (i=0; i< ep->e_proto->p_nodsno; i++) {
		p = skipwhites(p);
		tmp = sscanf(p, "%d", &gnode);
		if (tmp != 1)
			fatal("%s(): reading global correspon..", __F);
		gnode--;	/* node numbering starts at 0 */
		ep->e_nodes[i] = gnode;
		p = skipto(p, ' '); /* nodes should be separated by space */
	}
	elemcnt++;
	return elemcnt;
}
