#include <stdio.h>
#include <string.h>
#include <strings.h>
#ifdef sgi
#include <bstring.h>	/* for Irix5.2 */
#endif
#include <stdlib.h>
#include <ctype.h>

#include "common.h"
#include "finite.h"

int	dp_putheaders(FILE *fp);
int	dp_putnodes(FILE *);
int	dp_putements(FILE *);

/*
 * Puts the grid information at diffpack format. Remember that the format
 * produced is not exact format, so that I do not know that the diffpack program
 * can handle this produced file.
 */
void
dp_putmesh(const char *filename)
{
	FILE	*fp;

	if (filename == NULL)
		return;

	fp = fopen(filename, "w");
	dp_putheaders(fp);
	dp_putnodes(fp);
	dp_putements(fp);
	fclose(fp);

	return;
}


int
dp_putheaders(FILE *fp)
{
	fprintf(fp, " Number of space dim. = %2d\n", meshes->m_spacedim);
	fprintf(fp, " Number of elements  = %5d\n", meshes->m_ementsno);
	fprintf(fp, " Number of nodes = %5d\n", meshes->m_nodesno);
	fprintf(fp, " All elements are of the same type : %s\n", "dpFALSE");
	/*
	 * I dont care..
	 */
	fprintf(fp, " Max number of nodes in an element : %d\n", 100);
#if 0
	fprintf(fp, " Only one subdomain : %s\n",
		meshes->m_onesub ? "dpTRUE" : "dpFALSE");
#endif
	fprintf(fp, " is this mesh a lattice?\n");
	fprintf(fp, " 0 Boundary indicators:\n");
	fprintf(fp, "\n\n#\n");
	return 0;
}
void
dp_bdflag(int bd, char *buf)
{
	char	name[64], small[16];
	int	i, j, flag;
	int	cnt;

	flag = bd;
	*buf = 0;
	*name = 0;
	cnt = 0;
	for (i=0, j = 1; i < 32; i++, j <<= 1) {
		if (bd & j) {
			sprintf(small, "%d ", (i+1));
			strcat(name, small);
			cnt++;
		}
	}
	sprintf(buf, "[%d] ", cnt);
	if (cnt)
		strcat(buf, name);
	return;
}

int
dp_putnodes(FILE *fp)
{
	int	i;
	struct node_t	**npp;
	char	buf[128];

	npp = meshes->m_nodes;
	if (npp == NULL)
		fatal("Really fatal error in dp_putnodes()");
	for (i = 0; i< meshes->m_nodesno; i++) {
		(void) fprintf(fp, "%4d ", (npp[i]->n_no+1));
		if (meshes->m_spacedim == 3)
			(void) fprintf(fp, "(%8.4f,%8.4f,%8.4f) ",
				npp[i]->n_coor[0],
				npp[i]->n_coor[1],
				npp[i]->n_coor[2]);
		else if (meshes->m_spacedim == 2)
			(void) fprintf(fp, "(%8.4f,%8.4f)    ",
				npp[i]->n_coor[0],
				npp[i]->n_coor[1]);

		dp_bdflag(npp[i]->n_bdflag, buf);
#if 0
		(void) bitput(npp[i]->n_bdflag, (char *) buf);
		(void) fprintf(fp, "[%d] %s", npp[i]->n_bdindno, buf);
#endif
		(void) fprintf(fp, "%s\n", buf);
	}
	(void) fprintf(fp, "\n#\n");
	return 0;
}

int
dp_putements(FILE *fp)
{
	int	i, j;
	struct ement_t	**epp;

	epp = meshes->m_ements;
	if (epp == NULL)
		fatal("Really fatal error in dp_putements()");
	for (i = 0; i< meshes->m_ementsno; i++) {
		(void) fprintf(fp, "%4d ", (epp[i]->e_no+1));
		(void) fprintf(fp, "%s ", epp[i]->e_proto->p_name);
		(void) fprintf(fp, "%4d ", (epp[i]->e_sdomain+1));
		for (j=0; j < epp[i]->e_proto->p_nodsno; j++)
			(void) fprintf(fp, "%5d ", (epp[i]->e_nodes[j]+1));
		(void) fprintf(fp, "\n");
	}
	(void) fprintf(fp, "\n\n");
	return 0;
}
