/*
 * generate a mesh
 */

#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include "common.h"
#include "finite.h"

void	linear_2dim(const char *geometry, const char *partition);
void	linear_3dim(const char *geometry, const char *partition);
void	linear_2process(struct genmesh_t *gp);
void	linear_3process(struct genmesh_t *gp);

void
genm_linear(const char *geo, const char *part)
{
	char	*p;
	int	dim, readn;

	printf("+++++++++++++START of %s()+++++++++++++++++\n", __F);
	if (meshes == NULL)
		fatal("%s(\"%s\", \"%s\"): you must call mesh_init first",
			geo, part);
	if (meshes->m_nodes != NULL || meshes->m_ements != NULL) {
		fprintf(stderr, "%s():you can not call %s() 'cause ", __F, __F);
		fprintf(stderr, "you already have meshes->m_nodes structure\n");
		exit(1);
	}

	p = index(geo, '=');
	if (p == NULL)
		fatal("%s(\"%s\", b): not good typing", __F,  geo);

	readn = sscanf(++p, "%d", &dim);
	if (readn != 1 || dim < 2 || dim > 3)
		fatal("%s(\"%s\", b): not good typing", __F,  geo);

	switch (dim) {
		case 2:
			meshes->m_spacedim = 2;
			linear_2dim(geo, part);
			break;
		case 3:
			meshes->m_spacedim = 3;
			linear_3dim(geo, part);
			break;
	}
	mk_nearements();
	printf("+++++++++++++END   of %s()+++++++++++++++++\n", __F);
	return;
}

/*
 * generate 2-dimensional meshes
 * you should NOT call this function DIRECTLY, because it is suffice to call
 * genmesh(), so that this function is a sort of a private function of
 * genmesh().
 */
void linear_3dim(const char *geo, const char *par)
{
	char	*geom, *part;
	char	*p;
	struct genmesh_t genm;
	int	i, readn;

	geom = strdup(geo);	/* please free me */
	part = strdup(par);

	if (geom == NULL || part == NULL)
		memerr("%s(): at strdup()ing", __F);

	p = index(geom, '[');	/* dont need to check p == NULL */
	for (i=0; i<3;i++) {
		readn = sscanf(p++, "[%f:%f]",
			&genm.gn_mins[i], &genm.gn_maxs[i]);
		if (readn != 2)
			fatal("linear_3dim(\"%s\", b): wrong typing", __F,geo);
		p = index(p, '[');
	}

	/*
	 * manipulate the partition
	 */
	p = index(part, '=');
	if (p == NULL)
		fatal("%s(\"%s\", \"%s\"): not good 2nd argument",__F,geo, par);

	readn = sscanf(++p, "%d", &i);
	if (i != 3)
		fatal("%s(\"%s\", \"%s\"): two have different dim", __F,
			geo, par);

	p = index(part, '[');
	if (p == NULL)
		fatal("%s(\"%s\", \"%s\"): bad 2nd argum", __F, geo, par);

	readn = sscanf(p, "[%d:%d:%d]",
		&genm.gn_steps[0], &genm.gn_steps[1], &genm.gn_steps[2]);
	if (readn != 3)
		fatal("%s(\"%s\", \"%s\"): bad 2nd argum", __F, geo, par);

	if (genm.gn_steps[0] < 1 || genm.gn_steps[1] < 1 ||genm.gn_steps[2] < 1)
		fatal("%s(\"%s\", \"%s\"): bad steps", __F, geo, par);

	/*
	 * read gradings
	 */
	p++;
	p = index(p, '[');
	readn = sscanf(p, "[%f:%f:%f]",
		&genm.gn_grades[0], &genm.gn_grades[1], &genm.gn_grades[2]);
	if (readn != 3)
		fatal("%s(\"%s\", \"%s\"): bad grading", __F, geo, par);

	/*
	 * The end of reading of geometry and partitions
	 */
	linear_3process(&genm);

	free((void *) geom);
	free((void *) part);
	return;
}

void
linear_2dim(const char *geo, const char *par)
{
	char	*geom, *part;
	char	*p;
	struct genmesh_t genm;
	int	i, readn;

	geom = strdup(geo);	/* please free me */
	part = strdup(par);
	if (geom == NULL || part == NULL)
		memerr("%s(): at strdup()ing", __F);

	p = index(geom, '[');	/* dont need to check p == NULL */
	for (i=0; i < 2;i++) {
		readn = sscanf(p++, "[%f:%f]",
			&genm.gn_mins[i], &genm.gn_maxs[i]);
		if (readn != 2)
			fatal("%s(\"%s\", b): wrong typing", __F, geo);
		p = index(geom, 'x');
		p = index(p, '[');
	}

	p = index(part, '=');
	if (p == NULL)
	fatal("%s(\"%s\", \"%s\"): not good 2nd argument", __F, geo, par);

	readn = sscanf(++p, "%d", &i);
	if (i != 2)
		fatal("%s(\"%s\", \"%s\"): two have different dim", __F,
			geo, par);

	p = index(part, '[');
	if (p == NULL)
		fatal("%s(\"%s\", \"%s\"): bad 2nd argum", __F, geo, par);

	readn = sscanf(p, "[%d:%d]", &genm.gn_steps[0], &genm.gn_steps[1]);
	if (readn != 2)
		fatal("%s(\"%s\", \"%s\"): bad 2nd argum", __F, geo, par);

	if (genm.gn_steps[0] < 1 || genm.gn_steps[1] < 1)
		fatal("%s(\"%s\", \"%s\"): bad steps", __F, geo, par);

	/*
	 * read gradings
	 */
	p++;
	p = index(p, '[');
	readn = sscanf(p, "[%f:%f]", &genm.gn_grades[0], &genm.gn_grades[1]);
	if (readn != 2)
		fatal("%s(\"%s\", \"%s\"): bad grading", __F, geo, par);

	/*
	 * The end of reading of geometry and partitions
	 */
	linear_2process(&genm);
	free((void *) geom);
	free((void *) part);
	return;
}

/*
 * Do not call this function explicitly, 'cause this is a sort of private
 * function of linear_2dim();
 */
void
linear_2process(struct genmesh_t *gp)
{
	int	i, j, total, cnt, steps[2];

	float	mins[2], maxs[2], units[2], grading[2], plus[2], minus[2];
	float	x, y, x1, y1, var;
	nodno_t *nnp;
	struct ement_t *ep;
	int	ix, indx, xchecked=1;
	float	*xcoor, *ycoor;
	ushort	*xminnodes, *yminnodes;

	total = (gp->gn_steps[0] + 1) * (gp->gn_steps[1] + 1);

	/*
	 * To facilate the calculation of...
	 */
	m_for3dim = (struct for3dimcase *)
		malloc(sizeof(struct for3dimcase));

	for (i=0; i < 2; i++) {
		m_for3dim->f3d_howmany[i] = (gp->gn_steps[i] + 1);
		m_for3dim->f3d_nonodes[i] = (gp->gn_steps[(i+1)%2]+1);
		m_for3dim->f3d_coors[i] = (float *)
		    malloc(sizeof(float)*m_for3dim->f3d_howmany[i]);
		m_for3dim->f3d_minnodes[i] = (ushort *)
		    malloc(sizeof(ushort)*m_for3dim->f3d_howmany[i]);
	}

	m_for3dim->f3d_nincres[0] =  1;
	m_for3dim->f3d_nincres[1] =  m_for3dim->f3d_howmany[0];

	m_for3dim->f3d_eincres[0] =  1;
	m_for3dim->f3d_eincres[1] =
		(m_for3dim->f3d_howmany[0] - 1);

	/*
	 * real allocation of the node information
	 */
	meshes->m_nodes = mesh_nodesbuf(total);

	for (i=0;i<2;i++) {
		grading[i]= gp->gn_grades[i];
		steps[i]  = gp->gn_steps[i];
		mins[i]   = gp->gn_mins[i];
		maxs[i]   = gp->gn_maxs[i];
		plus[i]   = maxs[i] + mins[i];
		minus[i]  = maxs[i] - mins[i];
		units[i]  = minus[i] / (float) steps[i];
	}

	xcoor = m_for3dim->f3d_coors[0];
	ycoor = m_for3dim->f3d_coors[1];
	xminnodes = m_for3dim->f3d_minnodes[0];
	yminnodes = m_for3dim->f3d_minnodes[1];

	cnt = 0;
	xchecked = 1;
	for (i=0; i <= gp->gn_steps[1]; i++) { /* Y */
		y = mins[1] + units[1] * (float) i;
		var = 2.0 * y - plus[1];
		y1 = 0.5 * sign(var) * minus[1]
			* pow(fabs(var/minus[1]), 1.0/grading[1])
			+ 0.5 * plus[1];

		ycoor[i] = y1;
		yminnodes[i] = cnt;

		if (i == 0)
			m_coormin[1] = y1;
		if (i == gp->gn_steps[1])
			m_coormax[1] = y1;

		for (j=0; j <= gp->gn_steps[0]; j++) { /* X */
			x = mins[0] + units[0] * (float) j;
			var = 2.0 * x - plus[0];
			x1 = 0.5 * sign(var) * minus[0]
				* pow(fabs(var/minus[0]), 1.0/grading[0])
				+ 0.5 * plus[0];
			if (xchecked) {
				xcoor[j] = x1;
				xminnodes[j] = cnt;

				if (j == 0)
					m_coormin[0] = x1;
				if (j == gp->gn_steps[0])
					m_coormax[0] = x1;
			}

			meshes->m_nodes[cnt]->n_coor[0] = x1;
			meshes->m_nodes[cnt]->n_coor[1] = y1;

			meshes->m_nodes[cnt]->n_plane[0] = j;	/* X */
			meshes->m_nodes[cnt]->n_plane[1] = i;	/* Y */

			meshes->m_nodes[cnt]->n_no = cnt;

			/*
			 * boundary manipulations
			 *      |
			 *      | y  2 
			 *      +---------+
			 *      |         |
			 *      |         |
			 *    3 |         |  1
			 *      |         |
			 *      |         |
			 *      +---------+----> x
			 *           0
			 */
			if (i == 0)
				setbit(meshes->m_nodes[cnt]->n_bdflag, 0);
			else if (i == gp->gn_steps[1])
				setbit(meshes->m_nodes[cnt]->n_bdflag, 2);
			if (j == 0)
				setbit(meshes->m_nodes[cnt]->n_bdflag, 3);
			else if (j == gp->gn_steps[0])
				setbit(meshes->m_nodes[cnt]->n_bdflag, 1);

			/* XXX */
			cnt++;
		}
		xchecked = 0;
	}
	meshes->m_nodesno = cnt;

	/*
	 * start of ement structure
	 */

	total = (gp->gn_steps[0]) * (gp->gn_steps[1]);
	meshes->m_ements = mesh_ementsbuf(total);

	cnt = 0;
	for (i=0; i < gp->gn_steps[1]; i++) {
		ix = i * (gp->gn_steps[0]+1);

		for (j=0; j < gp->gn_steps[0]; j++) {
			indx = ix + j;
			ep = meshes->m_ements[cnt];
			ep->e_proto = getproto("ElmB4n2D");
			ep->e_nodesno = 4;
			ep->e_no = cnt;
			nnp = ep->e_nodes = (nodno_t *)
				malloc(sizeof(nodno_t) * 4);
			if (ep->e_nodes == NULL)
				fatal("linear_2process(): Hmm");
			*nnp++ = indx;
			*nnp++ = indx + 1;
			*nnp++ = indx + (gp->gn_steps[0]+1);
			*nnp++ = indx + (gp->gn_steps[0]+1) + 1; /* XXX */

			cnt++;
		}
	}
	meshes->m_ementsno = cnt;

	return;
}


/*
 * for 3-dimensional case
 *
 * Do not call this function explicitly, 'cause this is a sort of private
 * function of linear_2dim();
 */
void
linear_3process(struct genmesh_t *gp)
{
	int	i, j, k, total, cnt, steps[3];

	float	mins[3], maxs[3], units[3], grading[3], plus[3], minus[3];
	float	x, y, z, x1, y1, z1, var;
	nodno_t *nnp;
	struct ement_t *ep;
	int	simple, duple, indx;
	int	xchecked, ychecked;
	float	*xcoor, *ycoor, *zcoor;
	ushort	*xminnodes, *yminnodes, *zminnodes;

	total = (gp->gn_steps[0] + 1) * (gp->gn_steps[1] + 1)
		* (gp->gn_steps[2] + 1);

	m_for3dim = (struct for3dimcase *)
		malloc(sizeof(struct for3dimcase));

	for (i=0;i < 3;i++) {
		m_for3dim->f3d_howmany[i] = (gp->gn_steps[i] + 1);
		m_for3dim->f3d_nonodes[i] = ((gp->gn_steps[(i+1)%3] + 1)
			* (gp->gn_steps[(i+2)%3] + 1));

		m_for3dim->f3d_coors[i] = (float *)
			malloc(sizeof(float)*m_for3dim->f3d_howmany[i]);
		m_for3dim->f3d_minnodes[i] = (ushort *)
			malloc(sizeof(ushort)*m_for3dim->f3d_howmany[i]);
	}

	m_for3dim->f3d_nincres[0] =  1;
	m_for3dim->f3d_nincres[1] =  m_for3dim->f3d_howmany[0];
	m_for3dim->f3d_nincres[2] =  m_for3dim->f3d_howmany[0]
		* m_for3dim->f3d_howmany[1];

	m_for3dim->f3d_eincres[0] =  1;
	m_for3dim->f3d_eincres[1] =
		(m_for3dim->f3d_howmany[0] - 1);
	m_for3dim->f3d_eincres[2] =
		 (m_for3dim->f3d_howmany[0] - 1)
		* (m_for3dim->f3d_howmany[1] - 1);

	meshes->m_nodes = mesh_nodesbuf(total);

	for (i=0;i<3;i++) {
		grading[i]= gp->gn_grades[i];
		steps[i]  = gp->gn_steps[i];
		mins[i]   = gp->gn_mins[i];
		maxs[i]   = gp->gn_maxs[i];
		plus[i]   = maxs[i] + mins[i];
		minus[i]  = maxs[i] - mins[i];
		units[i]  = minus[i] / (float) steps[i];
	}

	/*
	 * NODE
	 */
	xchecked = ychecked = 1;
	xcoor = m_for3dim->f3d_coors[0];
	ycoor = m_for3dim->f3d_coors[1];
	zcoor = m_for3dim->f3d_coors[2];
	xminnodes = m_for3dim->f3d_minnodes[0];
	yminnodes = m_for3dim->f3d_minnodes[1];
	zminnodes = m_for3dim->f3d_minnodes[2];

	cnt = 0;
	for (i=0; i <= gp->gn_steps[2]; i++) {
	    z   = mins[2] + units[2] * (float) i;
	    var = 2.0 * z - plus[2];
	    z1  = 0.5 * sign(var) * minus[2]
		* pow(fabs(var/minus[2]), 1.0/grading[2])
		+ 0.5 * plus[2];

	    zcoor[i] = z1;
	    zminnodes[i] = cnt;

	    if (i == 0)
		m_coormin[2] = z1;
	    if (i == gp->gn_steps[2])
		m_coormax[2] = z1;

	    for (j=0; j <= gp->gn_steps[1]; j++) {
		y   = mins[1] + units[1] * (float) j;
		var = 2.0 * y - plus[1];
		y1  = 0.5 * sign(var) * minus[1]
		    * pow(fabs(var/minus[1]), 1.0/grading[1])
		    + 0.5 * plus[1];

		if (ychecked) {
		    ycoor[j] = y1;
		    yminnodes[j] = cnt;
		    if (j == 0)
			m_coormin[1] = y1;
		    if (j == gp->gn_steps[1])
			m_coormax[1] = y1;
		}

		for (k=0; k <= gp->gn_steps[0]; k++) {
		    x   = mins[0] + units[0] * (float) k;
		    var = 2.0 * x - plus[0];
		    x1  = 0.5 * sign(var) * minus[0]
			* pow(fabs(var/minus[0]), 1.0/grading[0])
			+ 0.5 * plus[0];

			if (xchecked) {/* do not need to cal each time */
				xcoor[k] = x1;
				xminnodes[k] = cnt;
				if (k == 0)
					m_coormin[0] = x1;
				if (k == gp->gn_steps[0])
					m_coormax[0] = x1;
			}

			meshes->m_nodes[cnt]->n_coor[0] = x1;
			meshes->m_nodes[cnt]->n_coor[1] = y1;
			meshes->m_nodes[cnt]->n_coor[2] = z1;

			meshes->m_nodes[cnt]->n_plane[0] = k;
			meshes->m_nodes[cnt]->n_plane[1] = j;
			meshes->m_nodes[cnt]->n_plane[2] = i;

			meshes->m_nodes[cnt]->n_no = cnt;

			/*
			 * boundary manipulations
			 */
			if (i == 0) /* bottom */
				setbit(meshes->m_nodes[cnt]->n_bdflag, 4);
			if (i == gp->gn_steps[2]) /* top */
				setbit(meshes->m_nodes[cnt]->n_bdflag, 5);

			if (j == 0) /* near */
				setbit(meshes->m_nodes[cnt]->n_bdflag, 0);
			if (j == gp->gn_steps[1]) /* far */
				setbit(meshes->m_nodes[cnt]->n_bdflag, 2);

			if (k == 0) /* left */
				setbit(meshes->m_nodes[cnt]->n_bdflag, 3);
			if (k == gp->gn_steps[0]) /* right */
				setbit(meshes->m_nodes[cnt]->n_bdflag, 1);

			/* XXX */
			cnt++;
		}
		xchecked = 0;
	    }
	    ychecked = 0;
	}
	meshes->m_nodesno = cnt;	/* total node no */

	/*
	 * ELEMENT
	 */
	total = (gp->gn_steps[0]) * (gp->gn_steps[1]) * (gp->gn_steps[2]);
	meshes->m_ements = mesh_ementsbuf(total);

	cnt = 0;
	for (i=0; i < gp->gn_steps[2]; i++) {
		duple = (gp->gn_steps[1]+1) * (gp->gn_steps[0]+1);
		for (j=0; j < gp->gn_steps[1]; j++) {
			simple = (gp->gn_steps[0]+1);

			for (k=0; k < gp->gn_steps[0]; k++) {
				indx = i * duple + j * simple + k;

				ep = meshes->m_ements[cnt];
				ep->e_proto = getproto("ElmB8n3D");
				ep->e_nodesno = 8;
				ep->e_no = cnt;
				nnp = ep->e_nodes = (nodno_t *)
					malloc(sizeof(nodno_t) * 8);
				if (ep->e_nodes == NULL)
					fatal("%s(): Hmm", __F);
				*nnp++ = indx;
				*nnp++ = indx + 1;
				*nnp++ = indx + simple;
				*nnp++ = indx + simple + 1;
				*nnp++ = indx + duple;
				*nnp++ = indx + duple + 1;
				*nnp++ = indx + duple + simple;
				*nnp++ = indx + duple + simple + 1;

				cnt++;
			}
		}
	}
	meshes->m_ementsno = cnt;	/* total element no */

	return;
}

/*
 * returns the element which contains the point (x,y) in the case of 2-dimension
 * and which contains the point (x, y, z) in the case of the 3-dimension
 *
 * If an error occurs, then negative integer will be returned, so that returned
 * -1 means that the x-coordinate is a wrong value, and returned -2 means that
 * y-coordinate is a sort of wrong value, etc.  -100 means that the mesh has
 * not data struture to calculate the element number of the point of interest. 
 */

int
get_elmno4point(float x, float y, float z)
{
	int	planes[3];
	float	coors[3], *fp;
	int	i, k, result;
	int	howmany;
	struct for3dimcase *for3;

	if (m_for3dim == NULL) {
		fprintf(stderr, "get_elmno4point(): m_for3dim == NULL");
		return -100;
	}

	for3 = m_for3dim;
	coors[0] = x;
	coors[1] = y;
	coors[2] = z;

	for (k=0;k < meshes->m_spacedim;k++) {
		fp = for3->f3d_coors[k];
		howmany = for3->f3d_howmany[k];

		if (coors[k] < fp[0] || coors[k] > fp[howmany-1])
			return -(k+1);

		for (i=0; i < (howmany-1); i++) {
			if (coors[k] < fp[i])
				break;
		}
		i--;
		if (i == (howmany-1))
			fatal("get_elmno4point(): CODING ERROR");
		planes[k] = i;
	}

	/*
	 * Now we have found the plane numbers;
	 */

	result = 0;
	for (k=0;k < meshes->m_spacedim;k++)
		result += (for3->f3d_eincres[k] * planes[k]);

	return result;
}
