#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <math.h>

#include "common.h"
#include "finite.h"


/*
 * Only puts the grid information on filename
 * this function can deal only with 2-dimensional case.
 * if needno is not zero, then numbers of the nodes are given
 *
 * filename is a full name
 * Now you can view the mesh using plotmtv utility.
 */
 
void
put_2dgrids(const char *filename, const int needno)
{
	FILE	*fp;
	char	**cpp;
	int	marker, subdo, *lseg;

	int	i, j, nodeno;
	nodno_t	*nnp;
	struct node_t *np;
	struct ement_t *ep;
	struct proto_t *proto;
	char	*headers[]= {"$ DATA=CURVE2D", "% equalscale=True", NULL};
	char	*foreach="% filltype=1 linecolor=5";

	if (meshes->m_spacedim != 2)
		return;

	fp = fopen(filename, "w");

	cpp = headers;
	while (*cpp)
		fprintf(fp, "%s\n", *cpp++);
	fprintf(fp, "\n");

	for (i=0;i<meshes->m_ementsno;i++) {
		ep    = meshes->m_ements[i];
		nnp   = ep->e_nodes;
		proto = ep->e_proto;
		lseg  = proto->p_lineseg;

		subdo = (ep->e_sdomain % 10)+1;
		fprintf(fp, "%s fillcolor=%d\n", foreach, subdo);
		for (j=0;j < (proto->p_linesegno+1); j++) {
			nodeno = ep->e_nodesno;
			np = meshes->m_nodes[nnp[lseg[j]]];
			fprintf(fp, "%7.4f %7.4f\n",np->n_coor[0],
					np->n_coor[1]);
		}
		fprintf(fp, "\n\n");
	}
	for (i=0;i<meshes->m_nodesno;i++) {
		np = meshes->m_nodes[i];
		marker = 13;
		if (np->n_bdflag)
			marker = 4;
		if (needno)
		fprintf(fp,
		    "@ point x1=%7.4f y1=%7.4f markertype=%2d linelabel=%2d\n",
			np->n_coor[0], np->n_coor[1], marker, i);
		else
			fprintf(fp,"@ point x1=%7.4f y1=%7.4f markertype=%2d\n",
				np->n_coor[0], np->n_coor[1], marker);

	}
	fclose(fp);
	printf("mtv-formatted grid info was printed at %s\n", filename);
	return;
}
