#include <stdio.h>
#include <string.h>
#include <strings.h>

#include "common.h"
#include "vector.h"
#include "finite.h"

void	vput_notcomplex(const char *filename, const struct vector_t *vp);
void	vput_complex(const char *filename, const struct vector_t *vp);

void
put_gridandvector(const struct vector_t *vp, const char *filename)
{
	if (filename == NULL || vp == NULL)
		fatal("%s(): Hmm..", __F);

	if (vp->vt_type == SP_INTEGER)
		not_yet("put_gridandvector for SP_INTEGER");

	switch (vp->vt_type) {
		case SP_SINGLE:
		case SP_DOUBLE:
			vput_notcomplex(filename, vp);
			break;
		case SP_SIMPLEX:
		case SP_DUPLEX:
			vput_complex(filename, vp);
			break;
	}
	return;
}

void
vput_notcomplex(const char *filename, const struct vector_t *vp)
{
	FILE	*fp;
	int	i, j;
	nodno_t	*nnp;
	float	*sdata;
	double	*ddata;
	struct node_t *np;
	struct ement_t *ep;
	char	*headers[]= {"$ DATA=CONTCURVE",
		"% contstyle=2  # 1=contour lines, 2=colors, 3=mesh",
		"% nsteps=50",
		"#% contours= 0.1 0.3 0.35  # each line value is specified",
		"#% comment=\"A comment in the upper right corner\"",
		"#% eyepos.x=0.5 eyepos.y=-1.5 eyepos.z=0.5 # another 3D view",
		NULL};
	char	**cpp;
	int	singleornot;

	if (vp->vt_type == SP_SINGLE)
		singleornot = 1;
	else
		singleornot = 0;

	if (vp->vt_size != meshes->m_nodesno)
		fatal("%s(): ERROR! vp->vt_size != meshes->m_nodesno", __F);

	if (meshes->m_spacedim != 2)
		return;

	fp = fopen(filename, "w");
	if (fp == NULL)
		fatal("%s(\"%s\"): can not fopen", __F, filename);

	cpp = headers;

	while (*cpp)
		fprintf(fp, "%s\n", *cpp++);

	fprintf(fp, "\n");

	if (singleornot)
		sdata =  svec(vp);
	else
		ddata =  dvec(vp);

	for (i=0;i<meshes->m_ementsno;i++) {
		ep = meshes->m_ements[i];
		nnp = ep->e_nodes;
		for (j=0;j < ep->e_nodesno; j++) {
			np = meshes->m_nodes[nnp[j]];

			if (singleornot)
				fprintf(fp, "%10.3e %10.3e %10.3e\n",
					np->n_coor[0], np->n_coor[1], 
					sdata[nnp[j]]);
			else
				fprintf(fp, "%10.3e %10.3e %10.3e\n",
					np->n_coor[0], np->n_coor[1], 
					ddata[nnp[j]]);
		}
		fprintf(fp, "\n\n");
	}
	fclose(fp);
	return;
}

void
vput_complex(const char *fname, const struct vector_t *vp)
{
	FILE	*fp[3];
	char	*tails[]={".real.mtv", ".imag.mtv", ".inten.mtv"};
	char	names[3][256];

	int	i, j;
	nodno_t	*nnp;
	struct simplex_t *cdata;
	struct duplex_t *zdata;

	struct node_t *np;
	struct ement_t *ep;
	char	*headers[]= {"$ DATA=CONTCURVE",
		"% contstyle=2  # 1=contour lines, 2=colors, 3=mesh",
		"% nsteps=50",
		"#% contours= 0.1 0.3 0.35  # each line value is specified",
		"#% comment=\"A comment in the upper right corner\"",
		"#% eyepos.x=0.5 eyepos.y=-1.5 eyepos.z=0.5 # another 3D view",
		NULL};
	char	**cpp;

	if (vp->vt_size != meshes->m_nodesno)
		fatal("%s(): ERROR! vp->vt_size != meshes->m_nodesno", __F);

	if (meshes->m_spacedim != 2)
		return;

	for (i=0;i<3;i++) {
		strcpy(names[i], fname);
		strcat(names[i], tails[i]);

		fp[i] = fopen(names[i], "w");
		if (fp[i] == NULL)
			fatal("%s(\"%s\"): can not fopen", __F, names[i]);

		cpp = headers;
		while (*cpp)
			fprintf(fp[i], "%s\n", *cpp++);

		fprintf(fp[i], "\n");

	}

	switch (vp->vt_type) {
		case SP_SIMPLEX:
			cdata = cvec(vp);
			for (i=0;i<meshes->m_ementsno;i++) {
				ep = meshes->m_ements[i];
				nnp = ep->e_nodes;
				for (j=0;j < ep->e_nodesno; j++) {
					np = meshes->m_nodes[nnp[j]];

					fprintf(fp[0], "%10.3e %10.3e %10.3e\n",
						np->n_coor[0], np->n_coor[1], 
						cdata[nnp[j]].real);
					fprintf(fp[1], "%10.3e %10.3e %10.3e\n",
						np->n_coor[0], np->n_coor[1], 
						cdata[nnp[j]].imag);
					fprintf(fp[2], "%10.3e %10.3e %10.3e\n",
						np->n_coor[0], np->n_coor[1], 
						cp_absolute(&cdata[nnp[j]]));
				}
				fprintf(fp[0], "\n\n");
				fprintf(fp[1], "\n\n");
				fprintf(fp[2], "\n\n");
			}
			break;
		case SP_DUPLEX:
			zdata = zvec(vp);
			for (i=0;i<meshes->m_ementsno;i++) {
				ep = meshes->m_ements[i];
				nnp = ep->e_nodes;
				for (j=0;j < ep->e_nodesno; j++) {
					np = meshes->m_nodes[nnp[j]];

					fprintf(fp[0], "%10.3e %10.3e %10.3e\n",
						np->n_coor[0], np->n_coor[1], 
						zdata[nnp[j]].real);
					fprintf(fp[1], "%10.3e %10.3e %10.3e\n",
						np->n_coor[0], np->n_coor[1], 
						zdata[nnp[j]].imag);
					fprintf(fp[2], "%10.3e %10.3e %10.3e\n",
						np->n_coor[0], np->n_coor[1], 
						zp_absolute(&zdata[nnp[j]]));
				}
				fprintf(fp[0], "\n\n");
				fprintf(fp[1], "\n\n");
				fprintf(fp[2], "\n\n");
			}
			break;
	}
	for (i=0;i<3;i++) {
		fclose(fp[i]);
	}
	return;
}
