/*
 * file: mesh_reset.c
 * The routines of this file will be used when someone want to change the
 * subdomain number of an element and when someone want to change the boudary
 * indicator number of a node. At each case, one has to provide a good function
 * pointer which returns the appropriate subdomain, or boundary number of
 */
#include <stdio.h>
#include "common.h"
#include "finite.h"

/*
 */
void
set_subdomain(int (*func)(const struct ement_t *ep))
{
	struct ement_t *ep;
	int	i;
	int	subdo;

	if (meshes == NULL || meshes->m_ements == NULL)
		fatal("%s(): please provide meshes first!!", __F);
		
	for (i = 0; i < meshes->m_ementsno; i++) {
		ep = meshes->m_ements[i];
		subdo = (*func)(ep);
		if (subdo > 31)
			fatal("%s(): subdo=%d is OOB", __F,  subdo);
		if (subdo < 0) {
			ep->e_sdomain = 0;
			continue;
		}
		ep->e_sdomain = subdo;
	}
	return;
}

/*
 */
void
set_boundary(int (*func)(const struct node_t *np))
{
	struct node_t *np;
	int	i;
	int	indic;

	if (meshes == NULL || meshes->m_nodes == NULL)
		fatal("%s(): please provide meshes first!!", __F);
		
	for (i = 0; i < meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		indic = (*func)(np);
		if (indic > 31)
			fatal("%s(): indic=%d is OOB", __F, indic);
		if (indic < 0) {
			np->n_bdflag = 0;
			continue;
		}

		np->n_bdflag = 0;
		setbit(np->n_bdflag, indic);
	}
	return;
}
