#include <stdio.h>
#include <math.h>
#include <strings.h>

#include "common.h"
#include "vector.h"
#include "matrix.h"
#include "sparse.h"
#include "finite.h"

void	mkright(void);
void	put_result(void);
void	radial(const struct vector_t *vp, const char *fname);
void	debugging(void);

#define TINY 1.0e-10

struct vector_t *fsrc;

int
whichbound(const struct node_t *np)
{
        float   x, y;
	float	len;

        x = np->n_coor[0];
        y = np->n_coor[1];

	len = sqrt(x*x + y*y);
        if (fabs(len - (20.0)) < TINY)
                return 0;
        if (fabs(len - (50.0)) < TINY)
                return 1;
#if 0
        if (fabs(x - (-1.0)) < TINY)
                return 0;
        if (fabs(x - (+1.0)) < TINY)
                return 2;
#endif
        return -1;
}

extern struct proto_t *currp;

void
problem(void)
{
	int        i;

	struct matrix_t *retmatf, *bmatf;
	struct proto_t  *prevproto = NULL;


	fsrc = vector(meshes->m_nodesno, NULL, SP_DOUBLE);
	mkright();
#if 0
	set_boundary(whichbound);
#endif
	sp_null(meshes->m_kmatrix);
	vec_null(meshes->m_bvector);
	for (i=0; i< meshes->m_ementsno; i++) {
		/*
		 * currp is only set at this location. You must
		 * not set currp at other places.
		 */
		currp = meshes->m_ements[i]->e_proto;
		if (prevproto != currp) {
			if (prevproto != NULL)
				m_mfree(2, retmatf, bmatf);
			create_smalls(currp, &retmatf, &bmatf);
		}

		integration(i, retmatf, bmatf);
		Kassemble(i, retmatf, meshes->m_kmatrix);
		bassemble(i, bmatf, meshes->m_bvector);
		prevproto = currp;
	}
	m_mfree(2, retmatf, bmatf);

	switch (meshes->m_datatype) {
		case SP_SINGLE:
		case SP_DOUBLE:
			ddirichlet(0, +0.0);
			ddirichlet(1, +0.0);
			ddirichlet(2, +0.0);
			ddirichlet(3, +0.0);
			ddirichlet(4, +0.0);
		break;
	}
	solveall();
	put_gridandvector(meshes->m_bvector, "a.mtv");
	put_result();
	return;
}

void
integration(const int elemno, struct matrix_t *kmatp, struct matrix_t *bmatp)
{
	int     i, j, k, l;
	double  **retmat, **bmat;
	double  **unit, **deriv, *weights, *jacob;
	double  beta, alpha, temp, weigh;

	m_mnull(2, kmatp, bmatp);
	retmat = dmat(kmatp);
	bmat   = dmat(bmatp);

	proto_einit(elemno);
	unit    = dmat(currp->p_intg->intg_trans);
	weights = currp->p_intg->intg_weights;
	jacob   = currp->p_intg->intg_jacob;

	currp = currp;
	for (k=0; k < currp->p_intgno; k++) {
		deriv = dmat(currp->p_intg->intg_deriv[k]);
		beta = dgetat(elemno, k, fsrc);
		weigh = weights[k];
		alpha = 1.0;
		if (0) {
			beta = 0.0;
		}

		for (i=0; i< currp->p_nodsno;i++) {
			for (j=0; j< currp->p_nodsno;j++) {
				temp = 0.0;
				for (l=0;l < currp->p_dimen;l++)
					temp += deriv[l][i] * deriv[l][j];

				retmat[i][j] += alpha * temp * jacob[k]
					* weigh;
			}
			bmat[i][0] +=
				-unit[k][i] * beta * jacob[k] * weigh;
		}
	}
	if ((elemno%100) == 0) {
		printf("%6d\n", elemno);
	}

	return;
}

void
mkright(void)
{
	int	i;
	float x, y;
	struct node_t *np;
	double	*dp;

	dp = dvec(fsrc);
	for (i=0; i<meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		x = np->n_coor[0];
		y = np->n_coor[1];
		dp[i] = 2.0*x*(x-1.0)+2.0*y*(y-1.0);
	}
	return;
}

void
put_result(void)
{
	int	i, issingle;
	float x, y, exact, diff;
	struct node_t *np;
	double	*dp;
	float	*sp;

	issingle = meshes->m_datatype == SP_SINGLE;
	if (issingle)
		sp = svec(meshes->m_bvector);
	else
		dp = dvec(meshes->m_bvector);

	for (i=0; i<meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		x = np->n_coor[0];
		y = np->n_coor[1];
		exact = x*(x-1.0)*y*(y-1.0);
		if (issingle) {
			diff = exact - sp[i];
			printf("%2d %7.4f %7.4f  %8.5f %8.5f %13.6e\n",
				i, x, y, sp[i], exact, diff);
		} else {
			diff = exact - dp[i];
			printf("%2d %7.4f %7.4f  %8.5f %8.5f %13.6e\n",
				i, x, y, dp[i], exact, diff);
		}
	}
	return;
}

void
radial(const struct vector_t *vp, const char *fname)
{
        FILE    *fp;
        char    names[132];
        char    *tailes=".res";
        int     i;
        struct node_t *np;
	float	x, y;
        double  radius, *dp;

	strcpy(names, fname);
	strcat(names, tailes);
	fp = fopen(names, "w");
	if (fp == NULL)
		fatal("%s():%s", __F, names);

        dp = dvec(vp);

        for (i=0; i < meshes->m_nodesno; i++) {
                np = meshes->m_nodes[i];
		x = np->n_coor[0];
		y = np->n_coor[1];

                radius = sqrt(x*x+y*y);
                fprintf(fp, "%7.5f %12.5e\n", radius, dp[i]);
        }

	fclose(fp);
        return;
}

/*
 * de0 and de1 should be same
 */
void
debugging(void)
{
	FILE	*fp;
	int	i, j;
	struct proto_t *proto;
	struct vector_t *xvec, *yvec;
	double	x, y, *dpx, *dpy;
	double	**intmat;

	fp = fopen("de0", "w");
	for (i=0;i<meshes->m_ementsno;i++) {
		proto = meshes->m_ements[i]->e_proto;
		proto_einit(i);
		intmat = dmat(proto->p_intg->intg_realintgcoors);
		for (j=0; j< proto->p_intgno; j++) {
			fprintf(fp, "%10.5e %10.5e\n",
				intmat[j][0], intmat[j][1]);
		}
	}
	fclose(fp);

	xvec = vector(meshes->m_nodesno, NULL, SP_DOUBLE);
	yvec = vector(meshes->m_nodesno, NULL, SP_DOUBLE);
	dpx = dvec(xvec);
	dpy = dvec(yvec);

	for (i=0;i<meshes->m_nodesno;i++) {
		dpx[i] = meshes->m_nodes[i]->n_coor[0];
		dpy[i] = meshes->m_nodes[i]->n_coor[1];
	}

	fp = fopen("de1", "w");
	for (i=0;i<meshes->m_ementsno;i++) {
		proto = meshes->m_ements[i]->e_proto;
		for (j=0; j< proto->p_intgno; j++) {
			x = dgetat(i, j, xvec);
			y = dgetat(i, j, yvec);
			fprintf(fp, "%10.5e %10.5e\n", x, y);
		}
	}
	fclose(fp);
	vec_mfree(2, xvec, yvec);
	return;
}
