#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include "common.h"
#include "vector.h"
#include "sparse.h"
#include "solver.h"
#include "rtc.h"

/*
 * It can solve real-symmetric case
 */

struct vector_t *
it_bicongradreal(struct sparse_t *sp, struct vector_t *x, struct vector_t *b,
	int itno, double omega, int pcd_type)
{
	int	it;
	int	dimen;
	double	alpha, beta, rho, rho_old;
	double	bnorm;
	struct vector_t *r, *r_, *p, *p_, *q,  *z, *z_;
	struct vector_t *tmp0, *tmp1;

	sp_symcheck(sp);
	iter_bontriplets(__F, sp, x, b);

	r    = vec_puredup(x);
	r_   = vec_puredup(x);
	p    = vec_puredup(x);
	p_   = vec_puredup(x);
	q    = vec_puredup(x);
	z    = vec_puredup(x);
	z_   = vec_puredup(x);
	tmp0 = vec_puredup(x);
	tmp1 = vec_puredup(x);

	dimen = x->vt_size;
	bnorm = vec_getnorm2(b);

	if (!iszero(vec_getnorm1(x)/(double) dimen))
		vec_minus(r, b, sp_dotvec(tmp0, 0, sp, x));
	else
		vec_copy(r, b);

	it = 0;
	rho_old = 0.0;

	switch (pcd_type) {
	default:
		vec_copy(r_, r);
		while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
			it++;
			if (it == 1) {
				vec_copy(p, r);
				vec_copy(p_, r_);
				rho = vec_ddotprod(r_, r);
			}

			rho_old = rho;
			sp_dotvec(q, 0, sp, p);
			alpha = rho/vec_ddotprod(p_, q);
			vec_pplus(x, vec_dcross(tmp0, alpha, p));

			vec_mminus(r, vec_dcross(tmp0, alpha, q));
			sp_transdotvec(tmp1, 0, sp, p_);
			vec_mminus(r_, vec_dcross(tmp0, alpha, tmp1));
			rho = vec_ddotprod(r_, r);

			beta = rho/rho_old;
			vec_plus(p,  r,  vec_dcross(tmp0, beta, p));
			vec_plus(p_, r_, vec_dcross(tmp0, beta, p_));


		}
		break;
	}
	
	vec_mfree(9, r, r_, p, p_, q, z, z_, tmp0, tmp1);
	return x;
}

struct vector_t *
it_bicongradcom(struct sparse_t *sp, struct vector_t *x, struct vector_t *b,
	int itno, double omega, int pcd_type)
{
	int	it;
	int	dimen;
	struct duplex_t alpha, beta, rho, rho_old, dtmp;
	struct sparse_t *diag, *diaginv;
	double	bnorm;
	struct vector_t *r, *r_, *p, *p_, *q,  *z, *z_;
	struct vector_t *tmp0, *tmp1;

	sp_symcheck(sp);
	iter_bontriplets(__F, sp, x, b);

	r    = vec_puredup(x);
	r_   = vec_puredup(x);
	p    = vec_puredup(x);
	p_   = vec_puredup(x);
	q    = vec_puredup(x);
	z    = vec_puredup(x);
	z_   = vec_puredup(x);
	tmp0 = vec_puredup(x);
	tmp1 = vec_puredup(x);

	dimen = x->vt_size;
	bnorm = vec_getnorm2(b);

	if (!iszero(vec_getnorm1(x)/(double) dimen))
		vec_minus(r, b, sp_dotvec(tmp0, 0, sp, x));
	else
		vec_copy(r, b);

	it = 0;
	rho_old = duplex_null;

	switch (pcd_type) {
	case 1:	/* not good */
		diag = sp_getdiagsp(sp);
		diaginv = sp_getinvdiag(diag);
		vec_copy(r_, r);
		while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
			it++;
			vec_dcross(z, omega, sp_dotvec(tmp0, 0, diaginv, r));
			vec_dcross(z_, omega,
					sp_transdotvec(tmp0, 0, diaginv, r_));
			vec_zdotprod2nd(&rho, 0, z, r_);

			if (it == 1) {
				vec_copy(p, z);
				vec_copy(p_, z_);
			} else {
				zp_slash(&beta, &rho, &rho_old);
				vec_plus(p, z, vec_zcross(tmp0, beta, p));
				vec_plus(p_, z_,
					vec_zcross2nd(tmp0, 0, beta, p_));
			}

			sp_dotvec(q, 0, sp, p);
			vec_zdotprod2nd(&dtmp, 0, p_, q);
			zp_slash(&alpha, &rho, &dtmp);

			vec_pplus(x, vec_zcross(tmp0, alpha, p));
			vec_mminus(r, vec_zcross(tmp0, alpha, q));
			sp_transdotvec(tmp1, 0, sp, p_);
			vec_mminus(r_, vec_zcross2nd(tmp0, 0, alpha, tmp1));

			rho_old = rho;
		}
		break;

	default:
		vec_copy(r_, r);
		while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
			it++;

			vec_zdotprod2nd(&rho, 0, r_, r);

			if (it == 1) {
				vec_copy(p, r);
				vec_copy(p_, r_);
			} else {
				zp_slash(&beta, &rho, &rho_old);
				vec_plus(p, r, vec_zcross(tmp0, beta, p));
				vec_plus(p_, r_,
					vec_zcross2nd(tmp0, 0, beta, p_));
			}

			sp_dotvec(q, 0, sp, p);
			vec_zdotprod2nd(&dtmp, 0, p_, q);
			zp_slash(&alpha, &rho, &dtmp);

			vec_pplus(x, vec_zcross(tmp0, alpha, p));
			vec_mminus(r, vec_zcross(tmp0, alpha, q));
			sp_transdotvec(tmp1, 0, sp, p_);
			vec_mminus(r_, vec_zcross2nd(tmp0, 0, alpha, tmp1));

			rho_old = rho;
		}
		break;
	}
	
	vec_mfree(9, r, r_, p, p_, q, z, z_, tmp0, tmp1);
	sp_mfree(2, diag, diaginv);
	return x;
}
