#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include "common.h"
#include "vector.h"
#include "sparse.h"
#include "solver.h"
#include "rtc.h"

struct vector_t *
it_bicongradstabcomplex(struct sparse_t *sp, struct vector_t *x,
	struct vector_t *b, int itno, double omega_notused, int pcd_type)
{
	int	it;
	int	dimen;
	struct duplex_t alpha, beta, rho, rho_old, omega;
	struct duplex_t ztmp0, ztmp1;
	double	bnorm;
	double	dtmp;
	struct vector_t *r, *r_, *p, *p_, *v, *s, *s_, *t;
	struct vector_t *tmp0, *tmp1, *tmp2;

	alpha = duplex_null;
	rho_old = duplex_null;
	omega = duplex_null;

	sp_symcheck(sp);
	iter_bontriplets(__F, sp, x, b);

	r    = vec_puredup(x);
	r_   = vec_puredup(x);
	p    = vec_puredup(x);
	p_   = vec_puredup(x);
	v    = vec_puredup(x);
	s    = vec_puredup(x);
	s_   = vec_puredup(x);
	t    = vec_puredup(x);
	tmp0 = vec_puredup(x);
	tmp1 = vec_puredup(x);
	tmp2 = vec_puredup(x);

	dimen = x->vt_size;
	bnorm = vec_getnorm2(b);

	if (!iszero(vec_getnorm1(x)/(double) dimen))
		vec_minus(r, b, sp_dotvec(tmp0, 0, sp, x));
	else
		vec_copy(r, b);

	it = 0;

	vec_copy(r_, r);

	while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
		it++;

		vec_zdotprod2nd(&rho, 0, r_, r);
		if (iszzero(&rho))
			fatal("%s():", __F);

		if (it == 1) {
			vec_copy(p, r);
		} else {
			zp_cross(&beta, zp_slash(&ztmp0, &rho, &rho_old),
				zp_slash(&ztmp1, &alpha, &omega));
			vec_zcross(tmp2, omega, v);
			vec_plus(p, r, vec_zcross(tmp0, beta,
				vec_minus(tmp1,  p, tmp2)));
		}

		sp_dotvec(v, 0, sp, p);

		vec_zdotprod2nd(&ztmp0, 0, r_, v);
		zp_slash(&alpha, &rho, &ztmp0);

		vec_zcross(tmp0, alpha, v);
		vec_minus(s, r, tmp0);
		sp_dotvec(t, 0, sp, s);

		/*
		 * 2nd arg's of followings should be 1's when hermitian cases
		 */
		vec_zdotprod2nd(&ztmp0, 1, t, s);	/* I DONT KNOW */
		vec_zdotprod2nd(&ztmp1, 1, t, t);
		zp_slash(&omega, &ztmp0, &ztmp1); 

		vec_pplus(x, vec_zcombi(tmp0, alpha, p, omega, s));
		vec_zcross(tmp0, omega, t);
		vec_minus(r, s, tmp0);

		rho_old = rho;

		if (iszzero(&omega))
			break;
	}

	vec_mfree(10, r, r_, p, p_, v, s, s_, t, tmp0, tmp1);
	return x;
}

struct vector_t *
it_bicongradstab(struct sparse_t *sp, struct vector_t *x,
	struct vector_t *b, int itno, double omega_notused, int pcd_type)
{
	int	it;
	int	dimen;
	double	alpha = 0.0, beta, rho, rho_old=0.0, omega = 0.0;
	double	bnorm;
	struct vector_t *r, *r_, *p, *p_, *v, *s, *s_, *t;
	struct vector_t *tmp0, *tmp1;

	sp_symcheck(sp);
	iter_bontriplets(__F, sp, x, b);

	r    = vec_puredup(x);
	r_   = vec_puredup(x);
	p    = vec_puredup(x);
	p_   = vec_puredup(x);
	v    = vec_puredup(x);
	s    = vec_puredup(x);
	s_   = vec_puredup(x);
	t    = vec_puredup(x);
	tmp0 = vec_puredup(x);
	tmp1 = vec_puredup(x);

	dimen = x->vt_size;
	bnorm = vec_getnorm2(b);

	if (!iszero(vec_getnorm1(x)/(double) dimen))
		vec_minus(r, b, sp_dotvec(tmp0, 0, sp, x));
	else
		vec_copy(r, b);

	it = 0;

	vec_copy(r_, r);

	while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
		it++;

		rho = vec_ddotprod(r_, r);
		if (iszero(rho))
			fatal("%s():", __F);

		if (it == 1) {
			vec_copy(p, r);
		} else {
			beta = (rho/rho_old) * (alpha/omega);
			vec_plus(p, r, vec_dcross(tmp0, beta,
				vec_dcombi(tmp1, 1.0, p, -omega, v)));
		}

		sp_dotvec(v, 0, sp, p);

		alpha = rho/vec_ddotprod(r_, v);

		vec_dcombi(s, 1.0, r, -alpha, v); 
		sp_dotvec(t, 0, sp, s);

		omega = vec_ddotprod(t, s)/pow(vec_getnorm2(t), 2.0);

		vec_pplus(x, vec_dcombi(tmp0, alpha, p, omega, s));
		vec_dcombi(r, 1.0, s, -omega, t);

		rho_old = rho;

		if (iszero(omega))
			break;
	}

	vec_mfree(10, r, r_, p, p_, v, s, s_, t, tmp0, tmp1);
	return x;
}
