#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include "common.h"
#include "vector.h"
#include "sparse.h"
#include "solver.h"
#include "rtc.h"

struct vector_t *
it_qmr(struct sparse_t *sp, struct vector_t *x, struct vector_t *b,
	int itno, double omega, int pcd_type)
{
	int	it;
	int	dimen;
	double	rho, rho_new, xi, gamma, gamma_old, eta, delta, beta;
	double	epsilon = 0.0, theta, theta_old = 0.0;
	double	bnorm, dtmp;
	struct vector_t *r, *p, *p_, *q, *y, *y_, *v, *v_;
	struct vector_t *w, *w_, *z, *z_, *s, *d;
	struct vector_t *tmp0, *tmp1;

	sp_symcheck(sp);
	iter_bontriplets(__F, sp, x, b);

	r    = vec_puredup(x);
	p    = vec_puredup(x);
	p_   = vec_puredup(x);
	q    = vec_puredup(x);
	y    = vec_puredup(x);
	y_   = vec_puredup(x);
	v    = vec_puredup(x);
	v_   = vec_puredup(x);
	w    = vec_puredup(x);
	w_   = vec_puredup(x);
	z    = vec_puredup(x);
	z_   = vec_puredup(x);
	s    = vec_puredup(x);
	d    = vec_puredup(x);
	tmp0 = vec_puredup(x);
	tmp1 = vec_puredup(x);

	dimen = x->vt_size;
	bnorm = vec_getnorm2(b);

	if (!iszero(vec_getnorm1(x)/(double) dimen))
		vec_minus(r, b, sp_dotvec(tmp0, 0, sp, x));
	else
		vec_copy(r, b);

	it = 0;

	vec_copy(v_, r);
	rho_new = vec_getnorm2(v_);
	vec_copy(w_, r);
	xi = vec_getnorm2(w_);

	gamma     =  1.0;
	eta       = -1.0;
	gamma_old = gamma;

	while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
		it++;

		rho = rho_new;
		if (iszero(rho) || iszero(xi))
			fatal("%s():", __F);

		vec_dcross(v, 1.0/rho, v_);
		vec_dcrossme(v_, 1.0/rho);

		vec_dcross(w, 1.0/xi, w_);
		vec_dcrossme(w_, 1.0/xi);

		delta = vec_ddotprod(w_, v_);

		if (iszero(delta))
			fatal("%s(): ", __F);


		if (it == 1) {
			vec_copy(p, v_);
			vec_copy(q, w_);
		} else {
			vec_minus(p, v_, vec_dcross(tmp0, xi*delta/epsilon, p));
			vec_minus(q, w_, vec_dcross(tmp0,rho*delta/epsilon, q));
		}

		sp_dotvec(p_, 0, sp, p);
		epsilon = vec_ddotprod(q, p_);

		if (iszero(epsilon))
			fatal("%s(): ", __F);

		beta = epsilon/delta;
		if (iszero(beta))
			fatal("%s():beta", __F);


		vec_minus(v_, p_, vec_dcross(tmp0, beta, v));

		rho_new = vec_getnorm2(v);
		vec_minus(w_, sp_transdotvec(tmp0, 0, sp, q),
			vec_dcross(tmp1, beta, w));

		xi = vec_getnorm2(w_);
		theta = rho_new/(gamma_old * fabs(beta));
		gamma = 1.0 / sqrt(1.0 + theta*theta);

		if (iszero(gamma))
			fatal("%s(): gamma", __F);

		eta = -eta * rho * gamma * gamma/(beta*gamma_old*gamma_old);

		if (it == 1) {
			vec_dcross(d, eta, p);
			vec_dcross(s, eta, p_);
		} else {
			dtmp = theta_old*theta_old*gamma*gamma;
			vec_dcombi(d, eta, p, dtmp, d);
			vec_dcombi(s, eta, p_, dtmp, s);
		}

		vec_pplus(x, d);
		vec_mminus(r, s);
		gamma_old = gamma;
		theta_old = theta;
	}
	vec_mfree(16, r, p, p_, q, y, y_, v, v_,
		w, w_, z, z_, s, d, tmp0, tmp1);
	return x;
}

#if 0
struct vector_t *
it_qmrcomplex(struct sparse_t *sp, struct vector_t *x, struct vector_t *b,
	int itno, double omega, int pcd_type)
{
	int	it;
	int	dimen;
	struct duplex_t rho, rho_new, xi, gamma, gamma_old, eta, delta, beta;
	struct duplex_t epsilon, theta, theta_old;
	double	bnorm, dtmp;
	struct vector_t *r, *p, *p_, *q, *y, *y_, *v, *v_;
	struct vector_t *w, *w_, *z, *z_, *s, *d;
	struct vector_t *tmp0, *tmp1;

	epsilon = duplex_null;
	theta_old = duplex_null;

	sp_symcheck(sp);
	iter_bontriplets(__F, sp, x, b);

	r    = vec_puredup(x);
	p    = vec_puredup(x);
	p_   = vec_puredup(x);
	q    = vec_puredup(x);
	y    = vec_puredup(x);
	y_   = vec_puredup(x);
	v    = vec_puredup(x);
	v_   = vec_puredup(x);
	w    = vec_puredup(x);
	w_   = vec_puredup(x);
	z    = vec_puredup(x);
	z_   = vec_puredup(x);
	s    = vec_puredup(x);
	d    = vec_puredup(x);
	tmp0 = vec_puredup(x);
	tmp1 = vec_puredup(x);

	dimen = x->vt_size;
	bnorm = vec_getnorm2(b);

	if (!iszero(vec_getnorm1(x)/(double) dimen))
		vec_minus(r, b, sp_dotvec(tmp0, 0, sp, x));
	else
		vec_copy(r, b);

	it = 0;

	vec_copy(v_, r);
	rho_new = vec_getnorm2(v_);
	vec_copy(w_, r);
	xi = vec_getnorm2(w_);

	gamma     = duplexv( 1.0, 0.0);
	eta       = duplexv(-1.0, 0.0);
	gamma_old = gamma;

	while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
		it++;

		rho = rho_new;
		if (iszzero(&rho) || iszzero(&xi))
			fatal("%s():", __F);

		zp_inverse(&ztmp0, &rho);
		vec_dcross(v, 1.0/rho, v_);
		vec_dcrossme(v_, 1.0/rho);

		vec_dcross(w, 1.0/xi, w_);
		vec_dcrossme(w_, 1.0/xi);

		vec_zdotprod2nd(&delta, 0, w_, v_);

		if (iszzero(&delta))
			fatal("%s(): ", __F);


		if (it == 1) {
			vec_copy(p, v_);
			vec_copy(q, w_);
		} else {
			
			vec_minus(p, v_, vec_dcross(tmp0, xi*delta/epsilon, p));
			vec_minus(q, w_, vec_dcross(tmp0,rho*delta/epsilon, q));
		}

		sp_dotvec(p_, 0, sp, p);
		epsilon = vec_ddotprod(q, p_);

		if (iszero(epsilon))
			fatal("%s(): ", __F);

		beta = epsilon/delta;
		if (iszero(beta))
			fatal("%s():beta", __F);


		vec_minus(v_, p_, vec_dcross(tmp0, beta, v));

		rho_new = vec_getnorm2(v);
		vec_minus(w_, sp_transdotvec(tmp0, 0, sp, q),
			vec_dcross(tmp1, beta, w));

		xi = vec_getnorm2(w_);
		theta = rho_new/(gamma_old * fabs(beta));
		gamma = 1.0 / sqrt(1.0 + theta*theta);

		if (iszzero(&gamma))
			fatal("%s(): gamma", __F);

		zp_3cross(&ztmp0, &rho, &gamma, &gamma);
		zp_3cross(&ztmp1, &beta, &gamma_old, &gamma_old);
		zp_slash(&ztmp2, &ztmp0, &ztmp1);
		zp_crossd(&ztmp0, &eta, -1.0);
		zp_cross(&eta, &ztmp0, &ztmp2);
	
		/*eta = -eta * rho * gamma * gamma/(beta*gamma_old*gamma_old)*/

		if (it == 1) {
			vec_zcross(d, eta, p);
			vec_zcross(s, eta, p_);
		} else {
			zp_cross(&ztmp0,
				&zp_cross(&ztmp1, &theta_old, &theta_old),
				&zp_cross(&ztmp2, &gamma, &gamma));
			vec_zcombi(d, eta, p, ztmp, d);
			vec_zcombi(s, eta, p_, ztmp, s);
		}

		vec_pplus(x, d);
		vec_mminus(r, s);
		gamma_old = gamma;
		theta_old = theta;
	}
	vec_mfree(16, r, p, p_, q, y, y_, v, v_,
		w, w_, z, z_, s, d, tmp0, tmp1);
	return x;
}
#endif
