#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include "sp_lu.h"
#include "sparse.h"

/*
 * a routine to be used to debug..
 */
void
splu_print(const struct spforlu_t *spf)
{
	int	i, j,  incre;
	struct splowup_t *lower, *upper;
	struct spentry_t **lsp, **usp;

	lower = spf->spf_lower;
	upper = spf->spf_upper;

	lsp = lower->splu_entries;
	usp = upper->splu_entries;

	for (i=0;i < lower->splu_size;i++) {
		incre = 0;
		for (j=0; j <  i; j++) {
			if (j == lsp[i]->spe_index[incre]) {
				printf("+");
				incre++;
			} else
				printf(".");
		}
		printf("%2d: %2d %2d\n", lsp[i]->spe_memno,
			lsp[i]->spe_mmax[0],lsp[i]->spe_mmax[1]);
	}
	for (i=0;i < upper->splu_size;i++) {
		incre = 0;
		for (j=0; j <= i; j++) {
			if (j == usp[i]->spe_index[incre]) {
				printf("+");
				incre++;
			} else
				printf(".");
		}
		printf("%2d: %2d %2d\n", usp[i]->spe_memno,
			usp[i]->spe_mmax[0],usp[i]->spe_mmax[1]);
	}
	return;
}

float
splu_all(const struct spforlu_t *spf)
{
	int	i, j;
	float	*data;
	struct splowup_t *lower, *upper;
	struct spentry_t **lsp, **usp;
	float	ret;

	lower = spf->spf_lower;
	upper = spf->spf_upper;

	lsp = lower->splu_entries;
	usp = upper->splu_entries;

	ret = 0.0;
	for (i=0;i < lower->splu_size;i++) {
		data = (float *) lsp[i]->spe_data;
		for (j=0; j <  lsp[i]->spe_memno; j++)
			ret += data[j];
	}
	for (i=0;i < upper->splu_size;i++) {
		data = (float *) usp[i]->spe_data;
		for (j=0; j <  usp[i]->spe_memno; j++)
			ret += data[j];
	}
	return ret;
}
void
splu_check(const struct spforlu_t *spf, const char *name)
{
	int	i, j,  before, error;
	struct splowup_t *lower, *upper;
	struct spentry_t **lsp, **usp;

	lower = spf->spf_lower;
	upper = spf->spf_upper;

	lsp = lower->splu_entries;
	usp = upper->splu_entries;

	error = 0;
	for (i=0;i < lower->splu_size;i++) {
		before = -1;
		for (j=0; j <  lsp[i]->spe_memno; j++) {
			if (before >= lsp[i]->spe_index[j])
				error++;
			before = lsp[i]->spe_index[j];
		}
	}
	for (i=0;i < upper->splu_size;i++) {
		before = -1;
		for (j=0; j <  usp[i]->spe_memno; j++) {
			if (before >= usp[i]->spe_index[j])
				error++;
			before = usp[i]->spe_index[j];
		}
	}
	if (error)
		fatal("%s(): at %s", __F, name);
	return;
}

/*
 * another debugging routines.
 */
void
splu_pattern(const struct spforlu_t *spf)
{
	int	i, j;
	float	*where;
	char	*p, line[80];
	int	ind;

	printf("++++++++%s++++\n", __F);
	if (spf->spf_size>79)
		return;
	ind = 0;
	for (i=0;i<spf->spf_size;i++) {
		p = line;
		for (j=0;j<spf->spf_size;j++) {
			where = splu_swhereis(spf, i, j);
			if (ind)
				*p++ = where ? '+' : '.';
			else
				if (where)
					printf("%5.1f+ ", *where);
				else
					printf("     + ");
		}
		if (ind) {
			*p = 0;
			printf("%s\n", line);
		} else {
			printf("\n");
		}
	}
	printf("++++++++%s++++\n", __F);
	return;
}

void
splu_rowput(const struct spforlu_t *spf, const int row)
{
	int	i;
	float	*where;

	if (spf->spf_size>79)
		return;
	printf("[%d] :", row);
	for (i=0;i<spf->spf_size;i++) {
		where = splu_swhereis(spf, row, i);
		if (where)
				printf("%5.1f+ ", *where);
			else
				printf("     + ");
	}
	printf("\n");
	return;
}
