#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "common.h"
#include "sp_lu.h"
#include "sparse.h"

void	splu_getcol(struct spentry_t *spep, const struct sparse_t *sp,
		const int which, const int guess);
void	splu_getrow(struct spentry_t *spep, const struct sparse_t *sp,
		const int which, const int guess);
float	splu_sgetproduct(struct spentry_t *low, struct spentry_t *upp,
	const int max);
void
toohard(void);

struct spforlu_t *
splu_init(struct sparse_t *sp)
{
	struct spforlu_t *spf;
	struct splowup_t *upp, *low;
	int	guess, offlines;
	struct spentry_t *spep;
	int	i, size;

	sp_check(__F, 1, sp);
	if (sp->sp_rown != sp->sp_coln)
		fatal("%s(): sparse_t should be square form");

	spf = (struct spforlu_t *) malloc(sizeof(struct spforlu_t));
	if (spf == NULL)
		memerr("%s(): spf", __F);

	spf->spf_lower = (struct splowup_t *)malloc(sizeof(struct splowup_t));
	if (spf->spf_lower == NULL)
		memerr("%s(): spf", __F);

	spf->spf_upper = (struct splowup_t *)malloc(sizeof(struct splowup_t));
	if (spf->spf_upper == NULL)
		memerr("%s(): spf", __F);

	offlines = ((sp->sp_nonzero / sp->sp_rown) + 1)/2;
	guess = 3 * offlines / 2;

	upp = spf->spf_upper;
	low = spf->spf_lower;
	upp->splu_size = sp->sp_coln;
	low->splu_size = sp->sp_rown;

	upp->splu_entries = (struct spentry_t **)
		malloc(sizeof(struct spentry_t *)*(sp->sp_coln));
	if (upp->splu_entries == NULL)
		memerr("%s(): [1]", __F);

	low->splu_entries = (struct spentry_t **)
		malloc(sizeof(struct spentry_t *)*(sp->sp_rown));
	if (upp->splu_entries == NULL)
		memerr("%s(): [1]", __F);

	if (sp->sp_roworder == NULL)
		sp_mkroworder(sp);

	size = sizeof(struct spentry_t) * (sp->sp_coln + sp->sp_rown);

	spep = (struct spentry_t *) malloc(size);
	if (spep == NULL)
		memerr("%s(): spep == NULL", __F);
	memset(spep, 0x00, size);	/* XXX */

	for (i=0; i < sp->sp_rown; i++, spep++) {
		low->splu_entries[i] = spep;
		splu_getrow(spep, sp, i, guess); /* ROW */
		spep->spe_flag |= SPE_ROW;
	}
	for (i=0; i < sp->sp_coln; i++, spep++) {
		upp->splu_entries[i] = spep;
		splu_getcol(spep, sp, i, guess); /* COLUMN */
		spep->spe_flag |= SPE_COL;
	}

	/*
	 * makes the space for the permutation information
	 */

	spf->spf_perm = (int *) malloc(sizeof(int) * sp->sp_rown);
	if (spf->spf_perm == NULL)
		fatal("%s(): spf->spf_perm == NULL", __F);
	
	spf->spf_parity = 1;
	spf->spf_type = sp->sp_type;
	spf->spf_size = sp->sp_rown;
	return spf;
}

/*
 * you dont have to call this function.
 */
void
splu_getcol(struct spentry_t *spep, const struct sparse_t *sp,
	const int which, const int guess)
{
	int	cnt;
	int	i;
	int	need;
	float		 *scol, *sspm;
	double		 *dcol, *dspm;
#if 0
	struct simplex_t *ccol, *cspm;
	struct duplex_t  *zcol, *zspm;
#endif

	cnt = 0;
	spep->spe_no = which;

	for (i=sp->sp_colx[which]; i < sp->sp_colx[which+1]; i++)
		if (which >= sp->sp_rowx[i])
			cnt++;

	need = (guess > cnt) ? guess : cnt;
	spep->spe_alloced = need;	/* how many entry alloced */
	spep->spe_memno = cnt;

	spep->spe_index = (int *) malloc(sizeof(int) * need);
	if (spep->spe_index == NULL)
		memerr("%s(): ", __F);
	for (i=0; i < need;i++)
		spep->spe_index[i] = -1;

	spep->spe_mmax[0] = spep->spe_mmax[1] = -1;
	switch (sp->sp_type) {
	case SP_SINGLE:
		spep->spe_data = (int *) malloc(sizeof(float) * need);
		scol = (float *) spep->spe_data;
		sspm = ssp(sp);
		cnt = 0;
		for (i=sp->sp_colx[which]; i < sp->sp_colx[which+1]; i++)
			if (which >= sp->sp_rowx[i]) {
				spep->spe_mmax[0] =
					sp->sp_rowx[sp->sp_colx[which]];

				spep->spe_index[cnt] = sp->sp_rowx[i];
				scol[cnt] = sspm[i];
				cnt++;
				spep->spe_mmax[1]=sp->sp_rowx[i];
			}
		break;
	case SP_DOUBLE:
		spep->spe_data = (int *) malloc(sizeof(double) * need);
		dcol = (double *) spep->spe_data;
		dspm = dsp(sp);
		cnt = 0;
		for (i=sp->sp_colx[which]; i < sp->sp_colx[which+1]; i++)
			if (which >= sp->sp_rowx[i]) {
				spep->spe_mmax[0] =
					sp->sp_rowx[sp->sp_colx[which]];

				spep->spe_index[cnt] = sp->sp_rowx[i];
				dcol[cnt] = dspm[i];
				cnt++;
				spep->spe_mmax[1]=sp->sp_rowx[i];
			}
		break;
	}
}

/*
 * you dont have to call this function.
 */
void
splu_getrow(struct spentry_t *spep, const struct sparse_t *sp,
	const int which, const int guess)
{
	int	cnt;
	int	i;
	int	need;
	float		 *scol, *sspm;
	double		 *dcol, *dspm;
#if 0
	struct simplex_t *ccol, *cspm;
	struct duplex_t  *zcol, *zspm;
#endif

	cnt = 0;
	spep->spe_no = which;

	for (i=sp->sp_rowindex[which]; i < sp->sp_rowindex[which+1]; i++)
		if (which >  sp->sp_colindex[i])
			cnt++;

	need = (guess > cnt) ? guess : cnt;
	spep->spe_alloced = need;	/* how many entry alloced */
	spep->spe_memno = cnt;

	spep->spe_index = (int *) malloc(sizeof(int) * need);
	if (spep->spe_index == NULL)
		memerr("%s(): ", __F);
	for (i=0;i < need;i++)
		spep->spe_index[i] = -1;


	spep->spe_mmax[0] = spep->spe_mmax[1] = -1;

	switch (sp->sp_type) {
	case SP_SINGLE:
		spep->spe_data = (int *) malloc(sizeof(float) * need);
		scol = (float *) spep->spe_data;
		sspm = ssp(sp);
		cnt = 0;

		for (i=sp->sp_rowindex[which]; i < sp->sp_rowindex[which+1];i++)
			if (which > sp->sp_colindex[i]) {
				spep->spe_mmax[0] =
					sp->sp_colindex[sp->sp_rowindex[which]];

				spep->spe_index[cnt] = sp->sp_colindex[i];
				scol[cnt] = sspm[sp->sp_dataindex[i]];
				cnt++;
				spep->spe_mmax[1]=sp->sp_colindex[i];
			}
		break;
	case SP_DOUBLE:
		spep->spe_data = (int *) malloc(sizeof(double) * need);
		dcol = (double *) spep->spe_data;
		dspm = dsp(sp);
		cnt = 0;
		for (i=sp->sp_rowindex[which]; i < sp->sp_rowindex[which+1];i++)
			if (which > sp->sp_colindex[i]) {
				spep->spe_mmax[0] =
					sp->sp_colindex[sp->sp_rowindex[which]];

				spep->spe_index[cnt] = sp->sp_colindex[i];
				dcol[cnt] = dspm[sp->sp_dataindex[i]];
				cnt++;
				spep->spe_mmax[1]=sp->sp_colindex[i];
			}
		break;
	}
	return;
}

float
splu_sgetproduct(struct spentry_t *low, struct spentry_t *upp, const int max)
{
	int	cnt0, cnt1, until0, until1, *array0, *array1;
	int	one, two;
	float	*first, *second;
	float	ret;

	array0 = low->spe_index;
	array1 = upp->spe_index;
	first  = (float *) low->spe_data;
	second = (float *) upp->spe_data;

	ret = 0.0;
	until0 = low->spe_memno;
	until1 = upp->spe_memno;
	cnt0 = cnt1 = 0;

	while (cnt0 < until0 && cnt1 < until1) {
		one = array0[cnt0];
		two = array1[cnt1];
		if (one >= max || two >= max)
			break;

		switch (INTCMP(one, two)) {
			case +1:
				cnt1++;	/* Hahaha.. you are a loser */
				break;
			case -1:
				cnt0++;	/* I can not match you */
				break;
			case  0:		/* we have equal power */
				ret += (first[cnt0] * second[cnt1]);
				cnt0++;
				cnt1++;
				break;
		}
	}
	return ret;
}

/*
 * find the location of (row, col) data pointer, if there is no such location
 * this function returns NULL pointer.
 */
float *
splu_swhereis(const struct spforlu_t *spf, const int row, const int col)
{
	struct spentry_t *spe;
	int	tail;
	int	*indx;
	int	i;
	float	*ret, *spointer;

	if (spf->spf_size <= row || spf->spf_size <= col)
		fatal("%s(): OOB", __F);

	if (row > col) {
		tail = col;
		spe = spf->spf_lower->splu_entries[row];
	} else {
		tail = row;
		spe = spf->spf_upper->splu_entries[col];
	}
	if (spe->spe_memno < 1)
		return NULL;

	ret = NULL;
	indx = spe->spe_index;
	spointer = (float *) spe->spe_data;

	for (i=0; i < spe->spe_memno; i++, indx++)
		if (*indx == tail)
			break;
	if (i != spe->spe_memno)
		ret = &spointer[i];
	return ret;
}

int
splu_findmeat(struct spentry_t *target, const int this)
{
	int	lim, cmp;
	int	*p, *base, *fixed;

	fixed = base = target->spe_index;
	for (lim = target->spe_memno; lim != 0; lim >>= 1) {
		p = base + (lim >> 1);
		cmp = *p < this ? 1 : *p > this ? -1 : 0;
		if (cmp == 0) {
			return p-fixed; /* already exists */
		}
		if (cmp > 0) {  /* key > p: move right */
			base = (p + 1);
			lim--;
		} /* else move left */
	}
	return -1;
}

/*
 * find the location of (row, col) data pointer, if there is no such location
 * alloc one
 */
float *
splu_sgiveme(const struct spforlu_t *spf, const int row, const int col)
{
	int	i, before, error;
	struct spentry_t *target;
	int	tail, lim, cmp;
	int	*p;
	float	*spointer;
	int	*base, *fixed, fromhere;

	if (spf->spf_size <= row || spf->spf_size <= col)
		fatal("%s(): OOB", __F);

	if (row > col) {
		tail = col;
		target = spf->spf_lower->splu_entries[row];
	} else {
		tail = row;
		target = spf->spf_upper->splu_entries[col];
	}

	spointer = (float *) target->spe_data;
	/*
	 * binary search
	 */
	fixed = base = target->spe_index;
	fromhere = 0;
	cmp = 0;
	for (lim = target->spe_memno; lim != 0; lim >>= 1) {
		p = base + (lim >> 1);
		fromhere = p - fixed;
		cmp = *p < tail ? 1 : *p > tail ? -1 : 0;
		if (cmp == 0) {
			return &spointer[(int)(p-fixed)];/* exists */
		}
		if (cmp > 0) {  /* key > p: move right */
			base = (p + 1);
			lim--;
		} /* else move left */
	}
	if (cmp > 0)
		fromhere++;

	/*
	 * increase the buffer
	 */
	if (target->spe_memno == target->spe_alloced) {
		target->spe_alloced += 20;
		target->spe_data = (int *) realloc(target->spe_data,
			sizeof(float) * target->spe_alloced);
		target->spe_index = (int *) realloc(target->spe_index,
			sizeof(int) * target->spe_alloced);
	}

	spointer = (float *) target->spe_data;
	memmove(&target->spe_index[fromhere+1], &target->spe_index[fromhere],
		sizeof(int) * (target->spe_memno - fromhere));
	memmove(&spointer[fromhere+1], &spointer[fromhere],
		sizeof(float) * (target->spe_memno - fromhere));
	target->spe_memno++;
	target->spe_index[fromhere] = tail;

#if DEBUG
	before = -2;
	error = 0;
	printf("[%2d:%2d] cmp=%2d HERE=%2d tot=%2d: ",
		row, col, cmp, fromhere, target->spe_memno);
	for (i=0;i<target->spe_memno;i++) {
		if (before > target->spe_index[i])
			error++;
		before = target->spe_index[i];
		printf("%3d", before);
	}
	if (error)
		fatal("%s(): [1]", __F);
	printf("\n");
#endif
		
	return &spointer[fromhere];
}

float *
splu_sgivemeone(struct spentry_t *target, const int this)
{
	int	i, before, error;
	int	lim, cmp;
	int	*p;
	float	*spointer;
	int	*base, *fixed, fromhere;


	spointer = (float *) target->spe_data;
	/*
	 * binary search
	 */
	fixed = base = target->spe_index;
	fromhere = 0;
	for (lim = target->spe_memno; lim != 0; lim >>= 1) {
		p = base + (lim >> 1);
		fromhere = p - fixed;
		cmp = *p < this ? 1 : *p > this ? -1 : 0;
		if (cmp == 0) {
			return &spointer[(int)(p-fixed)];/* exists */
		}
		if (cmp > 0) {  /* key > p: move right */
			base = (p + 1);
			lim--;
		} /* else move left */
	}
	if (cmp > 0)
		fromhere++;

	/*
	 * increase the buffer
	 */
	if (target->spe_memno == target->spe_alloced) {
		target->spe_alloced += 20;
		target->spe_data = (int *) realloc(target->spe_data,
			sizeof(float) * target->spe_alloced);
		target->spe_index = (int *) realloc(target->spe_index,
			sizeof(int) * target->spe_alloced);
	}

	spointer = (float *) target->spe_data;
	memmove(&target->spe_index[fromhere+1], &target->spe_index[fromhere],
		sizeof(int) * (target->spe_memno - fromhere));
	memmove(&spointer[fromhere+1], &spointer[fromhere],
		sizeof(float) * (target->spe_memno - fromhere));
	target->spe_memno++;
	target->spe_index[fromhere] = this;
#if DEBUG
	before = -2;
	error = 0;
	/*
	printf("[%2d:%2d] cmp=%2d HERE=%2d tot=%2d: ",
		row, this, cmp, fromhere, target->spe_memno);
	 */
	for (i=0;i<target->spe_memno;i++) {
		if (before > target->spe_index[i])
			error++;
		before = target->spe_index[i];
		printf("%3d", before);
	}
	printf("\n");
	if (error)
		fatal("%s(): [1]", __F);
#endif

	return &spointer[fromhere];
}

void
spe_free(struct spentry_t *spe)
{
	free(spe->spe_index);
	free(spe->spe_data);
	return;
}

struct spentry_t *
splu_getspentry(const int alloced)
{
	struct spentry_t *ret;
	ret = (struct spentry_t *) malloc(sizeof(struct spentry_t));
	if (ret == NULL)
		fatal("%s(): ret == NULL", __F);
	ret->spe_alloced = alloced;
	ret->spe_index = (int *) malloc(sizeof(int)*ret->spe_alloced);
	ret->spe_data  = (int *) malloc(sizeof(float)*ret->spe_alloced);
	ret->spe_memno = 0;

	if (ret->spe_index == NULL || ret->spe_data == NULL)
		memerr("%s(): [0]", __F);
	return ret;
} 

void
splue_print(const struct spentry_t *spe)
{
	int	i;
	int	*indx;
	float	*data;

	indx = spe->spe_index;
	data = (float *) spe->spe_data;

	printf("%3d\n", spe->spe_memno);

	for (i=0; i< spe->spe_memno;i++)
		printf("%6d ", indx[i]);
	printf("\n");
	for (i=0; i< spe->spe_memno;i++)
		printf("%6.2f ", data[i]);
	printf("\n");
	return;
}

/*
 * I think one is smaller than two
 */

void
spluup_changerow(struct spentry_t *onthis, const int one, const int two)
{
	int	locus0, locus1;
	float	*list, sone, *spoin;

	locus0 = splu_findmeat(onthis, one);
	locus1 = splu_findmeat(onthis, two);

	list = (float *) onthis->spe_data;

	if (locus0 == -1 && locus1 == -1)
		return;
	if (locus0 != -1 && locus1 != -1) {
		sone = list[locus0];
		list[locus0] = list[locus1];
		list[locus1] = sone;
		return;
	}

	/*
	 * only one does not exist.
	 * then locus'es should be changed.
	 * TAKE CARE OF THE REALLOCATION on splu_sgivemeone()!!!!
	 */

	if (locus0 == -1) {
		sone = list[locus1];
		spoin = splu_sgivemeone(onthis, one);
		*spoin = sone;
		list = (float *) onthis->spe_data;
		list[locus1+1] = 0.0;
		return;
	}
	if (locus1 == -1) {
		sone = list[locus0];
		spoin = splu_sgivemeone(onthis, two);
		*spoin = sone;
		list = (float *) onthis->spe_data;
		list[locus0] = 0.0;
		return;
	}
	return;
}


/*
 *  +------------------------------+
 *  +
 * change two rows
 * this and that can not be same
 */
void
splu_rowchange(struct spforlu_t *spf, const int this, const int that)
{
	struct splowup_t *low, *upp;
	struct spentry_t *thisp, *thatp, *thistmp, *thattmp;
	int	until, after;
	float	*srcdatax, *dstdatax, *spoin0, *spoin1;
	int	*srcindex, *dstindex;
	int	savedindex;
	int	i, cnt;

#if 0
	until = this < that ? this : that; /* min */
	after = this > that ? this : that; /* max */
#endif
	if (this >= that)
		fatal("%s(): this >= that", __F);
	until = this;
	after = that;

	low = spf->spf_lower;
	upp = spf->spf_upper;
	thisp = low->splu_entries[this];
	thatp = low->splu_entries[that];

	srcindex = thisp->spe_index;
	srcdatax = (float *) thisp->spe_data;

	cnt = 0;
	thattmp = splu_getspentry(5);
	thattmp->spe_no = that;

	for (i=0; i< thisp->spe_memno;i++) {
		savedindex = srcindex[i];
		if (savedindex >= until)
			break;
		if (thattmp->spe_memno == thattmp->spe_alloced) {
			thattmp->spe_alloced += 20;
			thattmp->spe_data = (int *) realloc(thattmp->spe_data,
				sizeof(float) * thattmp->spe_alloced);
			thattmp->spe_index = (int *) realloc(thattmp->spe_index,
				sizeof(int) * thattmp->spe_alloced);
		}
		dstindex = thattmp->spe_index;
		dstdatax = (float *) thattmp->spe_data;

		dstindex[cnt] = savedindex;
		dstdatax[cnt] = srcdatax[i];
		thattmp->spe_memno++;
		cnt++;
	}

	for (i=until; i <after; i++) {
		spoin0 = splu_swhereis(spf, this, i);
		if (spoin0 == NULL)
			continue;

		if (thattmp->spe_memno == thattmp->spe_alloced) {
			thattmp->spe_alloced += 20;
			thattmp->spe_data =(int *) realloc(thattmp->spe_data,
				sizeof(float) * thattmp->spe_alloced);
			thattmp->spe_index = (int *) realloc(thattmp->spe_index,
				sizeof(int) * thattmp->spe_alloced);
		}
		dstindex = thattmp->spe_index;
		dstdatax = (float *) thattmp->spe_data;

		dstindex[cnt] = i;
		dstdatax[cnt] = *spoin0;
		*spoin0 = 0.0;	/* XXX */
		thattmp->spe_memno++;
		cnt++;
	}

	/*
	 * end of thattmp, this will be that'th row in the end, now deal
	 * with thistmp;
	 */
	thistmp = splu_getspentry(5);
	thistmp->spe_no = this;

	srcindex = thatp->spe_index;
	srcdatax = (float *) thatp->spe_data;

	cnt = 0;	/* XXX */
	for (i=0; i< thatp->spe_memno;i++) {
		savedindex = srcindex[i];
		if (savedindex >= until)
			break;
		if (thistmp->spe_memno == thistmp->spe_alloced) {
			thistmp->spe_alloced += 20;
			thistmp->spe_data = (int *) realloc(thistmp->spe_data,
				sizeof(float) * thistmp->spe_alloced);
			thistmp->spe_index = (int *) realloc(thistmp->spe_index,
				sizeof(int) * thistmp->spe_alloced);
		}
		dstindex = thistmp->spe_index;
		dstdatax = (float *) thistmp->spe_data;

		dstindex[cnt] = savedindex;
		dstdatax[cnt] = srcdatax[i];
		thistmp->spe_memno++;
		cnt++;
	}

	/*
	 * end of thistmp, this will be this'th row in the end, now deal
	 * with thistmp;
	 */
	for (i=until; i < after; i++) {
		spoin0 = splu_swhereis(spf, this, i);
		spoin1 = splu_swhereis(spf, that, i);
		if (spoin1) {
			if (spoin0 == NULL)
				spoin0 = splu_sgiveme(spf, this, i);

			*spoin0 = *spoin1;
			continue;
		}
		if (spoin0)
			if ((thisp->spe_alloced - thisp->spe_memno) > 10)
				/* XXX --free me */ ;
	}


	for (i=after; i < spf->spf_size; i++)
		spluup_changerow(upp->splu_entries[i], this, that);

	spe_free(low->splu_entries[this]);
	spe_free(low->splu_entries[that]);

	low->splu_entries[this] = thistmp;
	low->splu_entries[that] = thattmp;

	return;
}

#define TINY 1.0e-20

struct spforlu_t *
sp_ludecomp(struct sparse_t *sp, const int dopivot)
{
	struct spforlu_t *spfret;
	struct spentry_t **low, **upp;
	int	i, j;
	int	imax;
	float	big, dum, sum, entre, *data;
	float	*here;
	int	*indx;
	int	coln;
	struct vector_t *svp;
	float	*sv;

	if (dopivot == 0) {
		printf("%s(): %s without pivot can result in wrong result\n",
			__F, __F);
	}
 
	spfret = splu_init(sp);
	printf("end of splu_init() at sp_ludecomp()\n");
	splu_check(spfret, __F);
	coln = sp->sp_coln;
	indx = spfret->spf_perm;

	if (dopivot) {
		svp = vector(coln, NULL, SP_SINGLE);
		sv  = svec(svp);
		data= ssp(sp);
		for (i=0;i < coln; i++) {
			big = 0.0;
			for (j=sp->sp_rowindex[i];j<sp->sp_rowindex[i+1]; j++) {
				entre = fabs(data[sp->sp_dataindex[j]]);
				if (entre > big)
					big = entre;
			}
			if (big == 0.0)
				fatal("%s(): at %d'th line", __F, __LINE__);
			sv[i] = 1.0/big;
		}
	}

	low = spfret->spf_lower->splu_entries;
	upp = spfret->spf_upper->splu_entries;

	for (j=0; j < coln; j++) {	/* for each col */
		for (i=0; i < j; i++) {
			here = splu_swhereis(spfret, i, j);
			sum  = here ? *here : 0.0;
			sum -= splu_sgetproduct(low[i], upp[j], i);

			if (fabs(sum) < 1.0e-20 && here == NULL)
				continue;

			if (here)
				*here = sum;
			else {
				here  = splu_sgiveme(spfret, i, j);
				*here = sum;
			}
		}
		big = 0.0;
		for (i=j; i < coln; i++) {
			here = splu_swhereis(spfret, i, j);
			sum  = here ? *here : 0.0;
			sum -= splu_sgetproduct(low[i], upp[j], j);

			if (fabs(sum) < 1.0e-20 && here == NULL)
				/* void */;

			else {
				if (here)
					*here = sum;
				else {
					here  = splu_sgiveme(spfret, i, j);
					*here = sum;
				}
			}

			if (dopivot)
				if ((dum=sv[i] * fabs(sum)) >= big) {
					big = dum;
					imax = i;
				}
		}
		if (dopivot)
			if (j != imax) {
				if (j > imax)
					splu_rowchange(spfret, imax, j);
				else if (j < imax)
					splu_rowchange(spfret, j, imax);

				spfret->spf_parity *= -1;
				sv[imax] = sv[j];
			}

		indx[j] = imax;

		here = splu_swhereis(spfret, j, j);
		if (*here == 0.0)
			*here = TINY;
	
		if (j != (coln-1)) {
			dum = 1.0/(*here);
			for (i=j+1; i < coln; i++) {
				here = splu_swhereis(spfret, i, j);
				if (here)
					*here = *here * dum;
			}
		}
	}
	if (dopivot)
		vec_free(svp);
	printf("end of sp_ludecomp()\n");
	return spfret;
}

void
sp_lubacksub(struct spforlu_t *spf, struct vector_t *bv)
{
	int	i, ii= -1, ip, j, n;
	struct spentry_t *spe;
	float	sum;
	float	*b, *data;
	int	*perm;
	int	*arrayx;

	n = spf->spf_size;
	b = svec(bv);
	perm = spf->spf_perm;

	for (i = 0;i < n;i++) {
		spe = spf->spf_lower->splu_entries[i];
		arrayx = spe->spe_index;
		data   = (float *) spe->spe_data;

		ip=perm[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii > -1) {
			for (j=0; j < spe->spe_memno; j++) {
				if (arrayx[j] < ii)
					continue;
				if (arrayx[j] >= i)
					break;
				sum -= data[j]*b[arrayx[j]];
			}
		}
		else if (sum)
			ii=i;
		b[i]=sum;
	}
	for (i = (n-1);i >= 0;i--) {
		sum=b[i];
		for (j=i+1; j < spf->spf_size; j++) {
			data = splu_swhereis(spf, i, j);
			if (data)
				sum -= (*data) * b[j];
		}
		data = splu_swhereis(spf, i, i);
		b[i]=sum/(*data);
	}
	return;
}
