#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>

struct token {
	char	*t_name;
	int	 t_namlen;
	float	 t_value;
	int	 t_setted;
	struct token *t_next;
	int	 t_donot;
	int	 t_lineno;
};

float mins[3], maxs[3], step[3], deltas[3];
struct token *troot;

void    usage(void);
void    help(void);
void	token_init(void);
void	token_add(char *, int donot);
void	initialize(char *filename);
void	getit(char *line, int lineno);
void	token_put(void);
void	output(char *getf, char *putf, char *line);
void	process(char *, char *);
char	*getstr(char *line);
float	token_get(const char *name);

int	lastline;

int
main(int argc, char **argv)
{
	FILE    *fin, *fout;
	char    fname[128], input[BUFSIZ];
	char	command[256];
	char    *p;
	int	i;

	if (argc != 2)
		usage();
	token_init();

	if (access(argv[1], R_OK)) {
		fprintf(stderr, "fopen(\"%s\"): error\n", argv[1]);
		usage();
	}
	strcpy((char *) fname, "/tmp/");
	strcat(fname, argv[1]);
	strcpy(command, "plotmtv -3d ");
	strcat(command, "-pbg white -pfg black ");
	strcat(command, fname);
	strcat(command, " &");

	initialize(argv[1]);
	token_put();

	mins[0] = token_get("xmin");
	maxs[0] = token_get("xmax");
	step[0] = token_get("nx");
	mins[1] = token_get("ymin");
	maxs[1] = token_get("ymax");
	step[1] = token_get("ny");
	mins[2] = token_get("zmin");
	maxs[2] = token_get("zmax");
	step[2] = token_get("nz");

	for (i=0;i<3;i++) 
		deltas[i] = (maxs[i] - mins[i])/step[i];

	printf("cmd> ");
	while (fgets(input, BUFSIZ, stdin)) {
		p = getstr(input);
		if (p == NULL) {
			printf("cmd> ");
			continue;
		}

		output(argv[1], fname, p);
		system(command);
		free((void *) p);
		printf("cmd> ");
		sleep(5);
	}

	return 0;
}

void
usage(void)
{
	fprintf(stderr, "usage: executable mtv-filename\n");
	exit(1);
}


void
token_add(char *name, int donot)
{
	struct token *tp, *tq;

	if (strlen(name) == 0)
		exit(1);

	if (troot == NULL) {
		troot = (struct token *)
			malloc((size_t) sizeof(struct token));
		tp = troot;
		tp->t_next = NULL;
	} else {
		for (tp = troot; tp; tp = tp->t_next) {
			if (strcmp(tp->t_name, name) == 0)
				fprintf(stderr,
					"token_add(\"%s\"): duplicated\n",
					name); 
			tq = tp;
		}
		tp = (struct token *) malloc((size_t)sizeof(struct token));
		tq->t_next = tp;
	}
	tp->t_name = strdup(name);
	tp->t_namlen = strlen(name);
	tp->t_next = NULL;
	tp->t_value = 0.0;
	tp->t_donot  = donot;
	tp->t_lineno = -1;
	return;
}

float
token_get(const char *name)
{
	struct token *tp;

	for (tp = troot; tp; tp = tp->t_next) {
		if (strlen(name) == tp->t_namlen && !strcmp(tp->t_name, name))
			break;
	}
	if (tp == NULL)
		return 0.0;	/* default */
	return tp->t_value;
}

void
token_init(void)
{
	token_add("vxmin", 1);
	token_add("vxmax", 1);
	token_add("vymin", 1);
	token_add("vymax", 1);
	token_add("vzmin", 1);
	token_add("vzmax", 1);

	token_add("xmin", 0);
	token_add("xmax", 0);
	token_add("ymin", 0);
	token_add("ymax", 0);
	token_add("zmin", 0);
	token_add("zmax", 0);
	token_add("nx", 0);
	token_add("ny", 0);
	token_add("nz", 0);
	return;
}

void
initialize(char *filename)
{
	char	line[BUFSIZ], *p;
	FILE	*fin;
	int	lineno;

	fin = fopen(filename, "r");
	if (fin == NULL) {
		fprintf(stderr,
			"initialize():fopen(\"%s\"): error\n", filename);
		exit(1);
	}

	lineno = 0;
	while (fgets(line, BUFSIZ, fin)) {
		lineno++;
		p = (char *) line;
		if (*p != '%') 
			continue;
		getit(line, lineno);
		lastline = lineno;
	}
	fclose(fin);
	return;
}

#define SEP	"\t "

void
getit(char *line, int lineno)
{
	char *p, *q;
	struct token *tp;
	int	len, readn;
	float	val;

	p = line;
	p++;	/* to skip % */

	for (q = strtok(p, SEP); q; q = strtok(NULL, SEP)) {
		p = rindex(q, '=');
		if (p == NULL) {
			continue;
		}
		*p = 0;
		readn = sscanf(++p, "%f", &val);
		len = strlen(q);
		if (readn != 1) {
			continue;
		}
		for (tp = troot; tp; tp = tp->t_next) {
			if (len == tp->t_namlen
			    && !strncmp(tp->t_name, q, len)) { /* found */
				tp->t_value = val;
				tp->t_lineno = lineno;
			}
		}
	}
	return;
}

void
token_put(void)
{
	struct token *tp;

	for (tp = troot; tp; tp = tp->t_next) {
		if (tp->t_donot && tp->t_lineno != -1)
			putchar(7);
		printf("%s: val = %8.4f\n", tp->t_name, tp->t_value);
	}
	return;
}

void
output(char *getf, char *putf, char *line)
{
	FILE	*fin, *fout;
	int	cnt;
	char	input[BUFSIZ];

	fin  = fopen(getf, "r");
	fout = fopen(putf, "w");
	if (fin == NULL || fout == NULL) {
		printf("AAAAAAAAAAAAAAAAAAAAAAAAAAACH!!\n");
		exit(1);
	}

	cnt = 0;
	while (fgets(input, BUFSIZ, fin)) {
		cnt++;
		if (cnt == (lastline+1))
			fprintf(fout, "%s\n", line);
		fprintf(fout, "%s", input);
	}
	fclose(fin);
	fclose(fout);
	return;
}


char *
getstr(char *line)
{
	char	*p, *q;
	char 	buf[BUFSIZ], label[32];
	int	readn;
	static float	value;
	static int indicator = -1;
	int	onoff;
	float	direc;

	onoff = 0;
	for (p = line; isspace(*p); p++)
		/* void */ ;
	if (*p == 'q')
		exit(0);
	if (*p == 'h' || *p == '?') {
		help();
		return NULL;
	} else if (*p == 'n' || *p == 'p') {
		onoff = 1;
		direc = (*p == 'n') ? 1.0 : -1.0;
		switch (indicator) {
			case 0:
				*p = 'x';
				break;
			case 1:
				*p = 'y';
				break;
			case 2:
				*p = 'z';
				break;
			case -1:
				*p = 'x';
				value = mins[0];
		}
	}


	if (onoff == 0) {
		q = index(p, '=');
		if (q == NULL)
			return NULL;

		readn = sscanf(++q, "%f", &value);
		if (readn != 1)
			return NULL;
	}
	
	switch (*p) {
		case 'x':
			if (onoff)
				value += direc * (deltas[0]/2.0);
			if (value < mins[0])
				value = mins[0];
			else if (value >= (maxs[0]-0.001))
				value = maxs[0];
				

			sprintf(label, "x=%5.3f", value);
			sprintf(buf,
				"%% toplabel=%s\n%% vxmin=%7.4f vxmax=%7.4f",
				label, value, value+0.001);
			indicator = 0;
			break;

		case 'y':
			if (onoff)
				value += direc * (deltas[1]/2.0);
			if (value < mins[1])
				value = mins[1];
			else if (value >= (maxs[1]-0.001))
				value = maxs[1];

			sprintf(label, "y=%5.3f", value);
			sprintf(buf,
				"%% toplabel=%s\n%% vymin=%7.4f vymax=%7.4f",
				label, value, value+0.001);
			indicator = 1;
			break;

		case 'z':
			if (onoff)
				value += direc * (deltas[2]/2.0);
			if (value < mins[2])
				value = mins[2];
			else if (value >= (maxs[2]-0.001))
				value = maxs[2];

			sprintf(label, "z=%5.3f", value);
			sprintf(buf,
				"%% toplabel=%s\n%% vzmin=%7.4f vzmax=%7.4f",
				label, value, value+0.001);
			indicator = 2;
			break;
		default:
			return NULL;
	}
	return strdup(buf);
}

void
help(void)
{
	printf("\nHELP message\n");
	printf("example: cmd> x=5\n");
	printf("example: cmd> y=5\n");
	printf("example: cmd> z=5\n");
	printf("example: cmd> p\n");
	printf("example: cmd> n\n");
	printf("example: cmd> quit\n");
	printf("example: cmd> help\n");
	return;
}
